#define INITGUID
#include <d3dx8.h>
#include <dinput.h>
#include <dmusicc.h>
#include <dmusici.h>
#include <cguid.h>
#include <mmsystem.h>
#include <stdio.h>
#include "dxutils/mesh.h"
#include "dxutils/utils.h"
#include "dxutils/control.h"
#include "dxutils/dmutil.h"
#include "dxutils/dsutil.h"
#include "dxutils/dxutil.h"
#include "dxutils/d3dfont.h"
#include "resource.h"

#define NUM_FIELD_TYPES 12
#define NUM_OBJECT_TYPES 9
#define NUM_SOUNDS 17
#define NUM_PARTICLE_TYPES 1
#define NUM_SOUND_CLONES 4
#define NUM_PLAYER_WALK_STEPS 8
#define NUM_PLAYER_PUSH_STEPS 8
#define NUM_FIRE_SPRITES 4
#define MAX_OBJECTS 64
#define MAX_SNOWFLAKES 256
#define MAX_PARTICLES 256
#define MAX_COMMANDS 256
#define MAX_PARAMETERS 16
#define PLAYER_IMMORTALITY 20.0f
#define PLAYER_WALK_STEP_SPEED 0.4f
#define PLAYER_PUSH_STEP_SPEED 0.2f
#define PLAYER_BEGIN_FALL 0.28f
#define PLAYER_BEGIN_HURT_FALL 0.35f
#define JUMP_SPEED 0.2f
#define JUMP_COUNT_DOWN 10.0f
#define GRAVITY 0.015f
#define SPEED_LOWERING 0.1f
#define ACCELERATION 0.02f
#define MAX_SPEED 0.2f
#define MAX_SPRING_SPEED 0.35f
#define LEVEL_FINISH_DELAY 50.0f
#define MENU_CURSOR_DELAY 4.0f
#define CONFIRM_QUIT_DELAY 10.0f
#define DIE_DELAY 150.0f
#define STANDARD_CAMERA_DISTANCE 10.0f
#define MAX_CAMERA_DISTANCE 20.0f
#define MIN_CAMERA_DISTANCE 1.0f
#define MAX_CAMERA_HEIGHT 5.0f
#define MIN_CAMERA_HEIGHT -2.0f
#define CAMERA_LOCK_DELAY 10.0f

#define FF_WALL 1
#define FF_TRANSPARENT 2
#define FF_RADIUS_XZ 4
#define FF_CUBE 8
#define FF_CUBEMESH 16
#define FF_SPRING 32

#define S_BLOCK_FALLING 0
#define S_STEP_LEFT 1
#define S_STEP_RIGHT 2
#define S_JUMP1 3
#define S_JUMP2 4
#define S_JUMP3 5
#define S_PUSH_OBJECT 6
#define S_DIAMOND 7
#define S_LEVEL_FINISHED 8
#define S_SHOOT 9
#define S_SNOWBALL_HIT 10
#define S_LOOSE_ENERGY1 11
#define S_LOOSE_ENERGY2 12
#define S_SWITCH 13
#define S_FIELD_CHANGES 14
#define S_LEVEL_EXIT_SIGNAL 15
#define S_SPRING 16

#define OF_WALL 1
#define OF_PUSHABLE 2
#define OF_FLYING 4
#define OF_DESTROYABLE 8
#define OF_TRANSPARENT 16
#define OF_EXTRA 32
#define OF_SHINING 64
#define OF_SWINGING_X 128
#define OF_SWINGING_Z 256
#define OF_RECEIVE_SHADOWS 512
#define OF_NO_MESH 1024
#define OF_BILLBOARD 2048
#define OF_SPIKE 4096
#define OF_FIRE 8192

#define OT_BLOCK1 0
#define OT_DIAMOND 1
#define OT_LEVEL_EXIT 2
#define OT_PLATFORM1 3
#define OT_PLATFORM2 4
#define OT_ICE1 5
#define OT_CANNON1 6
#define OT_SNOWBALL 7
#define OT_FIRE 8

#define CP_LEFT 0
#define CP_RIGHT 1
#define CP_BACK 2
#define CP_FRONT 3
#define CP_BOTTOM 4
#define CP_TOP 5

#define PT_SNOW 0

#define CT_SET_FIELD 0
#define CT_SET_CAMERA 1
#define CT_DELAY 2
#define CT_PLAY_SOUND 3

#define BOX_INSIDE 0
#define BOX_CLIPPED 1
#define BOX_OUTSIDE 2

class CFieldType
{
public:    
    int         m_iFlags;
    int         m_iType;
    float       m_fRadiusXZ;
    D3DXVECTOR3 m_vMinimum;
    D3DXVECTOR3 m_vMaximum;
    D3DXVECTOR3 m_vCenter;
    CMesh       *m_pMesh;
};

class CObjectType
{
public:
    int         m_iFlags;
    float       m_fWidth;
    float       m_fHeight;
    float       m_fDepth;
    float       m_fShadowSize;
    float       m_fRadius;
    float       m_fReload;
    D3DXVECTOR3 m_vAdd;
    D3DXVECTOR3 m_vShiningPosition;
    D3DCOLOR    m_cShiningColor;
    CMesh       *m_pMesh;
};

class CObject
{
public:
    int         m_iType;
    int         m_iParent;
    float       m_fCounter;
    float       m_fReload;
    float       m_fLife;
    CControl    m_Control;
    D3DXVECTOR3 m_vDirection;
    D3DXVECTOR3 m_vAdd;
    BOOL        m_bOnGround;
    BOOL        m_bExists;    
    CObjectType *m_pType;
};

class CPlayer
{
public:
    BOOL        m_bOnGround;
    BOOL        m_bPush;
    int         m_iFX;
    int         m_iFY;
    int         m_iFZ;    
    int         m_iNumDiamonds;
    float       m_fFalling;
    float       m_fSpeed;
    float       m_fStep;
    float       m_fJumpCountDown;
    float       m_fEnergy;
    float       m_fImmortality;
    CMesh       *m_pMeshWalk[NUM_PLAYER_WALK_STEPS];
    CMesh       *m_pMeshPush[NUM_PLAYER_PUSH_STEPS];
    CMesh       *m_pMeshJump;
    CMesh       *m_pMeshFall;
    CControl    m_Control;
};

class CSnowflake
{
public:
    D3DXVECTOR3 m_vPosition;
};

class CParticle
{
public:
    int         m_iType;
    float       m_fLife;
    D3DCOLOR    m_cColor;
    D3DXVECTOR3 m_vPosition;
    D3DXVECTOR3 m_vDirection;
};

class CCommand
{
public:
    BOOL        m_bExists;
    int         m_iType;
    int         m_iParameter[MAX_PARAMETERS];
    float       m_fParameter[MAX_PARAMETERS];
};

class CScript
{
public:
    int         m_iCurrentCommand;
    CCommand    m_Command[MAX_COMMANDS];

    void        Execute(float);
    void        AddSetFieldCommand(int, int, int, int);
    void        AddSetCameraCommand(D3DXVECTOR3, D3DXVECTOR3, float);
    void        AddDelayCommand(float);
    void        AddPlaySoundCommand(int, D3DXVECTOR3);
    int         GetFreeCommand();
};

LPDIRECT3D8             g_pD3D                                      = NULL;
LPDIRECT3DDEVICE8       g_pD3DDevice                                = NULL;
LPDIRECTINPUT8          g_pDInput                                   = NULL;
LPDIRECTINPUTDEVICE8    g_pKeyboard                                 = NULL;
LPDIRECTSOUND3DLISTENER g_pListener                                 = NULL;
LPD3DXSPRITE            g_pSprite                                   = NULL;
LPDIRECT3DTEXTURE8      g_pDiamondSprite                            = NULL;
LPDIRECT3DTEXTURE8      g_pCameraLockedSprite                       = NULL;
LPDIRECT3DTEXTURE8      g_pArrowSprite                              = NULL;
LPDIRECT3DTEXTURE8      g_pEnergyEmptySprite                        = NULL;
LPDIRECT3DTEXTURE8      g_pEnergyFullSprite                         = NULL;
LPDIRECT3DTEXTURE8      g_pShiningSprite                            = NULL;
LPDIRECT3DTEXTURE8      g_pSnowflakeSprite                          = NULL;
LPDIRECT3DTEXTURE8      g_pParticleSprite[NUM_PARTICLE_TYPES];
LPDIRECT3DTEXTURE8      g_pFireSprite[NUM_FIRE_SPRITES];
LPDIRECT3DTEXTURE8      g_pLastMeshTexture                          = NULL;
D3DMATERIAL8            *g_pmLastMeshMaterial                       = NULL;
D3DMATERIAL8            g_mWhiteMaterial;
D3DZBUFFERTYPE			g_ZBType;
JOYCAPS					g_JoyCaps;
BOOL                    g_bCameraLocked;
BOOL                    g_bObjectsOutOfMap;
BOOL                    g_bLevelFinished;
BOOL                    g_bLevelGroundRotating;
BOOL                    g_bLevelGroundMoving;
BOOL                    g_bSnowing;
BOOL                    g_bRenderShining;
BOOL                    g_bMenu;
BOOL                    g_bEnterPassword;
BOOL                    g_bConfirmQuit;
BOOL                    g_bConfirmRestart;
BOOL                    g_bPaused                                   = FALSE;
BOOL                    g_bCreditsMusicPlaying                      = FALSE;
BOOL					g_bUseJoystick;
char                    g_cKeyState[256];
char                    g_sLevelName[256];
char                    *g_sCopyright                               = "Blocks 2001 v1.1   Copyright (C) 2001, David Scherfgen";
char                    g_sPassword[256];
char                    *g_sLevelPassword[9]                        = {"", "FLOCKENGALOPP", "ANTIGAUSS", "BERHAUSEN", "HAARIGERHARRI", "SAUPFERD", "PHARAOXXIV", "HERMITDERSOCKE", "LASTLEVEL"};
int                     g_iLevel[50][20][50];
int                     g_iLevelWidth;
int                     g_iLevelDepth;
int                     g_iLevelHeight;
int                     g_iLevelDiamonds;
int                     g_iLevelNr;
int                     g_iNumSnowflakes;
int                     g_iFieldFlags[52][22][52];
int                     g_iSymbolTable[256];
int                     g_iViewMode;
int                     g_iMenuCursor;
int                     g_iMenuItemSelected;
int                     g_iPasswordCursor;
long                    g_lMusicVolume                              = -750;
DWORD                   g_dwLastRendering                           = 0;
float                   g_fCameraDistance;
float                   g_fCameraAngle;
float                   g_fCameraHeight;
float                   g_fLevelFinished;
float                   g_fLevelExitRotation;
float                   g_fGroundHeight;
float                   g_fLightFlickering;
float                   g_fResolutionX                              = 800.0f;
float                   g_fResolutionY                              = 600.0f;
float                   g_fCounter1;
float                   g_fCounter2;
float                   g_fEnergyDisplay;
float                   g_fCameraRadius;
float                   g_fLockCameraCountDown;
float                   g_fMoveCameraCountDown;
float                   g_fMenuCursorCountDown;
float                   g_fConfirmQuitCountDown;
float                   g_fDieCountDown;
CFieldType              g_FieldType[256];
CObjectType             g_ObjectType[256];
HWND                    g_hWindow                                   = NULL;
D3DCOLOR                g_cBackgroundColor                          = 0;
D3DCOLORVALUE           g_cLightColor;
HINSTANCE               g_hInstance                                 = NULL;
CPlayer                 g_Player;
D3DXVECTOR3             g_vCamera;
D3DXVECTOR3             g_vCameraDestination;
D3DXVECTOR3             g_vLookAt;
D3DXVECTOR3             g_vLevelExit;
D3DXVECTOR3             g_vCameraUp;
D3DXVECTOR3             g_vCameraDepth;
D3DXVECTOR3             g_vCameraDifference;
D3DXMATRIX              g_mCamera;
D3DXMATRIX              g_mProjection;
CObject                 g_Object[MAX_OBJECTS];
CControl                g_View1Camera;
CMusicManager           *g_pMusicManager                            = NULL;
CMusicSegment           *g_pMusic                                   = NULL;
CSoundManager           *g_pSoundManager                            = NULL;
CSound                  *g_pSound[NUM_SOUNDS][NUM_SOUND_CLONES];
CD3DFont                *g_pFontTahoma10                            = NULL;
CD3DFont                *g_pFontTahoma24                            = NULL;
CD3DFont                *g_pFontTahoma48                            = NULL;
CMesh                   *g_pLevelGround                             = NULL;
CMesh                   *g_pCubePart[6];
CMesh                   *g_pShadow                                  = NULL;
CSnowflake              g_Snowflake[MAX_SNOWFLAKES];
CParticle               g_Particle[MAX_PARTICLES];
CScript                 g_Script;

void Debug(char *sText)
{
    FILE *pFile;

    pFile = fopen("Debug.txt", "a");
    fprintf(pFile, sText);
    fclose(pFile);
}

HRESULT InitD3D()
{
    HRESULT r;
    D3DPRESENT_PARAMETERS d3dpp;
	D3DCAPS8 Caps;
	DWORD dwFlags;
	D3DFORMAT BBFormat[4] = {D3DFMT_X8R8G8B8, D3DFMT_R8G8B8, D3DFMT_R5G6B5, D3DFMT_X1R5G5B5};
	D3DFORMAT SFormat[4] = {D3DFMT_D32, D3DFMT_D24S8, D3DFMT_D16, D3DFMT_D16_LOCKABLE};

    g_pD3D = Direct3DCreate8(D3D_SDK_VERSION);
    if(g_pD3D == NULL) return E_FAIL;

	g_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &Caps);

	if(Caps.VertexProcessingCaps & D3DVTXPCAPS_DIRECTIONALLIGHTS)
	{
		dwFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
	}
	else dwFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    
    for(int a = 0; a < 4; a++)
	{
		for(int b = 0; b < 4; b++)
		{
			ZeroMemory(&d3dpp, sizeof(d3dpp));

			d3dpp.Windowed                  = FALSE;
			d3dpp.SwapEffect                = D3DSWAPEFFECT_DISCARD;
			d3dpp.BackBufferFormat          = BBFormat[a];
			d3dpp.EnableAutoDepthStencil    = TRUE;
			d3dpp.AutoDepthStencilFormat    = SFormat[b];
			d3dpp.BackBufferWidth           = (int)(g_fResolutionX);
			d3dpp.BackBufferHeight          = (int)(g_fResolutionY);

			r = g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_hWindow, dwFlags,
									 &d3dpp, &g_pD3DDevice);
			if(SUCCEEDED(r)) goto FormatsFound;
		}
	}

FormatsFound:
	if(FAILED(r)) return r;
	
	if(d3dpp.AutoDepthStencilFormat == D3DFMT_D16) g_ZBType = D3DZB_USEW;
	else g_ZBType = D3DZB_TRUE;

    g_pD3DDevice->SetRenderState(D3DRS_ZENABLE,				g_ZBType);
    g_pD3DDevice->SetRenderState(D3DRS_AMBIENT,             D3DCOLOR_XRGB(224, 224, 224));
    g_pD3DDevice->SetRenderState(D3DRS_DITHERENABLE,        TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND,            D3DBLEND_SRCALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND,           D3DBLEND_INVSRCALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_POINTSPRITEENABLE,   TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_POINTSCALEENABLE,    TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_POINTSIZE,           FloatToDWord(2.0f));
    g_pD3DDevice->SetRenderState(D3DRS_POINTSIZE_MIN,       FloatToDWord(0.0f));
    g_pD3DDevice->SetRenderState(D3DRS_POINTSCALE_A,        FloatToDWord(0.0f));
    g_pD3DDevice->SetRenderState(D3DRS_POINTSCALE_B,        FloatToDWord(0.0f));
    g_pD3DDevice->SetRenderState(D3DRS_POINTSCALE_C,        FloatToDWord(1.0f));

    g_pD3DDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    g_pD3DDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	g_pD3DDevice->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);

    return S_OK;
}

HRESULT InitFonts()
{
    g_pFontTahoma10 = new CD3DFont("Tahoma", 10);
    if(g_pFontTahoma10 == NULL) return E_FAIL;
    g_pFontTahoma10->InitDeviceObjects(g_pD3DDevice);
    g_pFontTahoma10->RestoreDeviceObjects();

    g_pFontTahoma24 = new CD3DFont("Tahoma", 24);
    if(g_pFontTahoma24 == NULL) return E_FAIL;
    g_pFontTahoma24->InitDeviceObjects(g_pD3DDevice);
    g_pFontTahoma24->RestoreDeviceObjects();

    g_pFontTahoma48 = new CD3DFont("Tahoma", 36);
    if(g_pFontTahoma48 == NULL) return E_FAIL;
    g_pFontTahoma48->InitDeviceObjects(g_pD3DDevice);
    g_pFontTahoma48->RestoreDeviceObjects();

    return S_OK;
}

void CleanUp()
{
    Debug("Cleaning up...\n");

    g_ObjectType[OT_FIRE].m_pMesh->m_pTexture[0] = NULL;

    Debug("Releasing field types...\n");
    for(int a = 0; a < NUM_FIELD_TYPES; a++) SAFE_RELEASE(g_FieldType[a].m_pMesh);
    for(a = 0; a < 6; a++) SAFE_RELEASE(g_pCubePart[a]);

    Debug("Releasing object types...\n");
    for(a = 0; a < NUM_OBJECT_TYPES; a++) SAFE_RELEASE(g_ObjectType[a].m_pMesh);

    Debug("Releasing player...\n");
    for(a = 0; a < NUM_PLAYER_WALK_STEPS; a++) SAFE_RELEASE(g_Player.m_pMeshWalk[a]);
    for(a = 0; a < NUM_PLAYER_PUSH_STEPS; a++) SAFE_RELEASE(g_Player.m_pMeshPush[a]);
    
    SAFE_RELEASE(g_Player.m_pMeshJump);
    SAFE_RELEASE(g_Player.m_pMeshFall);

    if(g_pLevelGround != NULL) g_pLevelGround->Release();

    Debug("Releasing keyboard device...\n");
    g_pKeyboard->Unacquire();

    Debug("Releasing sounds...\n");
    for(a = 0; a < NUM_SOUNDS; a++)
    {
        for(int b = 0; b < NUM_SOUND_CLONES; b++)
        {
            SAFE_DELETE(g_pSound[a][b]);
        }
    }
    
    Debug("Releasing fonts...\n");
    if(g_pFontTahoma10 != NULL)
    {
        g_pFontTahoma10->DeleteDeviceObjects();
        delete g_pFontTahoma10;
    }
    if(g_pFontTahoma24 != NULL)
    {
        g_pFontTahoma24->DeleteDeviceObjects();
        delete g_pFontTahoma24;
    }
    if(g_pFontTahoma48 != NULL)
    {
        g_pFontTahoma48->DeleteDeviceObjects();
        delete g_pFontTahoma48;
    }

    Debug("Releasing sprites...\n");
    SAFE_RELEASE(g_pSprite);
    SAFE_RELEASE(g_pDiamondSprite);
    SAFE_RELEASE(g_pCameraLockedSprite);
    SAFE_RELEASE(g_pArrowSprite);
    SAFE_RELEASE(g_pEnergyEmptySprite);
    SAFE_RELEASE(g_pEnergyFullSprite);
    SAFE_RELEASE(g_pShiningSprite);
    SAFE_RELEASE(g_pSnowflakeSprite);
    for(a = 0; a < NUM_PARTICLE_TYPES; a++) SAFE_RELEASE(g_pParticleSprite[a]);
    for(a = 0; a < NUM_FIRE_SPRITES; a++) SAFE_RELEASE(g_pFireSprite[a]);

    Debug("Releasing shadows...\n");
    SAFE_RELEASE(g_pShadow);

    Debug("Releasing DirectX devices...\n");
    SAFE_RELEASE(g_pD3DDevice);
    SAFE_RELEASE(g_pD3D);
    SAFE_RELEASE(g_pDInput);
    SAFE_RELEASE(g_pKeyboard);
    SAFE_DELETE (g_pSoundManager);
    SAFE_DELETE (g_pMusic);
    SAFE_DELETE (g_pMusicManager);
    
    Debug("Unregistering window class...\n");
    UnregisterClass("D3D Window", g_hInstance);

    Debug("Clean up sequence complete!\n");
}

HRESULT PlaySound(int iSound,
                  BOOL bLoop,
                  BOOL bReset,
                  D3DXVECTOR3 vPosition,
                  D3DXVECTOR3 vVelocity = D3DXVECTOR3(0.0f, 0.0f, 0.0f),
                  float fMinDistance = 5.0f,
                  float fMaxDistance = 100.0f)
{
    int                     iClone;
    int                     iNumSoundClones;
    LPDIRECTSOUND3DBUFFER8  p3DBuffer;
    HRESULT                 r;

    iNumSoundClones = NUM_SOUND_CLONES;
    if(iSound == S_PUSH_OBJECT) iNumSoundClones = 1;

    for(int a = 0; a < iNumSoundClones; a++)
    {
        if(!g_pSound[iSound][a]->IsSoundPlaying())
        {
            iClone = a;
            goto FreeCloneFound;
        }
    }

    if(bReset)
    {
        iClone = 0;
        g_pSound[iSound][0]->Stop();
    }
    else return S_OK;

FreeCloneFound:
 
    r = g_pSound[iSound][iClone]->Get3DBufferInterface(0, &p3DBuffer);
    if(FAILED(r)) return r;

    r = Set3DBuffer(p3DBuffer, vPosition, vVelocity, fMinDistance, fMaxDistance);
    if(FAILED(r)) return r;

    r = g_pSound[iSound][iClone]->Play(0, bLoop ? DSBPLAY_LOOPING : 0);
    if(FAILED(r)) return r;
    
    p3DBuffer->Release();

    return S_OK;
}

void PlayMusic(char *psFilename,
               BOOL bRepeat)
{
    if(g_pMusic != NULL)
    {
        g_pMusic->Stop(0);
        g_pMusic->Unload();
        delete g_pMusic;
    }

    g_pMusicManager->CollectGarbage();

    SAFE_DELETE(g_pMusicManager);
    g_pMusicManager = new CMusicManager();
    g_pMusicManager->Initialize(g_hWindow, 64, DMUS_APATH_SHARED_STEREOPLUSREVERB);

    g_pMusicManager->CreateSegmentFromFile(&g_pMusic, psFilename, TRUE, TRUE);

    if(bRepeat) g_pMusic->SetRepeats(DMUS_SEG_REPEAT_INFINITE);
    g_pMusic->Play(DMUS_SEGF_USE_AUDIOPATH | DMUS_SEGF_CONTROL);
}

CFieldType GetFieldType(int x,
                        int y,
                        int z)
{
    return g_FieldType[g_iLevel[x][y][z]];
}

int GetFieldFlags(int x,
                  int y,
                  int z)
{
    return g_iFieldFlags[x + 1][y + 1][z + 1];
}

int GetObject(D3DXVECTOR3 vPosition)
{
    D3DXVECTOR3 vObject;
    float       fHalfWidth;
    float       fHalfHeight;
    float       fHalfDepth;

    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(g_Object[a].m_bExists)
        {
            vObject = g_Object[a].m_Control.m_vPosition + g_Object[a].m_vAdd;
            fHalfWidth = g_Object[a].m_pType->m_fWidth / 2.0f;
            fHalfHeight = g_Object[a].m_pType->m_fHeight / 2.0f;
            fHalfDepth = g_Object[a].m_pType->m_fDepth / 2.0f;

            if(vPosition.x >= vObject.x - fHalfWidth && vPosition.x <= vObject.x + fHalfWidth &&
               vPosition.y >= vObject.y - fHalfHeight && vPosition.y <= vObject.y + fHalfHeight &&
               vPosition.z >= vObject.z - fHalfDepth && vPosition.z <= vObject.z + fHalfDepth)
            {
                return a;
            }
        }
    }

    return -1;
}

int GetRadiusObject(D3DXVECTOR3 vPosition,
                    float fRadius)
{
    float       fDistance;
    D3DXVECTOR3 vDifference;

    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(g_Object[a].m_bExists)
        {
            vDifference = vPosition - (g_Object[a].m_Control.m_vPosition + g_Object[a].m_vAdd);
            fDistance = D3DXVec3Length(&vDifference);

            if(fDistance <= g_Object[a].m_pType->m_fRadius + fRadius) return a;
        }
    }

    return -1;
}

BOOL IsFree(D3DXVECTOR3 vPosition,
            BOOL bOutOfMap,
            BOOL bCheckObjects,
            BOOL bExtras = FALSE,
            int *piObject = NULL,
            BOOL bCastShadow = FALSE)
{    
    int         x;
    int         y;
    int         z;
    int         iObject;
    float       fDistance;
    D3DXVECTOR3 vCenter;
    D3DXVECTOR3 vStart;
    CFieldType  FieldType;

    x = (int)(vPosition.x);
    y = (int)(vPosition.y);
    z = (int)(-vPosition.z);

    if(vPosition.x < 0.0f ||
       vPosition.y < 0.0f ||
       -vPosition.z < 0.0f ||
       x >= g_iLevelWidth ||
       y >= g_iLevelHeight ||
       z >= g_iLevelDepth)
    {
        if(piObject != NULL) *piObject = -1;
        return bOutOfMap;
    }

    FieldType = GetFieldType(x, y, z);

    if(FieldType.m_iFlags & FF_WALL)
    {
        vStart = D3DXVECTOR3((float)(x), (float)(y), (float)(-z));

        if(FieldType.m_iFlags & FF_RADIUS_XZ)
        {
            vCenter = vStart + FieldType.m_vCenter;
            vPosition.y = vCenter.y;
            fDistance = D3DXVec3Length(&D3DXVECTOR3(vPosition - vCenter));
            if(fDistance <= FieldType.m_fRadiusXZ)
            {
                if(piObject != NULL) *piObject = -1;
                return FALSE;
            }
        }
        else
        {
            if(vPosition.x >= vStart.x + FieldType.m_vMinimum.x && vPosition.x <= vStart.x + FieldType.m_vMaximum.x &&
               vPosition.y >= vStart.y + FieldType.m_vMinimum.y && vPosition.y <= vStart.y + FieldType.m_vMaximum.y &&
               vPosition.z >= vStart.z + FieldType.m_vMinimum.z && vPosition.z <= vStart.z + FieldType.m_vMaximum.z)
            {
                if(piObject != NULL) *piObject = -1;
                return FALSE;
            }
        }
    }

    if(bCheckObjects)
    {
        iObject = GetObject(vPosition);
        if(iObject != -1)
        {
            if(g_Object[iObject].m_pType->m_iFlags & OF_WALL)
            {
                if((!(g_Object[iObject].m_pType->m_iFlags & OF_EXTRA) ||
                   ((g_Object[iObject].m_pType->m_iFlags & OF_EXTRA) && !bExtras)) &&
                   g_Object[iObject].m_iType != OT_SNOWBALL)
                {
                    if(piObject != NULL) *piObject = iObject;
                    if(bCastShadow && (g_Object[iObject].m_pType->m_iFlags & OF_RECEIVE_SHADOWS)) return FALSE;
                    if(!bCastShadow) return FALSE;
                }
            }
        }
    }

    return TRUE;
}

BOOL IsFreeForPlayer(D3DXVECTOR3 vDirection,
                     BOOL bOutOfMap,
                     int iCheck,
                     BOOL bRelative = TRUE)
{
    float x1;
    float x2;
    float z1;
    float z2;
    float fMaxZ;

    if(g_Player.m_bPush) fMaxZ = 0.5f;
    else fMaxZ = 0.3f;

    if(iCheck == 0) {x1 = 0.0f; x2 = 0.0f; z1 = 0.0f; z2 = 0.0f;}
    else if(iCheck == 1) {x1 = -0.1f; x2 = 0.1f; z1 = -0.1f; z2 = 0.1f;}
    else if(iCheck == 2) {x1 = -0.4f; x2 = 0.4f; z1 = -0.3f; z2 = fMaxZ;}
    else if(iCheck == 3) {x1 = -0.4f; x2 = 0.4f; z1 = -0.3f; z2 = -0.2f;}
    else {x1 = -0.4f; x2 = 0.4f; z1 = 0.2f; z2 = fMaxZ;}

    for(float z = z1; z <= z2; z += 0.1f)
    {
        for(float x = x1; x <= x2; x += 0.1f)
        {
            if(bRelative)
            {
                if(!IsFree(g_Player.m_Control.GetRelativePosition(vDirection + D3DXVECTOR3(x, 0.0f, z)), bOutOfMap, TRUE, TRUE)) return FALSE;
            }
            else
            {
                if(!IsFree(g_Player.m_Control.m_vPosition + vDirection + D3DXVECTOR3(x, 0.0f, z), bOutOfMap, TRUE, TRUE)) return FALSE;
            }
        }
    }

    return TRUE;
}

BOOL ObjectFree(CObject Object,
                D3DXVECTOR3 vObject,
                BOOL bOutOfMap = FALSE,
                BOOL bPlayerCheck = TRUE)
{
    D3DXVECTOR3 vOtherObject;
    D3DXVECTOR3 vDifference;
    float       fHalfWidths;
    float       fHalfHeights;
    float       fHalfDepths;
    float       fHalfWidth;
    float       fHalfHeight;
    float       fHalfDepth;
    float       x;
    float       y;
    float       z;
    float       fDistance;

    fHalfWidth = Object.m_pType->m_fWidth / 2.0f;
    fHalfHeight = Object.m_pType->m_fHeight / 2.0f;
    fHalfDepth = Object.m_pType->m_fDepth / 2.0f;

    if(bPlayerCheck)
    {
        vDifference = vObject - g_Player.m_Control.m_vPosition;
        vDifference.y = 0.0f;
        fDistance = D3DXVec3Length(&vDifference);

        if(fDistance < 0.4f)
        {
            if(vObject.y - fHalfHeight <= g_Player.m_Control.m_vPosition.y + 1.0f &&
               vObject.y + fHalfHeight >= g_Player.m_Control.m_vPosition.y) return FALSE;
        }
    }

    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(g_Object[a].m_bExists)
        {
            if(g_Object[a].m_Control.m_vPosition != Object.m_Control.m_vPosition &&
               a != Object.m_iParent)
            {
                fHalfWidths = fHalfWidth + g_Object[a].m_pType->m_fWidth / 2.0f;
                fHalfHeights = fHalfHeight + g_Object[a].m_pType->m_fHeight / 2.0f;
                fHalfDepths = fHalfDepth + g_Object[a].m_pType->m_fDepth / 2.0f;

                vOtherObject = g_Object[a].m_Control.m_vPosition + g_Object[a].m_vAdd;

                if(vObject.x >= vOtherObject.x - (fHalfWidths - 0.05f) && vObject.x < vOtherObject.x + (fHalfWidths - 0.05f) &&
                   vObject.y >= vOtherObject.y - (fHalfHeights - 0.001f) && vObject.y < vOtherObject.y + (fHalfHeights - 0.001f) &&
                   vObject.z >= vOtherObject.z - (fHalfDepths - 0.05f) && vObject.z < vOtherObject.z + (fHalfDepths - 0.05f))
                {
                    return FALSE;
                }
            }
        }
    }

    for(float rz = -0.95f; rz <= 0.95f; rz += 0.95f)
    {
        z = rz * fHalfWidth;
        for(float ry = -1.0f; ry <= 0.95f; ry += 0.975f)
        {
            y = ry * fHalfHeight;
            for(float rx = -0.95f; rx <= 0.95f; rx += 0.95f)
            {
                x = rx * fHalfWidth;
                if(!IsFree(vObject + D3DXVECTOR3(x, y, z), bOutOfMap, FALSE)) return FALSE;
            }
        }
    }

    return TRUE;
}

BOOL RayIntersect(D3DXVECTOR3 vStart,
                  D3DXVECTOR3 vEnd,
                  D3DXVECTOR3 *pvIntersect = NULL,
                  BOOL bCamera = FALSE)
{
    int         iObject;
    int         x;
    int         y;
    int         z;
    float       fDistance;
    CFieldType  FieldType;
    D3DXVECTOR3 vRay;
    D3DXVECTOR3 vDirection;

    if(pvIntersect != NULL ||
       (bCamera && !g_bObjectsOutOfMap))
    {
        x = (int)(vStart.x);
        y = (int)(vStart.y);
        z = (int)(-vStart.z);

        if(vStart.x < 0.0f ||
           vStart.y < 0.0f ||
           -vStart.z < 0.0f ||
           x >= g_iLevelWidth ||
           y >= g_iLevelHeight ||
           z >= g_iLevelDepth)
        {
            if(pvIntersect != NULL) *pvIntersect = D3DXVECTOR3(vStart.x, g_fGroundHeight, vStart.z);
            return TRUE;
        }
    }

    vRay = vStart;
    vDirection = vEnd - vStart;
    D3DXVec3Normalize(&vDirection, &vDirection);
    vDirection /= 5.0f;

    while(TRUE)
    {
        if(pvIntersect != NULL)
        {
            if(vRay.y <= -1.0f)
            {
                *pvIntersect = D3DXVECTOR3(vStart.x, g_fGroundHeight, vStart.z);
                return TRUE;
            }
        }

        if(!IsFree(vRay, TRUE, TRUE, TRUE, &iObject, TRUE))
        {
            if(pvIntersect != NULL)
            {
                if(iObject == -1)
                {
                    FieldType = GetFieldType((int)(vRay.x), (int)(vRay.y), (int)(-vRay.z));
                    *pvIntersect = D3DXVECTOR3(vRay.x, (float)((int)(vRay.y)) + FieldType.m_vMaximum.y + 0.01f, vRay.z);
                }
                else
                {
                    *pvIntersect = D3DXVECTOR3(vRay.x, g_Object[iObject].m_Control.m_vPosition.y + g_Object[iObject].m_pType->m_fHeight / 2.0f + 0.01f, vRay.z);
                }
            }

            return TRUE;
        }

        fDistance = D3DXVec3Length(&D3DXVECTOR3(vRay - vEnd));
        if(fDistance < 0.5f) return FALSE;

        vRay += vDirection;
    }
}

int GetFreeObject()
{
    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(!g_Object[a].m_bExists) return a;
    }

    return -1;
}

int CreateObject(int iType,
                 D3DXVECTOR3 vPosition)
{
    int iObject;

    iObject = GetFreeObject();

    if(iObject != -1)
    {
        ZeroMemory(&g_Object[iObject], sizeof(g_Object[iObject]));

        g_Object[iObject].m_bExists = TRUE;
        g_Object[iObject].m_iType = iType;
        g_Object[iObject].m_iParent = -1;
        g_Object[iObject].m_fReload = g_ObjectType[iType].m_fReload;
        g_Object[iObject].m_fLife = 1.0f;
        g_Object[iObject].m_pType = &g_ObjectType[iType];
        g_Object[iObject].m_Control.Clear();
        g_Object[iObject].m_Control.m_vPosition = vPosition;
    }

    return iObject;
}

int GetFreeParticle()
{
    for(int a = 0; a < MAX_PARTICLES; a++)
    {
        if(g_Particle[a].m_fLife <= 0.0f) return a;
    }

    return -1;
}

int CreateParticle(int iType,
                   D3DXVECTOR3 vPosition,
                   float fLife,
                   float fSpeed,
                   D3DCOLOR cColor = D3DCOLOR_RGBA(255, 255, 255, 128),
                   D3DXVECTOR3 vDirection = D3DXVECTOR3(0.0f, 0.0f, 0.0f))
{
    int iParticle;

    iParticle = GetFreeParticle();

    if(iParticle != -1)
    {
        if(vDirection == D3DXVECTOR3(0.0f, 0.0f, 0.0f))
        {
            vDirection.x = FloatRandom(-1.0f, 1.0f);
            vDirection.y = FloatRandom(-1.0f, 1.0f);
            vDirection.z = FloatRandom(-1.0f, 1.0f);
            D3DXVec3Normalize(&vDirection, &vDirection);
            vDirection *= fSpeed;
        }

        if(vPosition.y < g_fGroundHeight &&
           vDirection.y > 0.0f) vPosition.y = g_fGroundHeight;

        g_Particle[iParticle].m_fLife = fLife;
        g_Particle[iParticle].m_iType = iType;
        g_Particle[iParticle].m_cColor = cColor;
        g_Particle[iParticle].m_vPosition = vPosition;
        g_Particle[iParticle].m_vDirection = vDirection;
    }

    return iParticle;
}

int ClassifyBox(D3DXVECTOR3 vMinimum,
				D3DXVECTOR3 vMaximum,
				D3DXPLANE *pPlane)
{
	D3DXVECTOR3 vNearPoint;
	D3DXVECTOR3 vFarPoint;

	vNearPoint = vMaximum;
	vFarPoint = vMinimum;

	if(pPlane->a > 0.0f) {vNearPoint.x = vMinimum.x; vFarPoint.x = vMaximum.x;}
	if(pPlane->b > 0.0f) {vNearPoint.y = vMinimum.y; vFarPoint.y = vMaximum.y;}
	if(pPlane->c > 0.0f) {vNearPoint.z = vMinimum.z; vFarPoint.z = vMaximum.z;}

	// Prfen, ob der nchste Punkt auerhalb liegt. Falls ja, liegt die ganze Box auerhalb.
	if(D3DXPlaneDotCoord(pPlane, &vNearPoint) > 0.0f) return BOX_OUTSIDE;

	// Prfen, ob der fernste Punkt auerhalb liegt. Falls ja, liegt die Box teilweise innerhalb.
	if(D3DXPlaneDotCoord(pPlane, &vFarPoint) > 0.0f) return BOX_CLIPPED;

	// Die Box befindet sich innerhalb
	return BOX_INSIDE;
}

void RenderLevel()
{
    int         iField;
    int         iFlags;
    int         iRenderMode;
    CFieldType  FieldType;
    D3DXMATRIX  mTranslation;
    D3DXMATRIX  mRotation;
    D3DXMATRIX  mTransformation;
	D3DXMATRIX	mCameraMatrix;
	D3DXMATRIX	mProjectionMatrix;
	D3DXMATRIX	mTemp;
	D3DXPLANE	ClipPlane[6];

    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    iRenderMode = 0;

    if(g_pLevelGround != NULL)
    {
        if(g_bLevelGroundRotating)
        {
            D3DXMatrixRotationY(&mRotation, g_fCounter1 / 100.0f);
            if(g_bLevelGroundMoving) D3DXMatrixTranslation(&mTranslation, (float)(g_iLevelWidth) / 2.0f + (float)(sin(g_fCounter1 / 20.0f)), g_fGroundHeight, (float)(g_iLevelDepth) / -2.0f - (float)(sin(g_fCounter1 / 30.0f)));
            else D3DXMatrixTranslation(&mTranslation, (float)(g_iLevelWidth) / 2.0f, g_fGroundHeight, (float)(g_iLevelDepth) / -2.0f);
            mTransformation = mRotation * mTranslation;
            g_pD3DDevice->SetTransform(D3DTS_WORLD, (D3DMATRIX*)(&mTransformation));
        }
        else
        {
            if(g_bLevelGroundMoving) SetWorldTranslation(D3DXVECTOR3((float)(g_iLevelWidth) / 2.0f + (float)(sin(g_fCounter1 / 20.0f)), g_fGroundHeight, (float)(g_iLevelDepth) / -2.0f - (float)(sin(g_fCounter1 / 30.0f))), g_pD3DDevice);
            else SetWorldTranslation(D3DXVECTOR3((float)(g_iLevelWidth) / 2.0f, g_fGroundHeight, (float)(g_iLevelDepth) / -2.0f), g_pD3DDevice);
        }

        g_pD3DDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT);
        g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
        g_pD3DDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);       
        g_pLevelGround->Render();
        g_pD3DDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
        g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
        g_pD3DDevice->SetRenderState(D3DRS_ZENABLE, g_ZBType);
    }

	// Sichtmatrix mit der Projektionsmatrix multiplizieren
	g_pD3DDevice->GetTransform(D3DTS_VIEW, (D3DMATRIX*)(&mCameraMatrix));
	g_pD3DDevice->GetTransform(D3DTS_PROJECTION, (D3DMATRIX*)(&mProjectionMatrix));
	mTemp = mCameraMatrix * mProjectionMatrix;

	// Linke Clippingebene
	ClipPlane[0].a = -(mTemp._14 + mTemp._11);
	ClipPlane[0].b = -(mTemp._24 + mTemp._21);
	ClipPlane[0].c = -(mTemp._34 + mTemp._31);
	ClipPlane[0].d = -(mTemp._44 + mTemp._41);

	// Rechte Clippingebene
	ClipPlane[1].a = -(mTemp._14 - mTemp._11);
	ClipPlane[1].b = -(mTemp._24 - mTemp._21);
	ClipPlane[1].c = -(mTemp._34 - mTemp._31);
	ClipPlane[1].d = -(mTemp._44 - mTemp._41);

	// Obere Clippingebene
	ClipPlane[2].a = -(mTemp._14 - mTemp._12);
	ClipPlane[2].b = -(mTemp._24 - mTemp._22);
	ClipPlane[2].c = -(mTemp._34 - mTemp._32);
	ClipPlane[2].d = -(mTemp._44 - mTemp._42);

	// Untere Clippingebene
	ClipPlane[3].a = -(mTemp._14 + mTemp._12);
	ClipPlane[3].b = -(mTemp._24 + mTemp._22);
	ClipPlane[3].c = -(mTemp._34 + mTemp._32);
	ClipPlane[3].d = -(mTemp._44 + mTemp._42);

	// Nahe Clippingebene
	ClipPlane[4].a = -(mTemp._14 + mTemp._13);
	ClipPlane[4].b = -(mTemp._24 + mTemp._23);
	ClipPlane[4].c = -(mTemp._34 + mTemp._33);
	ClipPlane[4].d = -(mTemp._44 + mTemp._43);

	// Ferne Clippingebene
	ClipPlane[5].a = -(mTemp._14 - mTemp._13);
	ClipPlane[5].b = -(mTemp._24 - mTemp._23);
	ClipPlane[5].c = -(mTemp._34 - mTemp._33);
	ClipPlane[5].d = -(mTemp._44 - mTemp._43);

    for(int y = g_iLevelHeight - 1; y >= 0; y--)
    {
        for(int z = 0; z < g_iLevelDepth; z++)
        {
            for(int x = 0; x < g_iLevelWidth; x++)
            {
                if(g_iLevel[x][y][z] != 255)
                {
					// Ist das Feld sichtbar?
					if(ClassifyBox(D3DXVECTOR3((float)(x) - 0.5f, (float)(y) - 0.5f, (float)(-z - 1) - 0.5f),
						           D3DXVECTOR3((float)(x + 1) + 0.5f, (float)(y + 1) + 0.5f, (float)(-z) + 0.5f),
								   ClipPlane) != BOX_OUTSIDE)
					{
	                    SetWorldTranslation(D3DXVECTOR3((float)(x), (float)(y), (float)(-z)), g_pD3DDevice);
		                iField = g_iLevel[x][y][z];
			            FieldType = g_FieldType[iField];
				        iFlags = FieldType.m_iFlags;

						if(iFlags & FF_TRANSPARENT)
						{
							if(iRenderMode != 1)
							{
								g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
								g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
								iRenderMode = 1;
							}
						}
						else
						{
							if(iRenderMode != 0)
							{
								g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
								g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
								iRenderMode = 0;
							}
						}

						if(iFlags & FF_CUBEMESH)
						{
							if(g_pmLastMeshMaterial != &FieldType.m_pMesh->m_Material[0])
							{
								g_pD3DDevice->SetMaterial(&FieldType.m_pMesh->m_Material[0]);
								g_pmLastMeshMaterial = &FieldType.m_pMesh->m_Material[0];
							}
							if(g_pLastMeshTexture != FieldType.m_pMesh->m_pTexture[0])
							{
								g_pD3DDevice->SetTexture(0, FieldType.m_pMesh->m_pTexture[0]);
								g_pLastMeshTexture = FieldType.m_pMesh->m_pTexture[0];
							}
                        
							if(!(GetFieldFlags(x - 1, y, z) & FF_CUBE)) g_pCubePart[CP_LEFT]->Render();
							if(!(GetFieldFlags(x + 1, y, z) & FF_CUBE)) g_pCubePart[CP_RIGHT]->Render();
							if(!(GetFieldFlags(x, y, z - 1) & FF_CUBE)) g_pCubePart[CP_FRONT]->Render();
							if(!(GetFieldFlags(x, y, z + 1) & FF_CUBE)) g_pCubePart[CP_BACK]->Render();
							if(!(GetFieldFlags(x, y - 1, z) & FF_CUBE)) g_pCubePart[CP_BOTTOM]->Render();
							if(!(GetFieldFlags(x, y + 1, z) & FF_CUBE)) g_pCubePart[CP_TOP]->Render();
						}
						else
						{
							FieldType.m_pMesh->Render();
						}
					}
                }
            }
        }
    }

    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
}

void RenderPlayer()
{
    CMesh *pMesh;

    g_Player.m_Control.UpdateTransformationMatrix();
    g_pD3DDevice->SetTransform(D3DTS_WORLD, (D3DMATRIX*)(&g_Player.m_Control.m_mTransformation));

    pMesh = g_Player.m_pMeshJump;
    
    if(!g_Player.m_bOnGround && g_Player.m_fFalling <= PLAYER_BEGIN_FALL) pMesh = g_Player.m_pMeshJump;
    else if(!g_Player.m_bOnGround && g_Player.m_fFalling > PLAYER_BEGIN_FALL) pMesh = g_Player.m_pMeshFall;
    else
    {
        if(g_Player.m_bPush) pMesh = g_Player.m_pMeshPush[(int)(g_Player.m_fStep)];
        else pMesh = g_Player.m_pMeshWalk[(int)(g_Player.m_fStep)];
    }

    if(g_Player.m_fEnergy <= 0.0f &&
       g_fDieCountDown <= DIE_DELAY * 0.75f) pMesh = g_Player.m_pMeshWalk[2];

    if(g_bLevelFinished)
    {
        for(int a = 0; a < (int)(pMesh->m_dwNumMaterials); a++)
        {
            if(g_fLevelFinished > 0.0f) pMesh->m_Material[a].Diffuse.a = g_fLevelFinished / LEVEL_FINISH_DELAY;
            else pMesh->m_Material[a].Diffuse.a = 0.0f;
        }

        g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
        g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
        g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
    }

    pMesh->Render();

    if(g_bLevelFinished) g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
}

void RenderObjects()
{
    float       fAngle;
    int         iFlags;
    D3DXMATRIX  mTransformation;
    D3DXMATRIX  mTranslation;
    D3DXMATRIX  mFaceToCamera;

    if(g_bRenderShining)
    {
        g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
        g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
        g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
        g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
        g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
        g_pD3DDevice->SetRenderState(D3DRS_POINTSIZE, FloatToDWord(2.0f));
    }

    for(int iRenderMode = (g_bRenderShining ? 0 : 1); iRenderMode < 4; iRenderMode++)
    {
        if(iRenderMode == 1)
        {
            g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
            g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
            g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
            g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
            g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
        }
        else if(iRenderMode == 2)
        {
            g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
            g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
            g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
            g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
            g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
        }
        else if(iRenderMode == 3)
        {
            g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
            g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
            g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
        }

        for(int a = 0; a < MAX_OBJECTS; a++)
        {
            if(g_Object[a].m_bExists)
            {
                iFlags = g_Object[a].m_pType->m_iFlags;

                if((iFlags & OF_BILLBOARD) && iRenderMode == 3)
                {
                    fAngle = (float)(atan2(g_Object[a].m_Control.m_vPosition.x - g_vCamera.x,
                                           g_Object[a].m_Control.m_vPosition.z - g_vCamera.z));
                    D3DXMatrixRotationY(&mFaceToCamera, fAngle);
                }
                else D3DXMatrixIdentity(&mFaceToCamera);

                g_Object[a].m_Control.UpdateTransformationMatrix();
                D3DXMatrixTranslation(&mTranslation, g_Object[a].m_vAdd.x, g_Object[a].m_vAdd.y, g_Object[a].m_vAdd.z);
                mTransformation = mFaceToCamera * g_Object[a].m_Control.m_mTransformation * mTranslation;
                g_pD3DDevice->SetTransform(D3DTS_WORLD, (D3DMATRIX*)(&mTransformation));

                if(((!(iFlags & OF_TRANSPARENT) && !(iFlags & OF_BILLBOARD) && iRenderMode == 1) ||
                   ((iFlags & OF_TRANSPARENT) && iRenderMode == 2) ||
                   ((iFlags & OF_BILLBOARD) && iRenderMode == 3)) &&
                   !(iFlags & OF_NO_MESH))
                {
                    g_Object[a].m_pType->m_pMesh->Render();
                }

                if((iFlags & OF_SHINING) && iRenderMode == 0)
                {
                    RenderPointSprite(g_pD3DDevice, g_pShiningSprite, g_Object[a].m_pType->m_vShiningPosition, g_Object[a].m_pType->m_cShiningColor);
                    g_pLastMeshTexture = g_pShiningSprite;
                }
            }
        }
    }

    g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
}

void RenderShadows()
{
    float       fScale;
    D3DXVECTOR3 vShadow;
    D3DXMATRIX  mTransformation;

    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
    g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_SRCALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_FOGENABLE, FALSE);
	g_pD3DDevice->SetRenderState(D3DRS_ZBIAS, 1);

    if(g_fLevelFinished > -LEVEL_FINISH_DELAY)
    {
        if(RayIntersect(g_Player.m_Control.m_vPosition, g_Player.m_Control.m_vPosition - D3DXVECTOR3(0.0f, 20.0f, 0.0f), &vShadow))
        {
            fScale = 0.75f / (g_Player.m_Control.m_vPosition.y - vShadow.y + 1.0f);
            D3DXMatrixTransformation(&mTransformation, NULL, NULL, &D3DXVECTOR3(fScale, 1.0f, fScale), NULL, NULL, &vShadow);
            g_pD3DDevice->SetTransform(D3DTS_WORLD, (D3DMATRIX*)(&mTransformation));
            g_pShadow->Render();
        }
    }

    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(g_Object[a].m_bExists)
        {
            if(RayIntersect(g_Object[a].m_Control.m_vPosition - D3DXVECTOR3(0.0f, g_Object[a].m_pType->m_fHeight, 0.0f), g_Object[a].m_Control.m_vPosition - D3DXVECTOR3(0.0f, 20.0f, 0.0f), &vShadow))
            {
                fScale = g_Object[a].m_pType->m_fShadowSize / ((g_Object[a].m_Control.m_vPosition.y + g_Object[a].m_vAdd.y) - vShadow.y + 1.0f);
                D3DXMatrixTransformation(&mTransformation, NULL, NULL, &D3DXVECTOR3(fScale, 1.0f, fScale), NULL, NULL, &vShadow);
                g_pD3DDevice->SetTransform(D3DTS_WORLD, (D3DMATRIX*)(&mTransformation));
                g_pShadow->Render();
            }
        }
    }

    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_DESTALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);
	g_pD3DDevice->SetRenderState(D3DRS_ZBIAS, 0);
}

void RenderSnow()
{
    for(int a = 0; a < g_iNumSnowflakes; a++)
    {
        RenderPointSprite(g_pD3DDevice, g_pSnowflakeSprite, g_Snowflake[a].m_vPosition, D3DCOLOR_RGBA(255, 255, 255, 128));
    }
}

void RenderParticles()
{
    for(int a = 0; a < MAX_PARTICLES; a++)
    {
        if(g_Particle[a].m_fLife > 0.0f)
        {
            RenderPointSprite(g_pD3DDevice, g_pParticleSprite[g_Particle[a].m_iType], g_Particle[a].m_vPosition, g_Particle[a].m_cColor);
        }
    }
}

void RenderPointSprites()
{
    g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_DESTALPHA);
    g_pD3DDevice->SetRenderState(D3DRS_POINTSIZE, FloatToDWord(0.1f));

    g_pD3DDevice->SetMaterial(&g_mWhiteMaterial);
    g_pmLastMeshMaterial = &g_mWhiteMaterial;

    SetWorldTranslation(D3DXVECTOR3(0.0f, 0.0f, 0.0f), g_pD3DDevice);

    if(g_bSnowing) RenderSnow();

    RenderParticles();

    g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    g_pD3DDevice->SetRenderState(D3DRS_POINTSIZE, FloatToDWord(2.0f));
}

void RenderDisplay()
{
    D3DXVECTOR3 vScreen;
    D3DXVECTOR2 vScaling;
    D3DXVECTOR2 vPosition;
    RECT        Rectangle;
    DWORD       dwTime;
    DWORD       dwDuration;
    float       fFPS;
    float       fRotation;
    float       fScroll;
    int         iAlpha;
    char        psText[64];

    dwTime = timeGetTime();
    dwDuration = dwTime - g_dwLastRendering;
    g_dwLastRendering = dwTime;

    g_pSprite->Begin();
    g_pD3DDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);

    if(g_iLevelNr == 8 && g_fLevelFinished < -LEVEL_FINISH_DELAY) goto DrawCredits;

    g_pSprite->Draw(g_pDiamondSprite, NULL, NULL, NULL, 0.0f, &D3DXVECTOR2(15.0f, g_fResolutionY - 104.0f), D3DCOLOR_RGBA(255, 255, 255, 160));
    if(g_bCameraLocked) g_pSprite->Draw(g_pCameraLockedSprite, NULL, NULL, NULL, 0.0f, &D3DXVECTOR2(g_fResolutionX - 75.0f, 10.0f), D3DCOLOR_RGBA(255, 255, 255, 160));

    if(g_Player.m_iNumDiamonds >= g_iLevelDiamonds)
    {
        D3DXVec3TransformCoord(&vScreen, &g_vLevelExit, &g_mCamera);
        if(vScreen.z > 0.0f)
        {
            D3DXVec3TransformCoord(&vScreen, &vScreen, &g_mProjection);
            vScreen.x = (g_fResolutionX / 2.0f) + (vScreen.x * g_fResolutionX / 2.0f);
            vScreen.y = (g_fResolutionY / 2.0f) - (vScreen.y * g_fResolutionY / 2.0f) + (float)(sin(g_fCounter1 / 10.0f)) * 16.0f;
            g_pSprite->Draw(g_pArrowSprite, NULL, NULL, NULL, 0.0f, &D3DXVECTOR2(vScreen.x - 16.0f, vScreen.y - 64.0f), D3DCOLOR_RGBA(255, 255, 255, 160));
        }
    }

    vScaling.x = vScaling.y = 1.0f + (float)(fabs(g_fEnergyDisplay - g_Player.m_fEnergy));
    fRotation = g_fEnergyDisplay - g_Player.m_fEnergy;
    vPosition = D3DXVECTOR2(g_fResolutionX - 300.0f, g_fResolutionY - 100.0f);

    g_pSprite->Draw(g_pEnergyEmptySprite, NULL, &vScaling, &D3DXVECTOR2(85.5f, 28.0f), fRotation, &vPosition, D3DCOLOR_RGBA(255, 255, 255, 160));

    Rectangle.left = 0;
    Rectangle.top = 0;
    Rectangle.right = (int)(vScaling.x * g_fEnergyDisplay * 255.0f);
    Rectangle.bottom = (int)(vScaling.y * 60.0f);
    g_pSprite->Draw(g_pEnergyFullSprite, &Rectangle, &vScaling, &D3DXVECTOR2(85.5f, 28.0f), fRotation, &vPosition, D3DCOLOR_RGBA(255, 255, 255, 128));
   
    if(g_cKeyState[DIK_F] ||
       g_cKeyState[DIK_V])
    {
        g_pFontTahoma10->DrawText(5.0f, 5.0f, D3DCOLOR_XRGB(255, 255, 255), "Blocks2001 v1.1");

        fFPS = 1000.0f / (float)(dwDuration);
        sprintf(psText, "FPS: %.2f", fFPS);
        g_pFontTahoma10->DrawText(5.0f, 20.0f, D3DCOLOR_XRGB(192, 192, 192), psText);
    }

    sprintf(psText, "%d / %d", g_Player.m_iNumDiamonds, g_iLevelDiamonds);
    g_pFontTahoma10->DrawText(30.0f, g_fResolutionY - 40.0f, D3DCOLOR_XRGB(255, 0, 0), psText);

    iAlpha = 0;
    if(g_fCounter1 < 128.0f) iAlpha = g_fCounter1 < 64.0f ? (int)(g_fCounter1 * 3.0f) : 192;
    else if(g_fCounter1 >= 128.0f && g_fCounter1 < 192.0f) iAlpha = (int)(192.0f - (g_fCounter1 - 128.0f) * 3.0f);
    
    if(iAlpha > 0)
    {
        g_pFontTahoma48->DrawText(200.0f, 250.0f, D3DCOLOR_RGBA(128, 128, 255, iAlpha), g_sLevelName, 0);
        g_pFontTahoma24->DrawText(200.0f, 300.0f, D3DCOLOR_RGBA(255, 255, 64, iAlpha), "Passwort:", 0);
        g_pFontTahoma24->DrawText(350.0f, 300.0f, D3DCOLOR_RGBA(128, 255, 128, iAlpha), g_sLevelPassword[g_iLevelNr], 0);
    }

    if(g_bConfirmQuit)
    {
        for(int a = 0; a < 5; a++)
        {
            g_pFontTahoma24->DrawText(180.0f, 250.0f + (float)(a), D3DCOLOR_RGBA(255, 255, 255, 64), "Zum Hauptmenue zurueckkehren?", 0);
            g_pFontTahoma48->DrawText(180.0f + (float)(a), 300.0f, D3DCOLOR_RGBA(128, 255, 128, 64), "Enter", 0);
            g_pFontTahoma48->DrawText(400.0f - (float)(a), 300.0f, D3DCOLOR_RGBA(255, 128, 128, 64), "Escape", 0);
        }
    }
    else if(g_bConfirmRestart)
    {
        for(int a = 0; a < 5; a++)
        {
            g_pFontTahoma24->DrawText(180.0f, 250.0f + (float)(a), D3DCOLOR_RGBA(255, 255, 255, 64), "Level neu starten?", 0);
            g_pFontTahoma48->DrawText(180.0f + (float)(a), 300.0f, D3DCOLOR_RGBA(128, 255, 128, 64), "Enter", 0);
            g_pFontTahoma48->DrawText(400.0f - (float)(a), 300.0f, D3DCOLOR_RGBA(255, 128, 128, 64), "Escape", 0);
        }
    }

    if(g_bPaused)
    {
        for(int a = 0; a < 5; a++)
        {
            g_pFontTahoma48->DrawText(280.0f - (float)(a), 250.0f, D3DCOLOR_RGBA(255, 128, 128, 64), "P A U S E", 0);
        }
        g_pFontTahoma24->DrawText(200.0f, 300.0f, D3DCOLOR_RGBA(255, 255, 64, 192), "Passwort:", 0);
        g_pFontTahoma24->DrawText(350.0f, 300.0f, D3DCOLOR_RGBA(128, 255, 128, 192), g_sLevelPassword[g_iLevelNr], 0);
    }

DrawCredits:

    if(g_iLevelNr == 8 &&
       g_fLevelFinished < -400.0f)
    {
        fScroll = -(g_fLevelFinished + 400.0f) * 2.0f;
        if(fScroll > 2270.0f) fScroll = 2270.0f;

        g_pFontTahoma48->DrawText(100.0f, 600.0f - fScroll, D3DCOLOR_RGBA(0, 255, 128, 255), "Vielen Dank,", 0);
        g_pFontTahoma24->DrawText(100.0f, 650.0f - fScroll, D3DCOLOR_RGBA(128, 255, 0, 255), "dass Sie Blocks 2001 gespielt haben!", 0);

        g_pFontTahoma48->DrawText(100.0f, 900.0f - fScroll, D3DCOLOR_RGBA(255, 128, 0, 255), "Idee", 0);
        g_pFontTahoma24->DrawText(200.0f, 960.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "David Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 1000.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Rolf Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 1040.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Eva Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 1080.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Gertrud Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 1120.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Jens Freche", 0);
        g_pFontTahoma24->DrawText(200.0f, 1160.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Hendrik Duesing", 0);
        g_pFontTahoma24->DrawText(200.0f, 1200.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Alexander Gehrke", 0);
        g_pFontTahoma24->DrawText(200.0f, 1240.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Maurice Kroh", 0);
        g_pFontTahoma24->DrawText(200.0f, 1280.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Albert Kalchmair", 0);

        g_pFontTahoma48->DrawText(100.0f, 1480.0f - fScroll, D3DCOLOR_RGBA(255, 128, 0, 255), "Programmierung", 0);
        g_pFontTahoma24->DrawText(200.0f, 1540.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "David Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 1580.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Microsoft", 0);

        g_pFontTahoma48->DrawText(100.0f, 1780.0f - fScroll, D3DCOLOR_RGBA(255, 128, 0, 255), "Grafik", 0);
        g_pFontTahoma24->DrawText(200.0f, 1840.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "David Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 1880.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Diverse Grafiker", 0);

        g_pFontTahoma48->DrawText(100.0f, 2080.0f - fScroll, D3DCOLOR_RGBA(255, 128, 0, 255), "Sound und Musik", 0);
        g_pFontTahoma24->DrawText(200.0f, 2140.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "David Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 2180.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Rolf Scherfgen", 0);
        g_pFontTahoma24->DrawText(200.0f, 2220.0f - fScroll, D3DCOLOR_RGBA(0, 128, 255, 255), "Diverse Musiker", 0);

        g_pFontTahoma24->DrawText(150.0f, 2520.0f - fScroll, D3DCOLOR_RGBA(255, 32, 0, 255), "Besuchen Sie mich im Internet:", 0);
        g_pFontTahoma24->DrawText(150.0f, 2560.0f - fScroll, D3DCOLOR_RGBA(255, 128, 0, 255), "http://www.scherfgensoftware.de", 0);

        g_pFontTahoma10->DrawText(250.0f, 2620.0f - fScroll, D3DCOLOR_RGBA(255, 255, 255, 255), "Druecken Sie ALT+F4, um das Spiel zu beenden.", 0);
    }

    g_pD3DDevice->SetRenderState(D3DRS_ZENABLE, g_ZBType);
    g_pSprite->End();
}

void RenderMenu()
{
    char    *sMenuEntry[3]          = {"Neues Spiel beginnen", "Passwort eingeben", "Spiel beenden"};
    char    sTemp[256];
    int     iMenuEntryColor[3 * 3]  = {255, 255, 255,
                                       255, 255, 255,
                                       255, 255, 255};
    int     iAlpha;
    DWORD   dwColor;
    float   fFactor;

    SetListener(g_pListener,
                g_vCamera,
                D3DXVECTOR3(0.0f, 0.0f, 0.0f),
                D3DXVECTOR3(0.0f, 0.0f, 1.0f),
                D3DXVECTOR3(0.0f, -1.0f, 0.0f),
                1.0f, 1.0f, 10.0f);

    g_pFontTahoma10->DrawText(340.0f, 20.0f, D3DCOLOR_RGBA(192, 192, 192, 192), "ScherfgenSoftware");

    if(g_fCounter1 >= 250.0f)
    {
        for(int a = 0; a < 3; a++)
        {
            iAlpha = ((int)(g_fCounter1) - 250) * 4 - a * 64;
            if(iAlpha < 0) iAlpha = 0;
            if(iAlpha > 224) iAlpha = 224;

            fFactor = g_iMenuCursor == a ? 1.0f : 0.5f;

            dwColor = D3DCOLOR_RGBA((int)((float)(iMenuEntryColor[a * 3]) * fFactor),
                                    (int)((float)(iMenuEntryColor[a * 3 + 1]) * fFactor),
                                    (int)((float)(iMenuEntryColor[a * 3 + 2]) * fFactor),
                                    iAlpha);
            g_pFontTahoma10->DrawText(320.0f, 490.0f + a * 20.0f, dwColor, sMenuEntry[a]);

            if(g_bEnterPassword)
            {
                strcpy(sTemp, g_sPassword);
                strcat(sTemp, "_");
                g_pFontTahoma10->DrawText(470.0f, 510.0f, D3DCOLOR_RGBA(0, 255, 0, 224), sTemp);
            }
        }

        iAlpha = ((int)(g_fCounter1) - 250) * 4 - 4 * 64;
        if(iAlpha < 0) iAlpha = 0;
        if(iAlpha > 224) iAlpha = 224;

        g_pFontTahoma10->DrawText(9.0f, 580.0f, D3DCOLOR_RGBA(64, 192, 64, iAlpha), g_sCopyright, 0);
    }
}

void Render()
{
    float       fFlickering;
    D3DLIGHT8   lLight;

    ZeroMemory(&lLight, sizeof(lLight));

    fFlickering = FloatRandom(-g_fLightFlickering, g_fLightFlickering);
    lLight.Type         = D3DLIGHT_DIRECTIONAL;
    lLight.Diffuse.r    = g_cLightColor.r + fFlickering;
    lLight.Diffuse.g    = g_cLightColor.g + fFlickering;
    lLight.Diffuse.b    = g_cLightColor.b + fFlickering;
    lLight.Direction    = g_vLookAt - g_vCamera;

    D3DXVec3Normalize((D3DXVECTOR3*)(&lLight.Direction), (D3DXVECTOR3*)(&lLight.Direction));

    g_pD3DDevice->SetLight(0, &lLight);
    g_pD3DDevice->LightEnable(0, TRUE);

    g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                        g_cBackgroundColor, 1.0f, 0);
    
    g_pD3DDevice->BeginScene();

    g_pLastMeshTexture = NULL;
    g_pmLastMeshMaterial = NULL;
  
    RenderLevel();

    if(g_bMenu) RenderMenu();
    else
    {
        if(!g_bLevelFinished ||
           (g_bLevelFinished && g_fLevelFinished > 0.0f))
        {
            if(g_iViewMode == 0) RenderPlayer();
            else if(g_iViewMode == 1)
            {
                if(D3DXVec3Length(&D3DXVECTOR3(g_View1Camera.m_vPosition - g_vCamera)) > 0.5f)
                {
                    RenderPlayer();
                }
            }
        }
    }

    RenderObjects();

    if(!g_bMenu) RenderShadows();

    RenderPointSprites();

    if(!g_bMenu) RenderDisplay();

    g_pD3DDevice->EndScene();
    g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
}

HRESULT UpdateKeyStates()
{
    ZeroMemory(&g_cKeyState, sizeof(g_cKeyState));

    return g_pKeyboard->GetDeviceState(sizeof(g_cKeyState), &g_cKeyState);
}

LRESULT WINAPI MessageProc(HWND hWindow,
                           UINT uMessage,
                           WPARAM wParam,
                           LPARAM lParam)
{
    switch(uMessage)
    {
    case WM_ACTIVATE:
        if(g_pKeyboard != NULL) g_pKeyboard->Acquire();
        break;

    case WM_TIMER:
        UpdateKeyStates();
        break;

    case WM_CLOSE:
        KillTimer(hWindow, 0);
        break;

    case WM_DESTROY:
        Debug("Exiting...\n");
        CleanUp();
        PostQuitMessage(0);
        break;
    }

    return DefWindowProc(hWindow, uMessage, wParam, lParam);
}

void InitLevelFieldTypes()
{
    char    sFilename[256];
    int     iField;

    for(int a = 0; a < 3; a++)
    {
        if(a == 0) iField = 0;
        if(a == 1) iField = 1;
        if(a == 2) iField = 7;

        SAFE_RELEASE(g_FieldType[iField].m_pMesh);
        sprintf(sFilename, "Data\\Field%d_%d.x", iField, g_iLevelNr);
        MeshCreate(&g_FieldType[iField].m_pMesh, sFilename, g_pD3DDevice, TRUE, TRUE, TRUE);

        if(g_FieldType[iField].m_pMesh != NULL)
        {
            for(int b = 0; b < MESH_MAX_MATERIALS; b++)
            {
                SAFE_RELEASE(g_FieldType[iField].m_pMesh->m_pTexture[b]);
            }
            sprintf(sFilename, "Data\\Field%d_%d.bmp", iField, g_iLevelNr);
            D3DXCreateTextureFromFile(g_pD3DDevice, sFilename, &g_FieldType[iField].m_pMesh->m_pTexture[0]);
        }
    }

    if(g_iLevelNr == 7)
    {
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_0.bmp", &g_FieldType[1].m_pMesh->m_pTexture[1]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_1.bmp", &g_FieldType[1].m_pMesh->m_pTexture[2]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_2.bmp", &g_FieldType[1].m_pMesh->m_pTexture[3]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_3.bmp", &g_FieldType[1].m_pMesh->m_pTexture[4]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_4.bmp", &g_FieldType[1].m_pMesh->m_pTexture[5]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_5.bmp", &g_FieldType[1].m_pMesh->m_pTexture[6]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_6.bmp", &g_FieldType[1].m_pMesh->m_pTexture[7]);
        D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field1_7_7.bmp", &g_FieldType[1].m_pMesh->m_pTexture[8]);
    }
}

HRESULT InitFieldTypes()
{
    char    sFilename[256];
    HRESULT r;

    ZeroMemory(&g_FieldType, sizeof(g_FieldType));

    g_FieldType[0].m_iFlags     = FF_WALL;
    g_FieldType[0].m_vMinimum   = D3DXVECTOR3(0.0f, 0.0f, -1.0f);
    g_FieldType[0].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);    

    g_FieldType[1].m_iFlags     = FF_WALL | FF_CUBE | FF_CUBEMESH;
    g_FieldType[1].m_vMinimum   = D3DXVECTOR3(0.0f, 0.0f, -1.0f);
    g_FieldType[1].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);

    g_FieldType[2].m_iFlags     = FF_WALL | FF_RADIUS_XZ;
    g_FieldType[2].m_vCenter    = D3DXVECTOR3(0.5f, 0.5f, -0.5f);
    g_FieldType[2].m_fRadiusXZ  = 0.25f;
    MeshCreate(&g_FieldType[2].m_pMesh, "Data\\Field2.x", g_pD3DDevice);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field2.bmp", &g_FieldType[2].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;

    g_FieldType[3].m_iFlags     = FF_WALL;
    g_FieldType[3].m_vMinimum   = D3DXVECTOR3(0.0f, 0.8f, -1.0f);
    g_FieldType[3].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
    MeshCreate(&g_FieldType[3].m_pMesh, "Data\\Field3.x", g_pD3DDevice, TRUE, FALSE, TRUE);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field3.bmp", &g_FieldType[3].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;

    g_FieldType[4].m_iFlags     = FF_WALL;
    g_FieldType[4].m_vMinimum   = D3DXVECTOR3(0.0f, 0.8f, -1.0f);
    g_FieldType[4].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
    MeshCreate(&g_FieldType[4].m_pMesh, "Data\\Field4.x", g_pD3DDevice, TRUE, FALSE, TRUE);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field4.bmp", &g_FieldType[4].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;

    g_FieldType[5].m_iFlags     = FF_WALL | FF_RADIUS_XZ;
    g_FieldType[5].m_vCenter    = D3DXVECTOR3(0.5f, 0.5f, -0.5f);
    g_FieldType[5].m_fRadiusXZ  = 0.35f;
    MeshCreate(&g_FieldType[5].m_pMesh, "Data\\Field5.x", g_pD3DDevice);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field5.bmp", &g_FieldType[5].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;

    g_FieldType[6].m_iFlags     = FF_WALL | FF_CUBE | FF_CUBEMESH;
    g_FieldType[6].m_vMinimum   = D3DXVECTOR3(0.0f, 0.0f, -1.0f);
    g_FieldType[6].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
    MeshCreate(&g_FieldType[6].m_pMesh, "Data\\Field6.x", g_pD3DDevice);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field6.bmp", &g_FieldType[6].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;
    D3DXComputeNormals(g_FieldType[6].m_pMesh->m_pMesh, NULL);

    g_FieldType[7].m_iFlags     = FF_WALL;
    g_FieldType[7].m_vMinimum   = D3DXVECTOR3(0.0f, 0.5f, -1.0f);
    g_FieldType[7].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);

    g_FieldType[8].m_iFlags     = FF_WALL;
    g_FieldType[8].m_vMinimum   = D3DXVECTOR3(0.25f, 0.0f, -0.75f);
    g_FieldType[8].m_vMaximum   = D3DXVECTOR3(0.75f, 1.0f, -0.25f);
    MeshCreate(&g_FieldType[8].m_pMesh, "Data\\Field8.x", g_pD3DDevice, TRUE, TRUE, FALSE);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field8_0.bmp", &g_FieldType[8].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field8_1.bmp", &g_FieldType[8].m_pMesh->m_pTexture[1]);
    if(FAILED(r)) return r;

    g_FieldType[9].m_iFlags     = FF_WALL;
    g_FieldType[9].m_vMinimum   = D3DXVECTOR3(0.3f, 0.0f, -0.7f);
    g_FieldType[9].m_vMaximum   = D3DXVECTOR3(0.7f, 0.3f, -0.3f);
    MeshCreate(&g_FieldType[9].m_pMesh, "Data\\Field9.x", g_pD3DDevice, TRUE, FALSE, TRUE);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field9.bmp", &g_FieldType[9].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;

    g_FieldType[10].m_iFlags     = FF_WALL;
    g_FieldType[10].m_vMinimum   = D3DXVECTOR3(0.3f, 0.0f, -0.7f);
    g_FieldType[10].m_vMaximum   = D3DXVECTOR3(0.7f, 0.1f, -0.3f);
    MeshCreate(&g_FieldType[10].m_pMesh, "Data\\Field10.x", g_pD3DDevice, TRUE, FALSE, TRUE);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field10.bmp", &g_FieldType[10].m_pMesh->m_pTexture[0]);
    if(FAILED(r)) return r;

    g_FieldType[11].m_iFlags     = FF_WALL | FF_SPRING;
    g_FieldType[11].m_vMinimum   = D3DXVECTOR3(0.0f, 0.0f, -1.0f);
    g_FieldType[11].m_vMaximum   = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
    MeshCreate(&g_FieldType[11].m_pMesh, "Data\\Field11.x", g_pD3DDevice);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field11_0.bmp", &g_FieldType[11].m_pMesh->m_pTexture[0]);
    r = D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Field11_1.bmp", &g_FieldType[11].m_pMesh->m_pTexture[1]);
    if(FAILED(r)) return r;

    g_FieldType[255].m_iFlags = 0;
    g_FieldType[255].m_iType = 255;
    g_FieldType[255].m_pMesh = NULL;

    for(int a = 0; a < NUM_FIELD_TYPES; a++)
    {
        g_FieldType[a].m_iType = a;

        if(a != 0 &&
           a != 1 &&
           a != 7 &&
           a != 255 &&
           g_FieldType[a].m_pMesh == NULL)
        {
            return E_FAIL;
        }
    }

    for(a = 0; a < 6; a++)
    {
        sprintf(sFilename, "Data\\CubePart%d.x", a);
        MeshCreate(&g_pCubePart[a], sFilename, g_pD3DDevice, FALSE);
        if(g_pCubePart[a] == NULL) return E_FAIL;
    }

    g_iSymbolTable[' '] = 255;
    g_iSymbolTable['.'] = 0;
    g_iSymbolTable['m'] = 1;
    g_iSymbolTable['i'] = 2;
    g_iSymbolTable['|'] = 3;
    g_iSymbolTable['-'] = 4;
    g_iSymbolTable['A'] = 5;
    g_iSymbolTable['M'] = 6;
    g_iSymbolTable['p'] = 7;
    g_iSymbolTable['f'] = 8;
    g_iSymbolTable['s'] = 9;
    g_iSymbolTable['S'] = 10;
    g_iSymbolTable['t'] = 11;

    g_iSymbolTable['0'] = -10;
    g_iSymbolTable['1'] = -11;
    g_iSymbolTable['2'] = -12;
    g_iSymbolTable['3'] = -13;
    g_iSymbolTable['4'] = -14;
    g_iSymbolTable['5'] = -15;
    g_iSymbolTable['6'] = -16;

    g_iSymbolTable['!'] = -100;
    g_iSymbolTable['?'] = -101;

    return S_OK;
}

HRESULT InitObjectTypes()
{
    ZeroMemory(&g_ObjectType, sizeof(g_ObjectType));

    g_ObjectType[OT_BLOCK1].m_iFlags = OF_WALL | OF_PUSHABLE | OF_DESTROYABLE | OF_RECEIVE_SHADOWS;
    g_ObjectType[OT_BLOCK1].m_fWidth = 1.0f;
    g_ObjectType[OT_BLOCK1].m_fHeight = 1.0f;
    g_ObjectType[OT_BLOCK1].m_fDepth = 1.0f;
    g_ObjectType[OT_BLOCK1].m_fShadowSize = 1.5f;
    MeshCreate(&g_ObjectType[OT_BLOCK1].m_pMesh, "Data\\Object0.x", g_pD3DDevice);
    if(FAILED(D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Object0.bmp", &g_ObjectType[OT_BLOCK1].m_pMesh->m_pTexture[0]))) return E_FAIL;

    g_ObjectType[OT_DIAMOND].m_iFlags = OF_WALL | OF_FLYING | OF_TRANSPARENT | OF_EXTRA | OF_SHINING;
    g_ObjectType[OT_DIAMOND].m_fWidth = 0.5f;
    g_ObjectType[OT_DIAMOND].m_fHeight = 1.0f;
    g_ObjectType[OT_DIAMOND].m_fDepth = 0.5f;
    g_ObjectType[OT_DIAMOND].m_fShadowSize = 0.5f;
    g_ObjectType[OT_DIAMOND].m_cShiningColor = D3DCOLOR_XRGB(255, 0, 0);
    MeshCreate(&g_ObjectType[OT_DIAMOND].m_pMesh, "Data\\Object1.x", g_pD3DDevice);
    if(FAILED(D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Object1.bmp", &g_ObjectType[OT_DIAMOND].m_pMesh->m_pTexture[0]))) return E_FAIL;
    g_ObjectType[OT_DIAMOND].m_pMesh->m_Material[0].Ambient.a = 0.75f;
    g_ObjectType[OT_DIAMOND].m_pMesh->m_Material[0].Diffuse.a = 0.75f;
    g_ObjectType[OT_DIAMOND].m_pMesh->m_Material[1].Ambient.a = 0.75f;
    g_ObjectType[OT_DIAMOND].m_pMesh->m_Material[1].Diffuse.a = 0.75f;

    g_ObjectType[OT_LEVEL_EXIT].m_iFlags = OF_WALL | OF_FLYING | OF_EXTRA;
    g_ObjectType[OT_LEVEL_EXIT].m_fWidth = 1.0f;
    g_ObjectType[OT_LEVEL_EXIT].m_fHeight = 1.1f;
    g_ObjectType[OT_LEVEL_EXIT].m_fDepth = 0.4f;
    g_ObjectType[OT_LEVEL_EXIT].m_fShadowSize = 1.0f;
    MeshCreate(&g_ObjectType[OT_LEVEL_EXIT].m_pMesh, "Data\\Object2.x", g_pD3DDevice);

    g_ObjectType[OT_PLATFORM1].m_iFlags = OF_WALL | OF_FLYING | OF_SWINGING_X | OF_RECEIVE_SHADOWS;
    g_ObjectType[OT_PLATFORM1].m_fWidth = 1.0f;
    g_ObjectType[OT_PLATFORM1].m_fHeight = 0.25f;
    g_ObjectType[OT_PLATFORM1].m_fDepth = 1.0f;
    g_ObjectType[OT_PLATFORM1].m_fShadowSize = 1.5f;
    MeshCreate(&g_ObjectType[OT_PLATFORM1].m_pMesh, "Data\\Object3.x", g_pD3DDevice);
    if(FAILED(D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Object3.bmp", &g_ObjectType[OT_PLATFORM1].m_pMesh->m_pTexture[0]))) return E_FAIL;

    g_ObjectType[OT_PLATFORM2].m_iFlags = OF_WALL | OF_FLYING | OF_SWINGING_Z | OF_RECEIVE_SHADOWS;
    g_ObjectType[OT_PLATFORM2].m_fWidth = 1.0f;
    g_ObjectType[OT_PLATFORM2].m_fHeight = 0.25f;
    g_ObjectType[OT_PLATFORM2].m_fDepth = 1.0f;
    g_ObjectType[OT_PLATFORM2].m_fShadowSize = 1.5f;
    MeshCreate(&g_ObjectType[OT_PLATFORM2].m_pMesh, "Data\\Object4.x", g_pD3DDevice);
    if(FAILED(D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Object4.bmp", &g_ObjectType[OT_PLATFORM2].m_pMesh->m_pTexture[0]))) return E_FAIL;

    g_ObjectType[OT_ICE1].m_iFlags = OF_WALL | OF_TRANSPARENT | OF_SPIKE;
    g_ObjectType[OT_ICE1].m_fWidth = 0.5f;
    g_ObjectType[OT_ICE1].m_fHeight = 2.0f;
    g_ObjectType[OT_ICE1].m_fDepth = 0.5f;
    MeshCreate(&g_ObjectType[OT_ICE1].m_pMesh, "Data\\Object5.x", g_pD3DDevice);
    g_ObjectType[OT_ICE1].m_pMesh->m_Material[0].Ambient.a = 0.5f;
    g_ObjectType[OT_ICE1].m_pMesh->m_Material[0].Diffuse.a = 0.5f;

    g_ObjectType[OT_CANNON1].m_iFlags = OF_WALL | OF_FLYING;
    g_ObjectType[OT_CANNON1].m_fReload = 50.0f;
    g_ObjectType[OT_CANNON1].m_fWidth = 1.0f;
    g_ObjectType[OT_CANNON1].m_fHeight = 0.5f;
    g_ObjectType[OT_CANNON1].m_fDepth = 1.0f;
    g_ObjectType[OT_CANNON1].m_vAdd = D3DXVECTOR3(0.0f, -0.25f, 0.0f);
    MeshCreate(&g_ObjectType[OT_CANNON1].m_pMesh, "Data\\Object6.x", g_pD3DDevice);
    if(FAILED(D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Object6.bmp", &g_ObjectType[OT_CANNON1].m_pMesh->m_pTexture[1]))) return E_FAIL;

    g_ObjectType[OT_SNOWBALL].m_iFlags = OF_WALL;
    g_ObjectType[OT_SNOWBALL].m_fWidth = 0.2f;
    g_ObjectType[OT_SNOWBALL].m_fHeight = 0.2f;
    g_ObjectType[OT_SNOWBALL].m_fDepth = 0.2f;
    g_ObjectType[OT_SNOWBALL].m_fShadowSize = 0.2f;
    MeshCreate(&g_ObjectType[OT_SNOWBALL].m_pMesh, "Data\\Object7.x", g_pD3DDevice);
    if(FAILED(D3DXCreateTextureFromFile(g_pD3DDevice, "Data\\Object7.bmp", &g_ObjectType[OT_SNOWBALL].m_pMesh->m_pTexture[0]))) return E_FAIL;

    g_ObjectType[OT_FIRE].m_iFlags = OF_FLYING | OF_SHINING | OF_BILLBOARD | OF_FIRE;
    g_ObjectType[OT_FIRE].m_fWidth = 0.25f;
    g_ObjectType[OT_FIRE].m_fHeight = 0.5f;
    g_ObjectType[OT_FIRE].m_fDepth = 0.25f;
    g_ObjectType[OT_FIRE].m_fShadowSize = 0.0f;
    g_ObjectType[OT_FIRE].m_vAdd = D3DXVECTOR3(0.0f, -0.05f, 0.0f);
    g_ObjectType[OT_FIRE].m_cShiningColor = D3DCOLOR_XRGB(255, 255, 0);
    g_ObjectType[OT_FIRE].m_vShiningPosition = D3DXVECTOR3(0.0f, -0.1f, 0.0f);
    MeshCreate(&g_ObjectType[OT_FIRE].m_pMesh, "Data\\Object8.x", g_pD3DDevice, TRUE, TRUE, TRUE);

    for(int a = 0; a < NUM_OBJECT_TYPES; a++)
    {
        g_ObjectType[a].m_fRadius = (g_ObjectType[a].m_fWidth + g_ObjectType[a].m_fHeight + g_ObjectType[a].m_fDepth) / 3.0f;

        if(!(g_ObjectType[a].m_iFlags & OF_NO_MESH) &&
           g_ObjectType[a].m_pMesh == NULL)
        {
            return E_FAIL;
        }
    }

    return S_OK;
}

HRESULT InitDInput()
{
    HRESULT r;

    r = DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION,
                           IID_IDirectInput8, (VOID**)&g_pDInput, NULL);
    if(FAILED(r)) return r;
    
    r = g_pDInput->CreateDevice(GUID_SysKeyboard, &g_pKeyboard, NULL);
    if(FAILED(r)) return r;
    
    r = g_pKeyboard->SetDataFormat(&c_dfDIKeyboard);
    if(FAILED(r)) return r;
    
    r = g_pKeyboard->SetCooperativeLevel(g_hWindow, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
    if(FAILED(r)) return r;

    g_pKeyboard->Acquire();

    return S_OK;
}

HRESULT InitDXAudio()
{
    HRESULT r;

    Debug("Initializing DirectSound...\n");

    g_pSoundManager = new CSoundManager();
    r = g_pSoundManager->Initialize(g_hWindow, DSSCL_PRIORITY, 1, 22050, 16);
    if(FAILED(r)) return r;

    r = g_pSoundManager->Get3DListenerInterface(&g_pListener);
    if(FAILED(r)) return r;

    Debug("Initializing DirectMusic...\n");

    g_pMusicManager = new CMusicManager();
    r = g_pMusicManager->Initialize(g_hWindow, 64, DMUS_APATH_SHARED_STEREOPLUSREVERB);
    if(FAILED(r))
	{
		r = g_pMusicManager->Initialize(g_hWindow, 32, DMUS_APATH_SHARED_STEREOPLUSREVERB);
		if(FAILED(r))
		{
			r = g_pMusicManager->Initialize(g_hWindow, 32, DMUS_APATH_DYNAMIC_STEREO);
			if(FAILED(r)) return r;
		}
		return r;
	}

    return S_OK;
}

HRESULT InitPlayer()
{
    MeshCreate(&g_Player.m_pMeshWalk[0], "Data\\PlayerWalk0.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[1], "Data\\PlayerWalk1.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[2], "Data\\PlayerWalk2.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[3], "Data\\PlayerWalk3.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[4], "Data\\PlayerWalk4.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[5], "Data\\PlayerWalk5.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[6], "Data\\PlayerWalk6.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshWalk[7], "Data\\PlayerWalk7.x", g_pD3DDevice, TRUE);
    for(int a = 0; a < NUM_PLAYER_WALK_STEPS; a++)
    {
        if(g_Player.m_pMeshWalk[a] == NULL) return E_FAIL;
    }

    MeshCreate(&g_Player.m_pMeshPush[0], "Data\\PlayerPush0.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[1], "Data\\PlayerPush1.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[2], "Data\\PlayerPush2.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[3], "Data\\PlayerPush3.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[4], "Data\\PlayerPush4.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[5], "Data\\PlayerPush5.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[6], "Data\\PlayerPush6.x", g_pD3DDevice, TRUE);
    MeshCreate(&g_Player.m_pMeshPush[7], "Data\\PlayerPush7.x", g_pD3DDevice, TRUE);
    for(a = 0; a < NUM_PLAYER_PUSH_STEPS; a++)
    {
        if(g_Player.m_pMeshPush[a] == NULL) return E_FAIL;
    }

    MeshCreate(&g_Player.m_pMeshJump, "Data\\PlayerJump.x", g_pD3DDevice, TRUE);
    if(g_Player.m_pMeshJump == NULL) return E_FAIL;

    MeshCreate(&g_Player.m_pMeshFall, "Data\\PlayerFall.x", g_pD3DDevice, TRUE);
    if(g_Player.m_pMeshFall == NULL) return E_FAIL;

    return S_OK;
}

HRESULT InitSounds()
{
    char    sFilename[256];
    HRESULT r;

    for(int a = 0; a < NUM_SOUNDS; a++)
    {
        sprintf(sFilename, "Data\\Sound%d.wav", a);

        for(int b = 0; b < NUM_SOUND_CLONES; b++)
        {
            g_pSound[a][b] = NULL;
            r = g_pSoundManager->Create(&g_pSound[a][b], sFilename, DSBCAPS_CTRL3D);
            if(FAILED(r)) return r;
        }
    }

    return S_OK;
}

HRESULT InitSprites()
{
    char    sFilename[256];
    HRESULT r;

    r = D3DXCreateSprite(g_pD3DDevice, &g_pSprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\DiamondSprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pDiamondSprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\CameraLockedSprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pCameraLockedSprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\ArrowSprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pArrowSprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\EnergyEmptySprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pEnergyEmptySprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\EnergyFullSprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pEnergyFullSprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\ShiningSprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pShiningSprite);
    if(FAILED(r)) return r;

    r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                    "Data\\SnowflakeSprite.bmp",
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    D3DX_DEFAULT,
                                    0,
                                    D3DFMT_UNKNOWN,
                                    D3DPOOL_DEFAULT,
                                    D3DX_FILTER_BOX,
                                    D3DX_FILTER_BOX,
                                    D3DCOLOR_XRGB(0, 0, 0),
                                    NULL,
                                    NULL,
                                    &g_pSnowflakeSprite);
    if(FAILED(r)) return r;

    for(int a = 0; a < NUM_PARTICLE_TYPES; a++)
    {
        sprintf(sFilename, "Data\\ParticleSprite%d.bmp", a);

        r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                        sFilename,
                                        D3DX_DEFAULT,
                                        D3DX_DEFAULT,
                                        D3DX_DEFAULT,
                                        0,
                                        D3DFMT_UNKNOWN,
                                        D3DPOOL_DEFAULT,
                                        D3DX_FILTER_BOX,
                                        D3DX_FILTER_BOX,
                                        D3DCOLOR_XRGB(0, 0, 0),
                                        NULL,
                                        NULL,
                                        &g_pParticleSprite[a]);
        if(FAILED(r)) return r;
    }

    for(a = 0; a < NUM_FIRE_SPRITES; a++)
    {
        sprintf(sFilename, "Data\\FireSprite%d.bmp", a);

        r = D3DXCreateTextureFromFileEx(g_pD3DDevice,
                                        sFilename,
                                        D3DX_DEFAULT,
                                        D3DX_DEFAULT,
                                        D3DX_DEFAULT,
                                        0,
                                        D3DFMT_UNKNOWN,
                                        D3DPOOL_DEFAULT,
                                        D3DX_FILTER_BOX,
                                        D3DX_FILTER_BOX,
                                        D3DCOLOR_XRGB(0, 0, 0),
                                        NULL,
                                        NULL,
                                        &g_pFireSprite[a]);
        if(FAILED(r)) return r;
    }

    return S_OK;
}

HRESULT InitShadow()
{
    MeshCreate(&g_pShadow, "Data\\Shadow.x", g_pD3DDevice);
    if(g_pShadow == NULL) return E_FAIL;
    g_pShadow->m_Material[0].Ambient.a = 0.75f;
    g_pShadow->m_Material[0].Diffuse.a = 0.75f;

    return S_OK;
}

HRESULT InitApplication()
{
    HRESULT     r;
    HINSTANCE   hD3D8DLL;
	MSG			Message;

    Debug("Searching for DirectX 8.0 or higher...\n");
    hD3D8DLL = LoadLibrary("D3D8.DLL");
    if(hD3D8DLL == NULL)
    {
        Debug("ERROR: DirectX 8.0 or higher could not be found!\n");
        return E_FAIL;
    }
    FreeLibrary(hD3D8DLL);
    Debug("DirectX 8.0 or higher found!\n");

    Debug("Initializing Direct3D...\n");
    r = InitD3D();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}

    Debug("Initializing fonts...\n");
    r = InitFonts();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing DirectInput...\n");
    r = InitDInput();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(0, 255, 0), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(255, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing DirectX Audio...\n");
    r = InitDXAudio();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(255, 255, 0), "Felder werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing field types...\n");
    r = InitFieldTypes();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(255, 255, 0), "Objekte werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing object types...\n");
    r = InitObjectTypes();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(255, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing player...\n");
    r = InitPlayer();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 270.0f, D3DCOLOR_XRGB(255, 255, 0), "Sounds werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing sounds...\n");
    r = InitSounds();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 270.0f, D3DCOLOR_XRGB(0, 255, 0), "Sounds werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 270.0f, D3DCOLOR_XRGB(0, 255, 0), "Sounds werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 290.0f, D3DCOLOR_XRGB(255, 255, 0), "Sprites werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing sprites...\n");
    r = InitSprites();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 270.0f, D3DCOLOR_XRGB(0, 255, 0), "Sounds werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 290.0f, D3DCOLOR_XRGB(0, 255, 0), "Sprites werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 270.0f, D3DCOLOR_XRGB(0, 255, 0), "Sounds werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 290.0f, D3DCOLOR_XRGB(0, 255, 0), "Sprites werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 310.0f, D3DCOLOR_XRGB(255, 255, 0), "Schatten werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);
    Debug("Initializing shadows...\n");
    r = InitShadow();
    if(FAILED(r)) return r;
	if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	g_pD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	g_pD3DDevice->BeginScene();
	g_pFontTahoma10->DrawText(50.0f, 50.0f, D3DCOLOR_XRGB(255, 255, 255), "ScherfgenSoftware", 0);
	g_pFontTahoma48->DrawText(46.0f, 60.0f, D3DCOLOR_XRGB(255, 0, 0), "Blocks 2001", 0);
	g_pFontTahoma10->DrawText(305.0f, 95.0f, D3DCOLOR_XRGB(255, 255, 255), "v1.1", 0);
	g_pFontTahoma10->DrawText(50.0f, 130.0f, D3DCOLOR_XRGB(0, 255, 0), "Direct3D wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 150.0f, D3DCOLOR_XRGB(0, 255, 0), "Schriftarten werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 170.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectInput wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 190.0f, D3DCOLOR_XRGB(0, 255, 0), "DirectX Audio wird initialisiert...", 0);
	g_pFontTahoma10->DrawText(50.0f, 210.0f, D3DCOLOR_XRGB(0, 255, 0), "Felder werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 230.0f, D3DCOLOR_XRGB(0, 255, 0), "Objekte werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 250.0f, D3DCOLOR_XRGB(0, 255, 0), "Spielfigurmodelle werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 270.0f, D3DCOLOR_XRGB(0, 255, 0), "Sounds werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 290.0f, D3DCOLOR_XRGB(0, 255, 0), "Sprites werden geladen...", 0);
	g_pFontTahoma10->DrawText(50.0f, 310.0f, D3DCOLOR_XRGB(0, 255, 0), "Schatten werden geladen...", 0);
	g_pD3DDevice->EndScene();
	g_pD3DDevice->Present(NULL, NULL, NULL, NULL);

	Debug("Looking for a joystick...\n");
	g_bUseJoystick = joyGetNumDevs() == 0 ? FALSE : TRUE;
	if(g_bUseJoystick)
	{
		Debug("Joystick found!\n");
		joyGetDevCaps(JOYSTICKID1, &g_JoyCaps, sizeof(JOYCAPS));
	}
	else Debug("No joystick found.\n");

    g_ObjectType[OT_FIRE].m_pMesh->m_pTexture[0] = g_pFireSprite[0];

    return S_OK;
}

void InitLevelRow(int y,
                  int z,
                  char *psRow)
{
    for(int x = 0; x < g_iLevelWidth; x++)
    {
        g_iLevel[x][y][z] = g_iSymbolTable[psRow[x]];
    }
}

void InitObjects()
{
    int     iField;
    int     iType;
    BOOL    bNewObject;

    for(int z = 0; z < g_iLevelDepth; z++)
    {
        for(int y = 0; y < g_iLevelHeight; y++)
        {
            for(int x = 0; x < g_iLevelWidth; x++)
            {
                iField = g_iLevel[x][y][z];

                bNewObject = FALSE;

                if(iField == -10) {iType = OT_BLOCK1; bNewObject = TRUE;}
                else if(iField == -11) {iType = OT_DIAMOND; bNewObject = TRUE;}
                else if(iField == -12) {iType = OT_PLATFORM1; bNewObject = TRUE;}
                else if(iField == -13) {iType = OT_PLATFORM2; bNewObject = TRUE;}
                else if(iField == -14) {iType = OT_ICE1; bNewObject = TRUE;}
                else if(iField == -15) {iType = OT_CANNON1; bNewObject = TRUE;}
                else if(iField == -16) {iType = OT_FIRE; bNewObject = TRUE;}

                if(bNewObject)
                {
                    CreateObject(iType, D3DXVECTOR3((float)(x) + 0.5f, (float)(y) + 0.5f, (float)(-z) - 0.5f) + g_ObjectType[iType].m_vAdd);
                    g_iLevel[x][y][z] = 255;
                }
            }
        }
    }
}

void InitPositions()
{
    int iField;

    for(int z = 0; z < g_iLevelDepth; z++)
    {
        for(int y = 0; y < g_iLevelHeight; y++)
        {
            for(int x = 0; x < g_iLevelWidth; x++)
            {
                iField = g_iLevel[x][y][z];

                if(iField == -100)
                {
                    g_Player.m_Control.SetPosition(D3DXVECTOR3((float)(x) + 0.5f, (float)(y), (float)(-z) - 0.5f));
                    g_iLevel[x][y][z] = 255;
                }
                else if(iField == -101)
                {
                    g_vLevelExit = D3DXVECTOR3((float)(x) + 0.5f, (float)(y) + 0.5f, (float)(-z) - 0.5f);
                    g_iLevel[x][y][z] = 255;
                }
            }
        }
    }    
}

void InitFieldFlagsArray()
{
    ZeroMemory(&g_iFieldFlags, sizeof(g_iFieldFlags));

    for(int z = 0; z <= g_iLevelDepth; z++)
    {
        for(int y = 0; y < g_iLevelHeight; y++)
        {
            for(int x = 0; x < g_iLevelWidth; x++)
            {
                g_iFieldFlags[x + 1][y + 1][z + 1] = g_FieldType[g_iLevel[x][y][z]].m_iFlags;
            }
        }
    }
}

void InitSnow()
{
    for(int a = 0; a < g_iNumSnowflakes; a++)
    {
        g_Snowflake[a].m_vPosition.x = FloatRandom(-5.0f, (float)(g_iLevelWidth) + 5.0f);
        g_Snowflake[a].m_vPosition.y = FloatRandom(0.0f, 20.0f);
        g_Snowflake[a].m_vPosition.z = FloatRandom((float)(-g_iLevelDepth) - 5.0f, 5.0f);
    }
}

void InitLevel()
{
    char                        sFilename[256];
    int                         iNumGroundTextures;
    float                       fFogStart;
    float                       fFogEnd;
    D3DXMATRIX                  mProjectionMatrix;
    IDirectMusicPerformance8    *pPerformance;

    pPerformance = g_pMusicManager->GetPerformance();
    pPerformance->SetGlobalParam(GUID_PerfMasterVolume, (void*)(&g_lMusicVolume), sizeof(g_lMusicVolume));

    sprintf(sFilename, "Data\\Level%d.mid", g_iLevelNr);
    PlayMusic(sFilename, TRUE);
    
    g_Player.m_fEnergy          = 1.0f;
    g_Player.m_fFalling         = 0.0f;
    g_Player.m_fSpeed           = 0.0f;
    g_Player.m_fStep            = 2.0f;
    g_Player.m_bOnGround        = TRUE;
    g_Player.m_bPush            = FALSE;
    g_Player.m_fJumpCountDown   = JUMP_COUNT_DOWN;
    g_Player.m_iNumDiamonds     = 0;

    g_Player.m_Control.Clear();

    g_iViewMode     = 0;
    g_fDieCountDown = DIE_DELAY;

    g_fCameraDistance       = STANDARD_CAMERA_DISTANCE;
    g_fCameraAngle          = 0.0f;
    g_fCameraHeight         = 0.0f;
    g_bCameraLocked         = FALSE;
    g_fLockCameraCountDown  = CAMERA_LOCK_DELAY;
    g_fMoveCameraCountDown  = 0.0f;
    g_bObjectsOutOfMap      = TRUE;

    g_fLevelFinished        = 0.0f;
    g_bLevelFinished        = FALSE;
    g_bLevelGroundRotating  = FALSE;
    g_bLevelGroundMoving    = FALSE;
    g_fGroundHeight         = 0.95f;
    g_fLightFlickering      = 0.0f;
    g_cLightColor           = D3DRGB(1.0f, 1.0f, 1.0f);
    g_bSnowing              = FALSE;
    g_iNumSnowflakes        = MAX_SNOWFLAKES;
    g_bRenderShining        = FALSE;
	g_fLevelExitRotation	= 0.0f;

    g_fCounter1             = 0.0f;
    g_fEnergyDisplay        = 0.0f;
    g_bMenu                 = FALSE;
    g_iMenuCursor           = 0;
    g_fMenuCursorCountDown  = MENU_CURSOR_DELAY;
    g_iMenuItemSelected     = -1;
    g_bEnterPassword        = FALSE;
    g_bConfirmQuit          = FALSE;
    g_bConfirmRestart       = FALSE;
    g_fConfirmQuitCountDown = CONFIRM_QUIT_DELAY;
    g_bCreditsMusicPlaying  = FALSE;

    fFogStart           = 0.0f;
    fFogEnd             = 40.0f;
    iNumGroundTextures  = 1;

    for(int z = 0; z < 50; z++)
    {
        for(int y = 0; y < 20; y++)
        {
            for(int x = 0; x < 50; x++)
            {
                g_iLevel[x][y][z] = 255;
            }
        }
    }

    ZeroMemory(&g_Object, sizeof(g_Object));
    ZeroMemory(&g_Particle, sizeof(g_Particle));
    ZeroMemory(&g_Snowflake, sizeof(g_Snowflake));
    ZeroMemory(&g_Script, sizeof(g_Script));

    if(g_iLevelNr == 0)
    {
        g_iLevelWidth           = 27;
        g_iLevelHeight          = 20;
        g_iLevelDepth           = 15;
        g_cBackgroundColor      = D3DCOLOR_XRGB(32, 32, 64);
        g_fGroundHeight         = -14.0f;
        fFogEnd                 = 100.0f;
        g_bMenu                 = TRUE;
        strcpy(g_sLevelName,      "");

        InitLevelRow(0,  0, "mmm  m    mm   mm m  m  mmm");
        InitLevelRow(0,  1, "m  m m   m  m m   m  m m   ");
        InitLevelRow(0,  2, "m  m m   m  m m   m m  m   ");
        InitLevelRow(0,  3, "mmm  m   m  m m   mm    mm ");
        InitLevelRow(0,  4, "m  m m   m  m m   m m     m");
        InitLevelRow(0,  5, "m  m m   m  m m   m  m    m");
        InitLevelRow(0,  6, "mmm  mmm  mm   mm m  m mmm ");

        InitLevelRow(0,  8, "    ...................    ");
        InitLevelRow(0,  9, "    ...................    ");
        InitLevelRow(0, 10, "    ...................    ");
        InitLevelRow(0, 11, "    ...................    ");
        InitLevelRow(0, 12, "    ...................    ");
        InitLevelRow(0, 13, "    ...................    ");
        InitLevelRow(0, 14, "    ...................    ");

        InitLevelRow(1,  8, "    m                 m    ");
        InitLevelRow(1,  9, "      000  00   00  0      ");
        InitLevelRow(1, 10, "        0 0  0 0  0 0      ");
        InitLevelRow(1, 11, "      000 0  0 0  0 0      ");
        InitLevelRow(1, 12, "      0   0  0 0  0 0      ");
        InitLevelRow(1, 13, "      000  00   00  0      ");
        InitLevelRow(1, 14, "    m                 m    ");

        InitLevelRow(2,  8, "    A                 A    ");
        InitLevelRow(2,  9, "                           ");
        InitLevelRow(2, 10, "                           ");
        InitLevelRow(2, 11, "                           ");
        InitLevelRow(2, 12, "                           ");
        InitLevelRow(2, 13, "                           ");
        InitLevelRow(2, 14, "    A                 A    ");
    }
    else if(g_iLevelNr == 1)
    {
        g_iLevelWidth           = 15;
        g_iLevelHeight          = 10;
        g_iLevelDepth           = 15;
        g_iLevelDiamonds        = 3;
        g_cBackgroundColor      = D3DCOLOR_XRGB(128, 128, 255);
		g_fLevelExitRotation    = D3DX_PI / 2.0f;
        strcpy(g_sLevelName,      "W i l l k o m m e n !");

        InitLevelRow(0,  0, "               ");
        InitLevelRow(0,  1, "               ");
        InitLevelRow(0,  2, "               ");
        InitLevelRow(0,  3, "               ");
        InitLevelRow(0,  4, "               ");
        InitLevelRow(0,  5, "     .....     ");
        InitLevelRow(0,  6, "     .....     ");
        InitLevelRow(0,  7, "     .....     ");
        InitLevelRow(0,  8, "     .....     ");
        InitLevelRow(0,  9, "     .....     ");
        InitLevelRow(0, 10, "               ");
        InitLevelRow(0, 11, "               ");
        InitLevelRow(0, 12, "               ");
        InitLevelRow(0, 13, "               ");
        InitLevelRow(0, 14, "               ");

        InitLevelRow(1,  0, "m             m");
        InitLevelRow(1,  1, "               ");
        InitLevelRow(1,  2, "               ");
        InitLevelRow(1,  3, "               ");
        InitLevelRow(1,  4, "    mmmmmmm    ");
        InitLevelRow(1,  5, "    m     m    ");
        InitLevelRow(1,  6, "    m     m    ");
        InitLevelRow(1,  7, "    m  1  m    ");
        InitLevelRow(1,  8, "    m     m    ");
        InitLevelRow(1,  9, "    m    mm    ");
        InitLevelRow(1, 10, "    mmmmmmm    ");
        InitLevelRow(1, 11, "               ");
        InitLevelRow(1, 12, "               ");
        InitLevelRow(1, 13, "               ");
        InitLevelRow(1, 14, "m             m");

        InitLevelRow(2,  0, "m.............m");
        InitLevelRow(2,  1, "...............");
        InitLevelRow(2,  2, "...............");
        InitLevelRow(2,  3, "...............");
        InitLevelRow(2,  4, "....mmmmmmm....");
        InitLevelRow(2,  5, "....m     m....");
        InitLevelRow(2,  6, "....m     m....");
        InitLevelRow(2,  7, "....m     m....");
        InitLevelRow(2,  8, "....m     m....");
        InitLevelRow(2,  9, "....m     m....");
        InitLevelRow(2, 10, "....mmmmmmm....");
        InitLevelRow(2, 11, "...............");
        InitLevelRow(2, 12, "...............");
        InitLevelRow(2, 13, "...............");
        InitLevelRow(2, 14, "m.............m");

        InitLevelRow(3,  0, "               ");
        InitLevelRow(3,  1, "               ");
        InitLevelRow(3,  2, "  !            ");
        InitLevelRow(3,  3, "               ");
        InitLevelRow(3,  4, "    i     i    ");
        InitLevelRow(3,  5, "               ");
        InitLevelRow(3,  6, "               ");
        InitLevelRow(3,  7, "            mm ");
        InitLevelRow(3,  8, " 0          ?  ");
        InitLevelRow(3,  9, "m1m         mm ");
        InitLevelRow(3, 10, "mmm i     i mm ");
        InitLevelRow(3, 11, "               ");
        InitLevelRow(3, 12, "               ");
        InitLevelRow(3, 13, "               ");
        InitLevelRow(3, 14, "               ");

        InitLevelRow(4,  0, "               ");
        InitLevelRow(4,  1, "               ");
        InitLevelRow(4,  2, "               ");
        InitLevelRow(4,  3, "               ");
        InitLevelRow(4,  4, "    i     i    ");
        InitLevelRow(4,  5, "               ");
        InitLevelRow(4,  6, "               ");
        InitLevelRow(4,  7, "            mm ");
        InitLevelRow(4,  8, " m          mm ");
        InitLevelRow(4,  9, "mmm         mm ");
        InitLevelRow(4, 10, "mmm i     i    ");
        InitLevelRow(4, 11, "               ");
        InitLevelRow(4, 12, "               ");
        InitLevelRow(4, 13, "               ");
        InitLevelRow(4, 14, "               ");

        InitLevelRow(5,  0, "               ");
        InitLevelRow(5,  1, "               ");
        InitLevelRow(5,  2, "               ");
        InitLevelRow(5,  3, "               ");
        InitLevelRow(5,  4, "    i     i    ");
        InitLevelRow(5,  5, "               ");
        InitLevelRow(5,  6, "            mm ");
        InitLevelRow(5,  7, "            mm ");
        InitLevelRow(5,  8, "            mm ");
        InitLevelRow(5,  9, "               ");
        InitLevelRow(5, 10, "    i     i    ");
        InitLevelRow(5, 11, "               ");
        InitLevelRow(5, 12, "               ");
        InitLevelRow(5, 13, "               ");
        InitLevelRow(5, 14, "               ");

        InitLevelRow(6,  0, "               ");
        InitLevelRow(6,  1, "               ");
        InitLevelRow(6,  2, "               ");
        InitLevelRow(6,  3, "               ");
        InitLevelRow(6,  4, "    MMmmmMM    ");
        InitLevelRow(6,  5, "    M     M    ");
        InitLevelRow(6,  6, "    m     m mm ");
        InitLevelRow(6,  7, "    m-mM  m mm ");
        InitLevelRow(6,  8, "    m     m    ");
        InitLevelRow(6,  9, "    M     M    ");
        InitLevelRow(6, 10, "    MMmmmMM    ");
        InitLevelRow(6, 11, "               ");
        InitLevelRow(6, 12, "               ");
        InitLevelRow(6, 13, "               ");
        InitLevelRow(6, 14, "               ");

        InitLevelRow(7,  0, "               ");
        InitLevelRow(7,  1, "               ");
        InitLevelRow(7,  2, "               ");
        InitLevelRow(7,  3, "               ");
        InitLevelRow(7,  4, "               ");
        InitLevelRow(7,  5, "               ");
        InitLevelRow(7,  6, "               ");
        InitLevelRow(7,  7, "       1       ");
        InitLevelRow(7,  8, "               ");
        InitLevelRow(7,  9, "               ");
        InitLevelRow(7, 10, "               ");
        InitLevelRow(7, 11, "               ");
        InitLevelRow(7, 12, "               ");
        InitLevelRow(7, 13, "               ");
        InitLevelRow(7, 14, "               ");
    }
    else if(g_iLevelNr == 2)
    {
        g_iLevelWidth           = 15;
        g_iLevelHeight          = 10;
        g_iLevelDepth           = 15;
        g_iLevelDiamonds        = 3;
        g_bRenderShining        = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(0, 0, 0);
		g_fLevelExitRotation    = D3DX_PI / 2.0f;
        fFogEnd                 = 30.0f;
        strcpy(g_sLevelName,      "Schloss Trauerstein");

        InitLevelRow(0,  0, "...............");
        InitLevelRow(0,  1, "...............");
        InitLevelRow(0,  2, "...............");
        InitLevelRow(0,  3, "...............");
        InitLevelRow(0,  4, "...............");
        InitLevelRow(0,  5, "...............");
        InitLevelRow(0,  6, "...............");
        InitLevelRow(0,  7, "...............");
        InitLevelRow(0,  8, "...............");
        InitLevelRow(0,  9, "...............");
        InitLevelRow(0, 10, "...............");
        InitLevelRow(0, 11, "...............");
        InitLevelRow(0, 12, "...............");
        InitLevelRow(0, 13, "...............");
        InitLevelRow(0, 14, "...............");

        InitLevelRow(1,  0, "mmmmmmmmmmmmmmm");
        InitLevelRow(1,  1, "mm     m     mm");
        InitLevelRow(1,  2, "m      m      m");
        InitLevelRow(1,  3, "m  !   ?      m");
        InitLevelRow(1,  4, "m      m      m");
        InitLevelRow(1,  5, "m    mmmmm    m");
        InitLevelRow(1,  6, "m    m   m    m");
        InitLevelRow(1,  7, "m    mm 1m    m");
        InitLevelRow(1,  8, "m    m   m    m");
        InitLevelRow(1,  9, "m    mmmmm    m");
        InitLevelRow(1, 10, "m             m");
        InitLevelRow(1, 11, "m  0          m");
        InitLevelRow(1, 12, "m             m");
        InitLevelRow(1, 13, "mm           mm");
        InitLevelRow(1, 14, "mmmmmmmmmmmmmmm");

        InitLevelRow(2,  0, "mmmmmmmmmmmmmmm");
        InitLevelRow(2,  1, "m      m      m");
        InitLevelRow(2,  2, "m      m      m");
        InitLevelRow(2,  3, "m      m      m");
        InitLevelRow(2,  4, "m      m      m");
        InitLevelRow(2,  5, "m    mmmmm    m");
        InitLevelRow(2,  6, "m    m   m    m");
        InitLevelRow(2,  7, "m    m   m    m");
        InitLevelRow(2,  8, "m    m   m    m");
        InitLevelRow(2,  9, "m    mmmmm    m");
        InitLevelRow(2, 10, "m             m");
        InitLevelRow(2, 11, "m             m");
        InitLevelRow(2, 12, "m             m");
        InitLevelRow(2, 13, "m             m");
        InitLevelRow(2, 14, "mmmmmmmmmmmmmmm");

        InitLevelRow(3,  0, "m             m");
        InitLevelRow(3,  1, "              1");
        InitLevelRow(3,  2, "               ");
        InitLevelRow(3,  3, "              m");
        InitLevelRow(3,  4, "              m");
        InitLevelRow(3,  5, "     mmmmm     ");
        InitLevelRow(3,  6, "     m   m     ");
        InitLevelRow(3,  7, "         m     ");
        InitLevelRow(3,  8, "     m   m     ");
        InitLevelRow(3,  9, "     mmmmm     ");
        InitLevelRow(3, 10, "               ");
        InitLevelRow(3, 11, "               ");
        InitLevelRow(3, 12, "               ");
        InitLevelRow(3, 13, "               ");
        InitLevelRow(3, 14, "m             m");

        InitLevelRow(4,  0, "m             m");
        InitLevelRow(4,  1, "               ");
        InitLevelRow(4,  2, "               ");
        InitLevelRow(4,  3, "               ");
        InitLevelRow(4,  4, "              m");
        InitLevelRow(4,  5, "     mmmmm    m");
        InitLevelRow(4,  6, "     m   m     ");
        InitLevelRow(4,  7, "     m         ");
        InitLevelRow(4,  8, "     m   m     ");
        InitLevelRow(4,  9, "     mm mm     ");
        InitLevelRow(4, 10, "               ");
        InitLevelRow(4, 11, "               ");
        InitLevelRow(4, 12, "               ");
        InitLevelRow(4, 13, "               ");
        InitLevelRow(4, 14, "m             m");

        InitLevelRow(5,  0, "A             A");
        InitLevelRow(5,  1, "               ");
        InitLevelRow(5,  2, "               ");
        InitLevelRow(5,  3, "               ");
        InitLevelRow(5,  4, "               ");
        InitLevelRow(5,  5, "     mmmmm    m");
        InitLevelRow(5,  6, "     m...m    m");
        InitLevelRow(5,  7, "     m...m     ");
        InitLevelRow(5,  8, "     m...m     ");
        InitLevelRow(5,  9, "     mmmmm     ");
        InitLevelRow(5, 10, "               ");
        InitLevelRow(5, 11, "               ");
        InitLevelRow(5, 12, "               ");
        InitLevelRow(5, 13, "               ");
        InitLevelRow(5, 14, "A             A");

        InitLevelRow(6,  0, "               ");
        InitLevelRow(6,  1, "               ");
        InitLevelRow(6,  2, "               ");
        InitLevelRow(6,  3, "               ");
        InitLevelRow(6,  4, "               ");
        InitLevelRow(6,  5, "     m m m     ");
        InitLevelRow(6,  6, "             mm");
        InitLevelRow(6,  7, "     m 1 m     ");
        InitLevelRow(6,  8, "               ");
        InitLevelRow(6,  9, "     m m m     ");
        InitLevelRow(6, 10, "               ");
        InitLevelRow(6, 11, "               ");
        InitLevelRow(6, 12, "               ");
        InitLevelRow(6, 13, "               ");
        InitLevelRow(6, 14, "               ");
    }  
    else if(g_iLevelNr == 3)
    {
        g_iLevelWidth           = 15;
        g_iLevelHeight          = 10;
        g_iLevelDepth           = 15;
        g_iLevelDiamonds        = 1;
        g_bLevelGroundRotating  = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(128, 0, 255);
        g_fLevelExitRotation    = D3DX_PI / 2.0f;
        strcpy(g_sLevelName,      "M a e r c h e n l a n d");

        InitLevelRow(0,  0, "...............");
        InitLevelRow(0,  1, "...............");
        InitLevelRow(0,  2, "...............");
        InitLevelRow(0,  3, "...............");
        InitLevelRow(0,  4, "...............");
        InitLevelRow(0,  5, "...............");
        InitLevelRow(0,  6, "...............");
        InitLevelRow(0,  7, "...............");
        InitLevelRow(0,  8, "...............");
        InitLevelRow(0,  9, "...............");
        InitLevelRow(0, 10, "...............");
        InitLevelRow(0, 11, "...............");
        InitLevelRow(0, 12, "...............");
        InitLevelRow(0, 13, "...............");
        InitLevelRow(0, 14, "...............");

        InitLevelRow(1,  0, "               ");
        InitLevelRow(1,  1, " !             ");
        InitLevelRow(1,  2, " ?        mm   ");
        InitLevelRow(1,  3, "               ");
        InitLevelRow(1,  4, "       mmmmmm  ");
        InitLevelRow(1,  5, "               ");
        InitLevelRow(1,  6, "               ");
        InitLevelRow(1,  7, "               ");
        InitLevelRow(1,  8, "           m   ");
        InitLevelRow(1,  9, "           m   ");
        InitLevelRow(1, 10, "               ");
        InitLevelRow(1, 11, "               ");
        InitLevelRow(1, 12, "          i    ");
        InitLevelRow(1, 13, "               ");
        InitLevelRow(1, 14, "               ");

        InitLevelRow(2,  0, "               ");
        InitLevelRow(2,  1, "               ");
        InitLevelRow(2,  2, "           m   ");
        InitLevelRow(2,  3, "               ");
        InitLevelRow(2,  4, "         0  m  ");
        InitLevelRow(2,  5, "               ");
        InitLevelRow(2,  6, "               ");
        InitLevelRow(2,  7, "           m   ");
        InitLevelRow(2,  8, "           m   ");
        InitLevelRow(2,  9, "               ");
        InitLevelRow(2, 10, "               ");
        InitLevelRow(2, 11, "               ");
        InitLevelRow(2, 12, "          i    ");
        InitLevelRow(2, 13, "               ");
        InitLevelRow(2, 14, "               ");

        InitLevelRow(3,  0, "               ");
        InitLevelRow(3,  1, "               ");
        InitLevelRow(3,  2, "               ");
        InitLevelRow(3,  3, "               ");
        InitLevelRow(3,  4, "       mmmmmm  ");
        InitLevelRow(3,  5, "               ");
        InitLevelRow(3,  6, "           m   ");
        InitLevelRow(3,  7, "           m   ");
        InitLevelRow(3,  8, "               ");
        InitLevelRow(3,  9, "               ");
        InitLevelRow(3, 10, "               ");
        InitLevelRow(3, 11, "               ");
        InitLevelRow(3, 12, "          i    ");
        InitLevelRow(3, 13, "               ");
        InitLevelRow(3, 14, "               ");

        InitLevelRow(4,  0, "               ");
        InitLevelRow(4,  1, "               ");
        InitLevelRow(4,  2, "               ");
        InitLevelRow(4,  3, "               ");
        InitLevelRow(4,  4, "         0  m  ");
        InitLevelRow(4,  5, "               ");
        InitLevelRow(4,  6, "           mm  ");
        InitLevelRow(4,  7, "               ");
        InitLevelRow(4,  8, "               ");
        InitLevelRow(4,  9, "               ");
        InitLevelRow(4, 10, "               ");
        InitLevelRow(4, 11, "               ");
        InitLevelRow(4, 12, "          M    ");
        InitLevelRow(4, 13, "               ");
        InitLevelRow(4, 14, "               ");

        InitLevelRow(5,  0, "               ");
        InitLevelRow(5,  1, "               ");
        InitLevelRow(5,  2, "               ");
        InitLevelRow(5,  3, "               ");
        InitLevelRow(5,  4, "       mmmm m  ");
        InitLevelRow(5,  5, "               ");
        InitLevelRow(5,  6, "               ");
        InitLevelRow(5,  7, "               ");
        InitLevelRow(5,  8, "               ");
        InitLevelRow(5,  9, "               ");
        InitLevelRow(5, 10, "               ");
        InitLevelRow(5, 11, "               ");
        InitLevelRow(5, 12, "       M       ");
        InitLevelRow(5, 13, "               ");
        InitLevelRow(5, 14, "               ");

        InitLevelRow(6,  0, "               ");
        InitLevelRow(6,  1, "               ");
        InitLevelRow(6,  2, "               ");
        InitLevelRow(6,  3, "               ");
        InitLevelRow(6,  4, "         0     ");
        InitLevelRow(6,  5, "               ");
        InitLevelRow(6,  6, "               ");
        InitLevelRow(6,  7, "               ");
        InitLevelRow(6,  8, "               ");
        InitLevelRow(6,  9, "               ");
        InitLevelRow(6, 10, "               ");
        InitLevelRow(6, 11, "               ");
        InitLevelRow(6, 12, " MM M          ");
        InitLevelRow(6, 13, "               ");
        InitLevelRow(6, 14, "               ");

        InitLevelRow(7,  0, "               ");
        InitLevelRow(7,  1, "               ");
        InitLevelRow(7,  2, "               ");
        InitLevelRow(7,  3, "               ");
        InitLevelRow(7,  4, "               ");
        InitLevelRow(7,  5, "               ");
        InitLevelRow(7,  6, "               ");
        InitLevelRow(7,  7, "               ");
        InitLevelRow(7,  8, "               ");
        InitLevelRow(7,  9, "               ");
        InitLevelRow(7, 10, "               ");
        InitLevelRow(7, 11, "               ");
        InitLevelRow(7, 12, " 1             ");
        InitLevelRow(7, 13, "               ");
        InitLevelRow(7, 14, "               ");
    }
    else if(g_iLevelNr == 4)
    {
        g_iLevelWidth           = 15;
        g_iLevelHeight          = 10;
        g_iLevelDepth           = 15;
        g_iLevelDiamonds        = 1;
        g_bLevelGroundMoving    = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(192, 64, 64);
        fFogEnd                 = 25.0f;
        strcpy(g_sLevelName,      "H o e l l e n f e u e r");

        InitLevelRow(0,  0, "...............");
        InitLevelRow(0,  1, "...............");
        InitLevelRow(0,  2, "...............");
        InitLevelRow(0,  3, "...............");
        InitLevelRow(0,  4, "...............");
        InitLevelRow(0,  5, "...............");
        InitLevelRow(0,  6, "...............");
        InitLevelRow(0,  7, "...............");
        InitLevelRow(0,  8, "...............");
        InitLevelRow(0,  9, "...............");
        InitLevelRow(0, 10, "...............");
        InitLevelRow(0, 11, "...............");
        InitLevelRow(0, 12, "...............");
        InitLevelRow(0, 13, "...............");
        InitLevelRow(0, 14, "...............");

        InitLevelRow(1,  0, "i              ");
        InitLevelRow(1,  1, "               ");
        InitLevelRow(1,  2, "mm             ");
        InitLevelRow(1,  3, "               ");
        InitLevelRow(1,  4, "               ");
        InitLevelRow(1,  5, "  m?m          ");
        InitLevelRow(1,  6, "  m            ");
        InitLevelRow(1,  7, "  m    !       ");
        InitLevelRow(1,  8, "               ");
        InitLevelRow(1,  9, "            i  ");
        InitLevelRow(1, 10, "               ");
        InitLevelRow(1, 11, "   i     i     ");
        InitLevelRow(1, 12, "               ");
        InitLevelRow(1, 13, "               ");
        InitLevelRow(1, 14, "               ");

        InitLevelRow(2,  0, "M              ");
        InitLevelRow(2,  1, "               ");
        InitLevelRow(2,  2, "mm             ");
        InitLevelRow(2,  3, "               ");
        InitLevelRow(2,  4, "               ");
        InitLevelRow(2,  5, "  mmm          ");
        InitLevelRow(2,  6, "  m            ");
        InitLevelRow(2,  7, "               ");
        InitLevelRow(2,  8, "               ");
        InitLevelRow(2,  9, "            i  ");
        InitLevelRow(2, 10, "               ");
        InitLevelRow(2, 11, "   i     i     ");
        InitLevelRow(2, 12, "               ");
        InitLevelRow(2, 13, "               ");
        InitLevelRow(2, 14, "               ");

        InitLevelRow(3,  0, "               ");
        InitLevelRow(3,  1, "               ");
        InitLevelRow(3,  2, "mm             ");
        InitLevelRow(3,  3, "               ");
        InitLevelRow(3,  4, "        3  3   ");
        InitLevelRow(3,  5, "  mmm          ");
        InitLevelRow(3,  6, "               ");
        InitLevelRow(3,  7, "               ");
        InitLevelRow(3,  8, "               ");
        InitLevelRow(3,  9, "           mm  ");
        InitLevelRow(3, 10, "   m       m   ");
        InitLevelRow(3, 11, "  mm  2  mmm   ");
        InitLevelRow(3, 12, "               ");
        InitLevelRow(3, 13, "               ");
        InitLevelRow(3, 14, "               ");

        InitLevelRow(4,  0, "1              ");
        InitLevelRow(4,  1, "               ");
        InitLevelRow(4,  2, "mm             ");
        InitLevelRow(4,  3, "               ");
        InitLevelRow(4,  4, "               ");
        InitLevelRow(4,  5, "               ");
        InitLevelRow(4,  6, "               ");
        InitLevelRow(4,  7, "               ");
        InitLevelRow(4,  8, "               ");
        InitLevelRow(4,  9, "               ");
        InitLevelRow(4, 10, "   0           ");
        InitLevelRow(4, 11, "               ");
        InitLevelRow(4, 12, "               ");
        InitLevelRow(4, 13, "               ");
        InitLevelRow(4, 14, "               ");

        InitLevelRow(5,  0, "               ");
        InitLevelRow(5,  1, "               ");
        InitLevelRow(5,  2, "mm             ");
        InitLevelRow(5,  3, "               ");
        InitLevelRow(5,  4, "               ");
        InitLevelRow(5,  5, "               ");
        InitLevelRow(5,  6, "               ");
        InitLevelRow(5,  7, "               ");
        InitLevelRow(5,  8, "               ");
        InitLevelRow(5,  9, "               ");
        InitLevelRow(5, 10, "               ");
        InitLevelRow(5, 11, "               ");
        InitLevelRow(5, 12, "               ");
        InitLevelRow(5, 13, "               ");
        InitLevelRow(5, 14, "               ");
    }
    else if(g_iLevelNr == 5)
    {
        g_iLevelWidth           = 15;
        g_iLevelHeight          = 10;
        g_iLevelDepth           = 15;
        g_iLevelDiamonds        = 1;
        g_bSnowing              = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(224, 224, 224);
        g_fLevelExitRotation    = D3DX_PI / 2.0f;
        fFogEnd                 = 30.0f;
        strcpy(g_sLevelName,      "Auf schmalem Pfad");

        InitLevelRow(0,  0, "...............");
        InitLevelRow(0,  1, "...............");
        InitLevelRow(0,  2, "...............");
        InitLevelRow(0,  3, "...............");
        InitLevelRow(0,  4, "...............");
        InitLevelRow(0,  5, "...............");
        InitLevelRow(0,  6, "...............");
        InitLevelRow(0,  7, "...............");
        InitLevelRow(0,  8, "...............");
        InitLevelRow(0,  9, "...............");
        InitLevelRow(0, 10, "...............");
        InitLevelRow(0, 11, "...............");
        InitLevelRow(0, 12, "...............");
        InitLevelRow(0, 13, "...............");
        InitLevelRow(0, 14, "...............");

        InitLevelRow(1,  0, "        m      ");
        InitLevelRow(1,  1, "               ");
        InitLevelRow(1,  2, "               ");
        InitLevelRow(1,  3, "  4 m 4        ");
        InitLevelRow(1,  4, "    m      4   ");
        InitLevelRow(1,  5, "               ");
        InitLevelRow(1,  6, "         4     ");
        InitLevelRow(1,  7, "               ");
        InitLevelRow(1,  8, "         !     ");
        InitLevelRow(1,  9, "               ");
        InitLevelRow(1, 10, "            4  ");
        InitLevelRow(1, 11, "       4       ");
        InitLevelRow(1, 12, "               ");
        InitLevelRow(1, 13, "    mmmmm      ");
        InitLevelRow(1, 14, "               ");

        InitLevelRow(2,  0, "        m      ");
        InitLevelRow(2,  1, "               ");
        InitLevelRow(2,  2, "               ");
        InitLevelRow(2,  3, "               ");
        InitLevelRow(2,  4, "    m          ");
        InitLevelRow(2,  5, "    m          ");
        InitLevelRow(2,  6, "               ");
        InitLevelRow(2,  7, "               ");
        InitLevelRow(2,  8, "               ");
        InitLevelRow(2,  9, "               ");
        InitLevelRow(2, 10, "               ");
        InitLevelRow(2, 11, "               ");
        InitLevelRow(2, 12, "               ");
        InitLevelRow(2, 13, "               ");
        InitLevelRow(2, 14, "               ");

        InitLevelRow(3,  0, "        m      ");
        InitLevelRow(3,  1, "               ");
        InitLevelRow(3,  2, "               ");
        InitLevelRow(3,  3, "               ");
        InitLevelRow(3,  4, "               ");
        InitLevelRow(3,  5, "    m          ");
        InitLevelRow(3,  6, "    m          ");
        InitLevelRow(3,  7, "               ");
        InitLevelRow(3,  8, "               ");
        InitLevelRow(3,  9, "               ");
        InitLevelRow(3, 10, "               ");
        InitLevelRow(3, 11, "               ");
        InitLevelRow(3, 12, "               ");
        InitLevelRow(3, 13, "       m       ");
        InitLevelRow(3, 14, "               ");

        InitLevelRow(4,  0, "        m      ");
        InitLevelRow(4,  1, "           ppp ");
        InitLevelRow(4,  2, "               ");
        InitLevelRow(4,  3, "               ");
        InitLevelRow(4,  4, "           3   ");
        InitLevelRow(4,  5, "               ");
        InitLevelRow(4,  6, "    m          ");
        InitLevelRow(4,  7, "    p      p   ");
        InitLevelRow(4,  8, "   ppppppppp   ");
        InitLevelRow(4,  9, "           p   ");
        InitLevelRow(4, 10, "           p   ");
        InitLevelRow(4, 11, "           p   ");
        InitLevelRow(4, 12, "        ppppp  ");
        InitLevelRow(4, 13, "       mpp     ");
        InitLevelRow(4, 14, "               ");

        InitLevelRow(5,  0, "        m      ");
        InitLevelRow(5,  1, "             ? ");
        InitLevelRow(5,  2, "               ");
        InitLevelRow(5,  3, "               ");
        InitLevelRow(5,  4, "               ");
        InitLevelRow(5,  5, "               ");
        InitLevelRow(5,  6, "               ");
        InitLevelRow(5,  7, "    0          ");
        InitLevelRow(5,  8, "               ");
        InitLevelRow(5,  9, "               ");
        InitLevelRow(5, 10, "               ");
        InitLevelRow(5, 11, "               ");
        InitLevelRow(5, 12, "               ");
        InitLevelRow(5, 13, "               ");
        InitLevelRow(5, 14, "               ");

        InitLevelRow(6,  0, "        m      ");
        InitLevelRow(6,  1, "               ");
        InitLevelRow(6,  2, "               ");
        InitLevelRow(6,  3, "               ");
        InitLevelRow(6,  4, "               ");
        InitLevelRow(6,  5, "               ");
        InitLevelRow(6,  6, "               ");
        InitLevelRow(6,  7, "    0          ");
        InitLevelRow(6,  8, "               ");
        InitLevelRow(6,  9, "               ");
        InitLevelRow(6, 10, "               ");
        InitLevelRow(6, 11, "               ");
        InitLevelRow(6, 12, "               ");
        InitLevelRow(6, 13, "1              ");
        InitLevelRow(6, 14, "               ");

        InitLevelRow(7,  0, "        5      ");
        InitLevelRow(7,  1, "               ");
        InitLevelRow(7,  2, "               ");
        InitLevelRow(7,  3, "               ");
        InitLevelRow(7,  4, "               ");
        InitLevelRow(7,  5, "               ");
        InitLevelRow(7,  6, "               ");
        InitLevelRow(7,  7, "               ");
        InitLevelRow(7,  8, "               ");
        InitLevelRow(7,  9, "               ");
        InitLevelRow(7, 10, "               ");
        InitLevelRow(7, 11, "               ");
        InitLevelRow(7, 12, "               ");
        InitLevelRow(7, 13, "               ");
        InitLevelRow(7, 14, "               ");
    }
    else if(g_iLevelNr == 6)
    {
        g_iLevelWidth           = 20;
        g_iLevelHeight          = 15;
        g_iLevelDepth           = 20;
        g_iLevelDiamonds        = 2;
        g_bRenderShining        = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(0, 0, 0);
        g_bObjectsOutOfMap      = FALSE;
        g_fGroundHeight         = 0.0f;
        g_fLightFlickering      = 0.25f;
        g_cLightColor           = D3DRGB(1.0f, 1.0f, 0.0f);
        fFogEnd                 = 20.0f;
        iNumGroundTextures      = 2;
        g_vCamera               = D3DXVECTOR3(0.5f, 2.0f, -0.5f);
        strcpy(g_sLevelName,      "K o e n i g s g r a b");

        InitLevelRow(0,  0, "                mm  ");
        InitLevelRow(0,  1, "                mm?m");
        InitLevelRow(0,  2, "                    ");
        InitLevelRow(0,  3, "                    ");
        InitLevelRow(0,  4, "                    ");
        InitLevelRow(0,  5, "                    ");
        InitLevelRow(0,  6, "                    ");
        InitLevelRow(0,  7, "          f         ");
        InitLevelRow(0,  8, "                    ");
        InitLevelRow(0,  9, "                    ");
        InitLevelRow(0, 10, "                    ");
        InitLevelRow(0, 11, "          !         ");
        InitLevelRow(0, 12, "                    ");
        InitLevelRow(0, 13, "                    ");
        InitLevelRow(0, 14, "                    ");
        InitLevelRow(0, 15, "               f    ");
        InitLevelRow(0, 16, "                    ");
        InitLevelRow(0, 17, "                 mmm");
        InitLevelRow(0, 18, "       m         mmm");
        InitLevelRow(0, 19, "                 mmm");

        InitLevelRow(1,  0, "            pp      ");
        InitLevelRow(1,  1, "            pp      ");
        InitLevelRow(1,  2, "                    ");
        InitLevelRow(1,  3, "                    ");
        InitLevelRow(1,  4, "                    ");
        InitLevelRow(1,  5, "                    ");
        InitLevelRow(1,  6, "                    ");
        InitLevelRow(1,  7, "          6         ");
        InitLevelRow(1,  8, "                    ");
        InitLevelRow(1,  9, "                    ");
        InitLevelRow(1, 10, "                    ");
        InitLevelRow(1, 11, "                    ");
        InitLevelRow(1, 12, "                    ");
        InitLevelRow(1, 13, "                    ");
        InitLevelRow(1, 14, "                    ");
        InitLevelRow(1, 15, "               6    ");
        InitLevelRow(1, 16, "                    ");
        InitLevelRow(1, 17, "                    ");
        InitLevelRow(1, 18, "       m          s ");
        InitLevelRow(1, 19, "                    ");

        InitLevelRow(2,  0, "         pp         ");
        InitLevelRow(2,  1, "         pp         ");
        InitLevelRow(2,  2, "                    ");
        InitLevelRow(2,  3, "                    ");
        InitLevelRow(2,  4, "                    ");
        InitLevelRow(2,  5, "                    ");
        InitLevelRow(2,  6, "                    ");
        InitLevelRow(2,  7, "                    ");
        InitLevelRow(2,  8, "                    ");
        InitLevelRow(2,  9, "                    ");
        InitLevelRow(2, 10, "                    ");
        InitLevelRow(2, 11, "                    ");
        InitLevelRow(2, 12, "                    ");
        InitLevelRow(2, 13, "                    ");
        InitLevelRow(2, 14, "                    ");
        InitLevelRow(2, 15, "                    ");
        InitLevelRow(2, 16, "                    ");
        InitLevelRow(2, 17, "                    ");
        InitLevelRow(2, 18, "       s            ");
        InitLevelRow(2, 19, "                    ");

        InitLevelRow(3,  0, "     pp             ");
        InitLevelRow(3,  1, "     pp             ");
        InitLevelRow(3,  2, "                    ");
        InitLevelRow(3,  3, "                    ");
        InitLevelRow(3,  4, "                    ");
        InitLevelRow(3,  5, "                    ");
        InitLevelRow(3,  6, "                    ");
        InitLevelRow(3,  7, "                    ");
        InitLevelRow(3,  8, "                    ");
        InitLevelRow(3,  9, "                    ");
        InitLevelRow(3, 10, "                    ");
        InitLevelRow(3, 11, "                    ");
        InitLevelRow(3, 12, "                    ");
        InitLevelRow(3, 13, "                    ");
        InitLevelRow(3, 14, "                    ");
        InitLevelRow(3, 15, "                    ");
        InitLevelRow(3, 16, "                    ");
        InitLevelRow(3, 17, "                    ");
        InitLevelRow(3, 18, "                    ");
        InitLevelRow(3, 19, "                    ");

        InitLevelRow(4,  0, "mm                  ");
        InitLevelRow(4,  1, "mm                  ");
        InitLevelRow(4,  2, "                    ");
        InitLevelRow(4,  3, "                    ");
        InitLevelRow(4,  4, "                    ");
        InitLevelRow(4,  5, "                    ");
        InitLevelRow(4,  6, "                    ");
        InitLevelRow(4,  7, "                    ");
        InitLevelRow(4,  8, "                    ");
        InitLevelRow(4,  9, "                    ");
        InitLevelRow(4, 10, "                    ");
        InitLevelRow(4, 11, "                    ");
        InitLevelRow(4, 12, "                    ");
        InitLevelRow(4, 13, "                    ");
        InitLevelRow(4, 14, "                    ");
        InitLevelRow(4, 15, "                    ");
        InitLevelRow(4, 16, "                    ");
        InitLevelRow(4, 17, "                    ");
        InitLevelRow(4, 18, "                    ");
        InitLevelRow(4, 19, "                    ");

        InitLevelRow(5,  0, "                    ");
        InitLevelRow(5,  1, "                    ");
        InitLevelRow(5,  2, "                    ");
        InitLevelRow(5,  3, "                    ");
        InitLevelRow(5,  4, "                    ");
        InitLevelRow(5,  5, "pp                  ");
        InitLevelRow(5,  6, "pp                  ");
        InitLevelRow(5,  7, "                    ");
        InitLevelRow(5,  8, "                    ");
        InitLevelRow(5,  9, "                    ");
        InitLevelRow(5, 10, "                    ");
        InitLevelRow(5, 11, "                    ");
        InitLevelRow(5, 12, "                    ");
        InitLevelRow(5, 13, "                    ");
        InitLevelRow(5, 14, "                    ");
        InitLevelRow(5, 15, "                    ");
        InitLevelRow(5, 16, "                    ");
        InitLevelRow(5, 17, "                    ");
        InitLevelRow(5, 18, "                    ");
        InitLevelRow(5, 19, "                    ");

        InitLevelRow(6,  0, "                    ");
        InitLevelRow(6,  1, "                    ");
        InitLevelRow(6,  2, "                    ");
        InitLevelRow(6,  3, "                    ");
        InitLevelRow(6,  4, "                    ");
        InitLevelRow(6,  5, "                    ");
        InitLevelRow(6,  6, "                    ");
        InitLevelRow(6,  7, "                    ");
        InitLevelRow(6,  8, "pp                  ");
        InitLevelRow(6,  9, "pp                  ");
        InitLevelRow(6, 10, "                   p");
        InitLevelRow(6, 11, "                 ppm");
        InitLevelRow(6, 12, "                 ppm");
        InitLevelRow(6, 13, "                 ppm");
        InitLevelRow(6, 14, "                    ");
        InitLevelRow(6, 15, "                    ");
        InitLevelRow(6, 16, "                    ");
        InitLevelRow(6, 17, "                    ");
        InitLevelRow(6, 18, "                    ");
        InitLevelRow(6, 19, "                    ");

        InitLevelRow(7,  0, "                    ");
        InitLevelRow(7,  1, "                    ");
        InitLevelRow(7,  2, "                    ");
        InitLevelRow(7,  3, "         ppp        ");
        InitLevelRow(7,  4, "         pmp        ");
        InitLevelRow(7,  5, "         ppp        ");
        InitLevelRow(7,  6, "                    ");
        InitLevelRow(7,  7, "                    ");
        InitLevelRow(7,  8, "                    ");
        InitLevelRow(7,  9, "                    ");
        InitLevelRow(7, 10, "                   f");
        InitLevelRow(7, 11, "                    ");
        InitLevelRow(7, 12, "mmmpp  2   ppp    s ");
        InitLevelRow(7, 13, "mmm                 ");
        InitLevelRow(7, 14, "ppp                 ");
        InitLevelRow(7, 15, "                    ");
        InitLevelRow(7, 16, "                 ppp");
        InitLevelRow(7, 17, " pp              ppp");
        InitLevelRow(7, 18, " pppppp          ppp");
        InitLevelRow(7, 19, "     mm          ppm");

        InitLevelRow(8,  0, "                    ");
        InitLevelRow(8,  1, "                    ");
        InitLevelRow(8,  2, "                    ");
        InitLevelRow(8,  3, "                    ");
        InitLevelRow(8,  4, "          1         ");
        InitLevelRow(8,  5, "                    ");
        InitLevelRow(8,  6, "                    ");
        InitLevelRow(8,  7, "                    ");
        InitLevelRow(8,  8, "                    ");
        InitLevelRow(8,  9, "                    ");
        InitLevelRow(8, 10, "                   6");
        InitLevelRow(8, 11, "                    ");
        InitLevelRow(8, 12, "                    ");
        InitLevelRow(8, 13, "f                   ");
        InitLevelRow(8, 14, "  0                 ");
        InitLevelRow(8, 15, "                    ");
        InitLevelRow(8, 16, "                 mmm");
        InitLevelRow(8, 17, "                    ");
        InitLevelRow(8, 18, "                  1 ");
        InitLevelRow(8, 19, "                    ");

        InitLevelRow(9,  0, "                    ");
        InitLevelRow(9,  1, "                    ");
        InitLevelRow(9,  2, "                    ");
        InitLevelRow(9,  3, "                    ");
        InitLevelRow(9,  4, "                    ");
        InitLevelRow(9,  5, "                    ");
        InitLevelRow(9,  6, "                    ");
        InitLevelRow(9,  7, "                    ");
        InitLevelRow(9,  8, "                    ");
        InitLevelRow(9,  9, "                    ");
        InitLevelRow(9, 10, "                    ");
        InitLevelRow(9, 11, "                    ");
        InitLevelRow(9, 12, "                    ");
        InitLevelRow(9, 13, "6                   ");
        InitLevelRow(9, 14, "                    ");
        InitLevelRow(9, 15, "                    ");
        InitLevelRow(9, 16, "                    ");
        InitLevelRow(9, 17, "                    ");
        InitLevelRow(9, 18, "                    ");
        InitLevelRow(9, 19, "                    ");
    }
    else if(g_iLevelNr == 7)
    {
        g_iLevelWidth           = 15;
        g_iLevelHeight          = 10;
        g_iLevelDepth           = 15;
        g_iLevelDiamonds        = 1;
        g_bRenderShining        = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(0, 0, 0);
        g_fLevelExitRotation    = D3DX_PI / 2.0f;
        strcpy(g_sLevelName,      "B  e  e  i  l  u  n  g");

        InitLevelRow(0,  0, "               ");
        InitLevelRow(0,  1, "               ");
        InitLevelRow(0,  2, "               ");
        InitLevelRow(0,  3, "   ............");
        InitLevelRow(0,  4, "   ............");
        InitLevelRow(0,  5, "   ............");
        InitLevelRow(0,  6, "   ............");
        InitLevelRow(0,  7, "   ............");
        InitLevelRow(0,  8, "   ............");
        InitLevelRow(0,  9, "   ............");
        InitLevelRow(0, 10, "   ............");
        InitLevelRow(0, 11, "   ............");
        InitLevelRow(0, 12, "   ............");
        InitLevelRow(0, 13, "   ............");
        InitLevelRow(0, 14, "   ............");

        InitLevelRow(1,  0, "              m");
        InitLevelRow(1,  1, "              m");
        InitLevelRow(1,  2, "  mmmmmmmmmmmmm");
        InitLevelRow(1,  3, "  m            ");
        InitLevelRow(1,  4, "  m            ");
        InitLevelRow(1,  5, "  mmm          ");
        InitLevelRow(1,  6, "  m            ");
        InitLevelRow(1,  7, "  m            ");
        InitLevelRow(1,  8, "  m            ");
        InitLevelRow(1,  9, "  m            ");
        InitLevelRow(1, 10, "  m       mmmmm");
        InitLevelRow(1, 11, "  m       m  m ");
        InitLevelRow(1, 12, "  m       m  ? ");
        InitLevelRow(1, 13, "  m       m  m ");
        InitLevelRow(1, 14, "mmm       m    ");

        InitLevelRow(2,  0, "..............m");
        InitLevelRow(2,  1, "..............m");
        InitLevelRow(2,  2, "..mmmmmmmmmmmmm");
        InitLevelRow(2,  3, "..m            ");
        InitLevelRow(2,  4, "..m            ");
        InitLevelRow(2,  5, "..mmm          ");
        InitLevelRow(2,  6, "..m            ");
        InitLevelRow(2,  7, "..m            ");
        InitLevelRow(2,  8, "..m            ");
        InitLevelRow(2,  9, "..m            ");
        InitLevelRow(2, 10, "..m            ");
        InitLevelRow(2, 11, "..m            ");
        InitLevelRow(2, 12, "..m            ");
        InitLevelRow(2, 13, "..m          m ");
        InitLevelRow(2, 14, "mmm            ");

        InitLevelRow(3,  0, "               ");
        InitLevelRow(3,  1, "             s ");
        InitLevelRow(3,  2, "               ");
        InitLevelRow(3,  3, "               ");
        InitLevelRow(3,  4, "               ");
        InitLevelRow(3,  5, "   mm          ");
        InitLevelRow(3,  6, "               ");
        InitLevelRow(3,  7, "               ");
        InitLevelRow(3,  8, "               ");
        InitLevelRow(3,  9, "               ");
        InitLevelRow(3, 10, "               ");
        InitLevelRow(3, 11, "               ");
        InitLevelRow(3, 12, "               ");
        InitLevelRow(3, 13, " !           m ");
        InitLevelRow(3, 14, "               ");

        InitLevelRow(4,  0, "               ");
        InitLevelRow(4,  1, "               ");
        InitLevelRow(4,  2, "               ");
        InitLevelRow(4,  3, "               ");
        InitLevelRow(4,  4, "               ");
        InitLevelRow(4,  5, "   mm          ");
        InitLevelRow(4,  6, "               ");
        InitLevelRow(4,  7, "               ");
        InitLevelRow(4,  8, "               ");
        InitLevelRow(4,  9, "               ");
        InitLevelRow(4, 10, "               ");
        InitLevelRow(4, 11, "               ");
        InitLevelRow(4, 12, "               ");
        InitLevelRow(4, 13, "               ");
        InitLevelRow(4, 14, "               ");

        InitLevelRow(5,  0, "               ");
        InitLevelRow(5,  1, "               ");
        InitLevelRow(5,  2, "               ");
        InitLevelRow(5,  3, "               ");
        InitLevelRow(5,  4, "               ");
        InitLevelRow(5,  5, "   0           ");
        InitLevelRow(5,  6, "               ");
        InitLevelRow(5,  7, "               ");
        InitLevelRow(5,  8, "               ");
        InitLevelRow(5,  9, "               ");
        InitLevelRow(5, 10, "               ");
        InitLevelRow(5, 11, "               ");
        InitLevelRow(5, 12, "               ");
        InitLevelRow(5, 13, "               ");
        InitLevelRow(5, 14, "               ");

        InitLevelRow(6,  0, "               ");
        InitLevelRow(6,  1, "               ");
        InitLevelRow(6,  2, "               ");
        InitLevelRow(6,  3, "               ");
        InitLevelRow(6,  4, "               ");
        InitLevelRow(6,  5, "               ");
        InitLevelRow(6,  6, "               ");
        InitLevelRow(6,  7, "               ");
        InitLevelRow(6,  8, "               ");
        InitLevelRow(6,  9, "               ");
        InitLevelRow(6, 10, "               ");
        InitLevelRow(6, 11, "               ");
        InitLevelRow(6, 12, "               ");
        InitLevelRow(6, 13, "             1 ");
        InitLevelRow(6, 14, "               ");
    }
    else if(g_iLevelNr == 8)
    {
        g_iLevelWidth           = 20;
        g_iLevelHeight          = 15;
        g_iLevelDepth           = 20;
        g_iLevelDiamonds        = 3;
        g_bRenderShining        = TRUE;
        g_cBackgroundColor      = D3DCOLOR_XRGB(0, 0, 0);
        g_bObjectsOutOfMap      = FALSE;
        g_fGroundHeight         = 0.0f;
        g_fLightFlickering      = 0.25f;
        g_cLightColor           = D3DRGB(1.0f, 1.0f, 0.0f);
        g_vCamera               = D3DXVECTOR3(0.5f, 2.0f, -0.5f);
        strcpy(g_sLevelName,      "H o c h   h i n a u s");

        InitLevelRow(0,  0, "                    ");
        InitLevelRow(0,  1, " f                f ");
        InitLevelRow(0,  2, "                    ");
        InitLevelRow(0,  3, "    mmm           t ");
        InitLevelRow(0,  4, "    mtm             ");
        InitLevelRow(0,  5, "    mmm             ");
        InitLevelRow(0,  6, "                    ");
        InitLevelRow(0,  7, "                    ");
        InitLevelRow(0,  8, "                  m ");
        InitLevelRow(0,  9, "                    ");
        InitLevelRow(0, 10, "          t         ");
        InitLevelRow(0, 11, "                    ");
        InitLevelRow(0, 12, "                    ");
        InitLevelRow(0, 13, "          !       m ");
        InitLevelRow(0, 14, "                    ");
        InitLevelRow(0, 15, "                    ");
        InitLevelRow(0, 16, "                    ");
        InitLevelRow(0, 17, "       mmmm         ");
        InitLevelRow(0, 18, " f  s  mppmmt     f ");
        InitLevelRow(0, 19, "       mppm         ");

        InitLevelRow(1,  0, "                    ");
        InitLevelRow(1,  1, " 6                6 ");
        InitLevelRow(1,  2, "                    ");
        InitLevelRow(1,  3, "    mmm             ");
        InitLevelRow(1,  4, "    m m             ");
        InitLevelRow(1,  5, "    mmm             ");
        InitLevelRow(1,  6, "                    ");
        InitLevelRow(1,  7, "                    ");
        InitLevelRow(1,  8, "                  m ");
        InitLevelRow(1,  9, "                    ");
        InitLevelRow(1, 10, "                    ");
        InitLevelRow(1, 11, "                    ");
        InitLevelRow(1, 12, "                    ");
        InitLevelRow(1, 13, "                  m ");
        InitLevelRow(1, 14, "                    ");
        InitLevelRow(1, 15, "                    ");
        InitLevelRow(1, 16, "                    ");
        InitLevelRow(1, 17, "       mmmm         ");
        InitLevelRow(1, 18, " 6     mttmm      6 ");
        InitLevelRow(1, 19, "       m  m         ");

        InitLevelRow(2,  0, "                    ");
        InitLevelRow(2,  1, "                    ");
        InitLevelRow(2,  2, "                    ");
        InitLevelRow(2,  3, "    mmm             ");
        InitLevelRow(2,  4, "    m1m             ");
        InitLevelRow(2,  5, "    m m             ");
        InitLevelRow(2,  6, "                    ");
        InitLevelRow(2,  7, "                    ");
        InitLevelRow(2,  8, "                  m ");
        InitLevelRow(2,  9, "                    ");
        InitLevelRow(2, 10, "                    ");
        InitLevelRow(2, 11, "                    ");
        InitLevelRow(2, 12, "                    ");
        InitLevelRow(2, 13, "                  m ");
        InitLevelRow(2, 14, "                    ");
        InitLevelRow(2, 15, "                    ");
        InitLevelRow(2, 16, "                    ");
        InitLevelRow(2, 17, "       mmmm         ");
        InitLevelRow(2, 18, "    p  m  mm        ");
        InitLevelRow(2, 19, "       m  m         ");

        InitLevelRow(3,  0, "                    ");
        InitLevelRow(3,  1, "                    ");
        InitLevelRow(3,  2, "                    ");
        InitLevelRow(3,  3, "    mmm             ");
        InitLevelRow(3,  4, "    m m             ");
        InitLevelRow(3,  5, "    mmm             ");
        InitLevelRow(3,  6, "                    ");
        InitLevelRow(3,  7, "                    ");
        InitLevelRow(3,  8, "                  m ");
        InitLevelRow(3,  9, "                    ");
        InitLevelRow(3, 10, "                    ");
        InitLevelRow(3, 11, "                    ");
        InitLevelRow(3, 12, "                    ");
        InitLevelRow(3, 13, "                  m ");
        InitLevelRow(3, 14, "                    ");
        InitLevelRow(3, 15, "                    ");
        InitLevelRow(3, 16, "                    ");
        InitLevelRow(3, 17, "                    ");
        InitLevelRow(3, 18, "          0         ");
        InitLevelRow(3, 19, "                    ");

        InitLevelRow(4,  0, "pp                  ");
        InitLevelRow(4,  1, "pp                  ");
        InitLevelRow(4,  2, "pp                  ");
        InitLevelRow(4,  3, "    mmm             ");
        InitLevelRow(4,  4, "    m m             ");
        InitLevelRow(4,  5, "    mmm             ");
        InitLevelRow(4,  6, "                    ");
        InitLevelRow(4,  7, "                    ");
        InitLevelRow(4,  8, "                  t ");
        InitLevelRow(4,  9, "                    ");
        InitLevelRow(4, 10, "                    ");
        InitLevelRow(4, 11, "                    ");
        InitLevelRow(4, 12, "                    ");
        InitLevelRow(4, 13, "                  m ");
        InitLevelRow(4, 14, "                    ");
        InitLevelRow(4, 15, "                    ");
        InitLevelRow(4, 16, "                    ");
        InitLevelRow(4, 17, "                    ");
        InitLevelRow(4, 18, "     ppp  pppp      ");
        InitLevelRow(4, 19, "     ppp  pppp      ");

        InitLevelRow(5,  0, " m                  ");
        InitLevelRow(5,  1, "?m                  ");
        InitLevelRow(5,  2, "                    ");
        InitLevelRow(5,  3, "      s             ");
        InitLevelRow(5,  4, "                    ");
        InitLevelRow(5,  5, "                    ");
        InitLevelRow(5,  6, "                    ");
        InitLevelRow(5,  7, "                    ");
        InitLevelRow(5,  8, "                    ");
        InitLevelRow(5,  9, "  m                 ");
        InitLevelRow(5, 10, " 1m       1         ");
        InitLevelRow(5, 11, " mm                 ");
        InitLevelRow(5, 12, "                    ");
        InitLevelRow(5, 13, "                  m ");
        InitLevelRow(5, 14, "                    ");
        InitLevelRow(5, 15, "                    ");
        InitLevelRow(5, 16, "                    ");
        InitLevelRow(5, 17, "                    ");
        InitLevelRow(5, 18, "                    ");
        InitLevelRow(5, 19, "             s      ");

        InitLevelRow(6,  0, "                    ");
        InitLevelRow(6,  1, "                    ");
        InitLevelRow(6,  2, "                    ");
        InitLevelRow(6,  3, "                    ");
        InitLevelRow(6,  4, "                    ");
        InitLevelRow(6,  5, "                    ");
        InitLevelRow(6,  6, "                    ");
        InitLevelRow(6,  7, "                    ");
        InitLevelRow(6,  8, "                    ");
        InitLevelRow(6,  9, "ppm                 ");
        InitLevelRow(6, 10, "pmm                 ");
        InitLevelRow(6, 11, "pmm                 ");
        InitLevelRow(6, 12, "        ppp         ");
        InitLevelRow(6, 13, "        ppp       m ");
        InitLevelRow(6, 14, "        ppp         ");
        InitLevelRow(6, 15, "                    ");
        InitLevelRow(6, 16, "                    ");
        InitLevelRow(6, 17, "                    ");
        InitLevelRow(6, 18, "                    ");
        InitLevelRow(6, 19, "                    ");

        InitLevelRow(7,  0, "                    ");
        InitLevelRow(7,  1, "                    ");
        InitLevelRow(7,  2, "                    ");
        InitLevelRow(7,  3, "                    ");
        InitLevelRow(7,  4, "                    ");
        InitLevelRow(7,  5, "                    ");
        InitLevelRow(7,  6, "                    ");
        InitLevelRow(7,  7, "                    ");
        InitLevelRow(7,  8, "                    ");
        InitLevelRow(7,  9, "                    ");
        InitLevelRow(7, 10, "                    ");
        InitLevelRow(7, 11, "                    ");
        InitLevelRow(7, 12, "                    ");
        InitLevelRow(7, 13, "         t        m ");
        InitLevelRow(7, 14, "                    ");
        InitLevelRow(7, 15, "                    ");
        InitLevelRow(7, 16, "                    ");
        InitLevelRow(7, 17, "                    ");
        InitLevelRow(7, 18, "                    ");
        InitLevelRow(7, 19, "                    ");

        InitLevelRow(8,  0, "                    ");
        InitLevelRow(8,  1, "                    ");
        InitLevelRow(8,  2, "                    ");
        InitLevelRow(8,  3, "                    ");
        InitLevelRow(8,  4, "                    ");
        InitLevelRow(8,  5, "                    ");
        InitLevelRow(8,  6, "                    ");
        InitLevelRow(8,  7, "                    ");
        InitLevelRow(8,  8, "                    ");
        InitLevelRow(8,  9, "                    ");
        InitLevelRow(8, 10, "                    ");
        InitLevelRow(8, 11, "                    ");
        InitLevelRow(8, 12, "                    ");
        InitLevelRow(8, 13, "                  t ");
        InitLevelRow(8, 14, "                    ");
        InitLevelRow(8, 15, "                    ");
        InitLevelRow(8, 16, "                    ");
        InitLevelRow(8, 17, "                    ");
        InitLevelRow(8, 18, "                    ");
        InitLevelRow(8, 19, "                    ");
    }

    InitPositions();
    InitFieldFlagsArray();
    InitObjects();
    if(g_bSnowing) InitSnow();

    g_vCamera = D3DXVECTOR3((g_iLevelWidth / 2) + 0.5f, 50.0f, -(g_iLevelDepth / 2) + 0.5f);
    g_vCameraDestination = g_vCamera;
    g_vLookAt = g_vCamera;
    g_vLookAt.y = 0.0f;

    g_pD3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);
    g_pD3DDevice->SetRenderState(D3DRS_FOGCOLOR, g_cBackgroundColor);
    g_pD3DDevice->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_LINEAR);
    g_pD3DDevice->SetRenderState(D3DRS_FOGSTART, FloatToDWord(fFogStart));
    g_pD3DDevice->SetRenderState(D3DRS_FOGEND, FloatToDWord(fFogEnd));
    
    D3DXMatrixPerspectiveFovLH(&g_mProjection, D3DX_PI / 4.0f, 800.0f / 600.0f, 0.01f, fFogEnd);
    g_pD3DDevice->SetTransform(D3DTS_PROJECTION, (D3DMATRIX*)(&g_mProjection));

    if(g_pLevelGround != NULL)
    {
        g_pLevelGround->Release();
    }
    sprintf(sFilename, "Data\\LevelGround%d.x", g_iLevelNr);
    if(iNumGroundTextures == 1) MeshCreate(&g_pLevelGround, sFilename, g_pD3DDevice, TRUE, TRUE, TRUE);
    else MeshCreate(&g_pLevelGround, sFilename, g_pD3DDevice);
    for(int a = 0; a < iNumGroundTextures; a++)
    {
        sprintf(sFilename, "Data\\LevelGround%d_%d.bmp", g_iLevelNr, a);
        D3DXCreateTextureFromFile(g_pD3DDevice, sFilename, &g_pLevelGround->m_pTexture[a]);
    }

    if(g_iLevelNr == 0)
    {
        for(a = 0; a < MAX_OBJECTS; a++)
        {
            if(g_Object[a].m_bExists)
            {
                g_Object[a].m_Control.m_vPosition.y = 15.0f + FloatRandom(0.0f, 5.0f);
            }
        }

        g_vCamera.y = 0.0f;
        g_vCameraDestination.y = 0.0f;
        g_vLookAt.y = 0.0f;
    }
    
    if(g_iLevelNr == 6)
    {
        g_pLevelGround->m_pTexture[2] = g_pLevelGround->m_pTexture[3] = g_pLevelGround->m_pTexture[4] = g_pLevelGround->m_pTexture[1]; g_pLevelGround->m_pTexture[5] = g_pLevelGround->m_pTexture[0];
    }
}

void UpdatePlayerField()
{
    g_Player.m_iFX = (int)(g_Player.m_Control.m_vPosition.x);
    g_Player.m_iFY = (int)(g_Player.m_Control.m_vPosition.y);
    g_Player.m_iFZ = (int)(-g_Player.m_Control.m_vPosition.z);
}

BOOL MakeMove(D3DXVECTOR3 vDirection)
{
    D3DXVECTOR3 vAdd;
    int         iCheck;

    if(vDirection.y > 0.0f) vAdd = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
    else vAdd = D3DXVECTOR3(0.0f, 0.0f, 0.0f);

    if(vDirection.y < 0.0f) iCheck = 1;
    else if(vDirection.y > 0.0f) iCheck = 1;
    else if(vDirection.z < 0.0f) iCheck = 3;
    else iCheck = 4;

    if(IsFreeForPlayer(vDirection + vAdd, FALSE, iCheck))
    {
        if(vDirection.y >= 0.0f) g_Player.m_Control.Translate(vDirection);
        else {g_Player.m_Control.m_vPosition += vDirection; g_Player.m_Control.UpdateTransformationMatrix();}
        return TRUE;
    }
    else
    {
        while(IsFreeForPlayer(vDirection / 10.0f + vAdd, FALSE, iCheck))
        {
            if(vDirection.y >= 0.0f) g_Player.m_Control.Translate(vDirection / 10.0f);
            else {g_Player.m_Control.m_vPosition += vDirection / 10.0f; g_Player.m_Control.UpdateTransformationMatrix();}
        }
        return FALSE;
    }
}

void ShowLevelExit()
{
    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(!g_Object[a].m_bExists)
        {
            ZeroMemory(&g_Object[a], sizeof(g_Object[a]));

            g_Object[a].m_bExists = TRUE;
            g_Object[a].m_iType = OT_LEVEL_EXIT;
            g_Object[a].m_pType = &g_ObjectType[OT_LEVEL_EXIT];
            g_Object[a].m_Control.Clear();
            g_Object[a].m_Control.SetPosition(g_vLevelExit);
            g_Object[a].m_Control.RotateY(g_fLevelExitRotation);

            g_pMusic->Stop(0);
            PlaySound(S_LEVEL_EXIT_SIGNAL, TRUE, TRUE, g_Object[a].m_Control.m_vPosition, D3DXVECTOR3(0.0f, 0.0f, 0.0f), 10.0f);

            return;
        }
    }
}

void AddEnergy(float fEnergy)
{
    if(g_Player.m_fImmortality <= 0.0f)
    {
        g_Player.m_fEnergy += fEnergy;

        if(fEnergy < 0.0f)
        {
            PlaySound(S_LOOSE_ENERGY1 + (rand() % 2), FALSE, TRUE, g_Player.m_Control.m_vPosition);
            g_Player.m_fImmortality = PLAYER_IMMORTALITY;
        }
    }
}

BOOL Switch(int x,
            int y,
            int z,
            BOOL bBlock)
{
    BOOL bSwitch;

    bSwitch = FALSE;

    if(g_iLevelNr == 6)
    {
        if(x == 18 && y == 7 && z == 12)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(7.0f, 8.5f, -16.0f), D3DXVECTOR3(2.0f, 7.5f, -16.0f), 150.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(1, 7, 15, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(2, 7, 15, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(1.5f, 8.0f, -15.5f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(1, 7, 16, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(2, 7, 16, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(1.5f, 8.0f, -16.5f));
            bSwitch = TRUE;
        }
        else if(x == 7 && y == 2 && z == 18)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(11.5f, 9.5f, -10.0f), D3DXVECTOR3(11.5f, 7.5f, -18.5f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(10, 7, 18, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(11, 7, 18, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(12, 7, 18, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(11.5f, 8.0f, -18.5f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(5.0f, 6.5f, -5.0f), D3DXVECTOR3(0.0f, 4.5f, 0.0f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(0, 4, 0, 255);
            g_Script.AddSetFieldCommand(1, 4, 0, 255);
            g_Script.AddSetFieldCommand(0, 4, 1, 255);
            g_Script.AddSetFieldCommand(1, 4, 1, 255);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(1.0f, 4.5f, -1.0f));
            bSwitch = TRUE;
        }
        else if(x == 18 && y == 1 && z == 18)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(13.5f, 10.5f, -16.0f), D3DXVECTOR3(11.5f, 8.0f, -5.0f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(11, 7, 8, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(11, 7, 9, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(11.5f, 8.0f, -9.0f));
            bSwitch = TRUE;
        }
    }
    else if(g_iLevelNr == 7)
    {
        if(x == 13 && y == 3 && z == 1)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(13.5f, 6.0f, -1.5f), D3DXVECTOR3(1.5f, 4.0f, -5.5f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(0, 3, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(1, 3, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(2, 3, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(0, 4, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(1, 4, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(2, 4, 5, g_iSymbolTable['m']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(1.5f, 4.0f, -5.5f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(7.5f, 25.0f, -7.5f), D3DXVECTOR3(7.5f, 0.0f, -7.5f), 130.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(13, 3, 5, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(13.5f, 2.5f, -5.5f));
            g_Script.AddDelayCommand(10.0f);
            g_Script.AddSetFieldCommand(13, 3, 8, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(13.5f, 2.5f, -8.5f));
            g_Script.AddDelayCommand(10.0f);
            g_Script.AddSetFieldCommand(10, 3, 11, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(10.5f, 2.5f, -11.5f));
            g_Script.AddDelayCommand(10.0f);
            g_Script.AddSetFieldCommand(7, 3, 11, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(7.5f, 2.5f, -11.5f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(3.5f, 10.0f, -20.0f), D3DXVECTOR3(1.5f, 3.5f, -13.5f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(1, 3, 13, g_iSymbolTable['s']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(1.5f, 3.5f, -13.5f));
            g_Script.AddDelayCommand(500.0f);
            g_Script.AddSetFieldCommand(13, 3, 5, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(13.5f, 2.5f, -5.5f));
            g_Script.AddDelayCommand(10.0f);
            g_Script.AddSetFieldCommand(13, 3, 8, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(13.5f, 2.5f, -8.5f));
            g_Script.AddDelayCommand(10.0f);
            g_Script.AddSetFieldCommand(10, 3, 11, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(10.5f, 2.5f, -11.5f));
            g_Script.AddDelayCommand(10.0f);
            g_Script.AddSetFieldCommand(7, 3, 11, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(7.5f, 2.5f, -11.5f));
            bSwitch = TRUE;
        }
        else if(x == 1 && y == 3 && z == 13)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(0.5f, 10.0f, -10.0f), D3DXVECTOR3(2.5f, 3.5f, -6.5f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(2, 3, 6, g_iSymbolTable['m']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(2.5f, 3.5f, -6.5f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(0.5f, 15.0f, -0.5f), D3DXVECTOR3(10.5f, 1.5f, -6.0f), 200.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(4, 1, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(5, 1, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(6, 1, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(7, 1, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(8, 1, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(9, 1, 5, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(10, 1, 5, g_iSymbolTable['m']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(7.5f, 1.5f, -5.5f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(10, 1, 6, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(10, 1, 7, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(10, 1, 8, g_iSymbolTable['m']);
            g_Script.AddSetFieldCommand(10, 1, 9, g_iSymbolTable['m']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(10.5f, 1.5f, -8.0f));
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(9, 1, 6, g_iSymbolTable['m']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(9.5f, 1.5f, -6.5f));
            g_Script.AddDelayCommand(500.0f);
            g_Script.AddSetFieldCommand(4, 2, 5, g_iSymbolTable[' ']);
            g_Script.AddSetFieldCommand(4, 3, 5, g_iSymbolTable[' ']);
            g_Script.AddSetFieldCommand(4, 4, 5, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(4.5f, 3.0f, -5.5f));
            g_Script.AddDelayCommand(1000.0f);
            g_Script.AddSetFieldCommand(13, 2, 13, g_iSymbolTable[' ']);
            g_Script.AddSetFieldCommand(13, 3, 13, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(13.5f, 2.5f, -13.5f));
            bSwitch = TRUE;
        }
    }
    else if(g_iLevelNr == 8)
    {
        if(x == 13 && y == 5 && z == 19)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetFieldCommand(8, 4, 19, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(9, 4, 19, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(8, 4, 18, g_iSymbolTable['p']);
            g_Script.AddSetFieldCommand(9, 4, 18, g_iSymbolTable['p']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(9.0f, 4.5f, -19.0f));
            g_Script.AddDelayCommand(200.0f);
            g_Script.AddSetFieldCommand(8, 4, 19, g_iSymbolTable[' ']);
            g_Script.AddSetFieldCommand(9, 4, 19, g_iSymbolTable[' ']);
            g_Script.AddSetFieldCommand(8, 4, 18, g_iSymbolTable[' ']);
            g_Script.AddSetFieldCommand(9, 4, 18, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(9.0f, 4.5f, -19.0f));
            g_Script.AddSetFieldCommand(13, 5, 19, g_iSymbolTable['s']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(13.5f, 5.5f, -19.5f));
            bSwitch = TRUE;
        }
        else if(x == 6 && y == 5 && z == 3)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(6.5f, 10.0f, -3.5f), D3DXVECTOR3(4.5f, 2.5f, -18.5f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(4, 2, 18, g_iSymbolTable[' ']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(4.5f, 2.5f, -18.5f));
            bSwitch = TRUE;
        }
        else if(x == 4 && y == 0 && z == 18 && bBlock)
        {
            g_Script.AddDelayCommand(30.0f);
            g_Script.AddSetCameraCommand(D3DXVECTOR3(5.5f, 10.0f, -18.5f), D3DXVECTOR3(1.5f, 0.5f, -7.5f), 100.0f);
            g_Script.AddDelayCommand(50.0f);
            g_Script.AddSetFieldCommand(1, 0, 7, g_iSymbolTable['t']);
            g_Script.AddPlaySoundCommand(S_FIELD_CHANGES, D3DXVECTOR3(1.5f, 0.5f, -14.5f));
            bSwitch = TRUE;
        }
    }

    if(bSwitch)
    {
        PlaySound(S_SWITCH, FALSE, TRUE, D3DXVECTOR3((float)(x) + 0.5f, (float)(y) + 0.25f, (float)(-z - 0.5f)));
        g_iLevel[x][y][z] = 10;
        InitFieldFlagsArray();
    }

    return bSwitch;
}

void CheckForSwitches(int x,
                      int y,
                      int z,
                      BOOL bBlock)
{
    CFieldType FieldType;

    FieldType = GetFieldType(x, y, z);
    if(FieldType.m_iType == 9)
    {
        Switch(x, y, z, bBlock);
    }
}

void MovePlayer(float fUnit)
{
    CFieldType  FieldType;
    int         iFlags;
    int         iObject;
    int         iStep;
    int         iOldStep;
    int         x;
    int         y;
    int         z;
    float       fSpeedLowering;
    float       fSpeedFactor;
    float       fOldStep;
    BOOL        bWalk;
    BOOL        bRotate;
    D3DXVECTOR3 vDirection;
	JOYINFO		JoystickInfo;
	D3DXVECTOR2	vJoystick;

    fSpeedLowering = SPEED_LOWERING;
    fOldStep = g_Player.m_fStep;

    UpdateKeyStates();
	if(g_bUseJoystick)
	{
		joyGetPos(JOYSTICKID1, &JoystickInfo);
		vJoystick.x = ((float)(JoystickInfo.wXpos - g_JoyCaps.wXmin) / (float)(g_JoyCaps.wXmax) - 0.5f) * 2.0f;
		vJoystick.y = ((float)(JoystickInfo.wYpos - g_JoyCaps.wYmin) / (float)(g_JoyCaps.wYmax) - 0.5f) * 2.0f;

		if(vJoystick.x <= -0.3f) g_cKeyState[DIK_LEFT] = TRUE;
		if(vJoystick.x >= 0.3f) g_cKeyState[DIK_RIGHT] = TRUE;
		if(vJoystick.y <= -0.4f) g_cKeyState[DIK_UP] = TRUE;
		if(vJoystick.y >= 0.4f) g_cKeyState[DIK_DOWN] = TRUE;
		if(JoystickInfo.wButtons & JOY_BUTTON1) g_cKeyState[DIK_LSHIFT] = TRUE;
		if(JoystickInfo.wButtons & JOY_BUTTON2) g_cKeyState[DIK_LCONTROL] = TRUE;
	}
	else vJoystick = D3DXVECTOR2(0.0f, 0.0f);

    if(g_cKeyState[DIK_ESCAPE] &&
       !g_bMenu &&
       !g_bConfirmQuit &&
       !g_bConfirmRestart &&
       g_fConfirmQuitCountDown <= 0.0f)
    {
        g_bConfirmQuit = TRUE;
        g_bConfirmRestart = FALSE;
        g_fConfirmQuitCountDown = CONFIRM_QUIT_DELAY;
    }

    if(g_cKeyState[DIK_N] &&
       !g_bMenu)
    {
        g_bConfirmRestart = TRUE;
        g_bConfirmQuit = FALSE;
    }

    if(g_cKeyState[DIK_F1] &&
       g_iViewMode != 0)
    {
        g_iViewMode = 0;
    }
    if(g_cKeyState[DIK_F2] &&
       g_Player.m_bOnGround &&
       g_iViewMode != 1)
    {
        g_iViewMode = 1;
        g_View1Camera.Clear();
        g_View1Camera.m_vAxis[0] = g_Player.m_Control.m_vAxis[0];
        g_View1Camera.m_vAxis[2] = g_Player.m_Control.m_vAxis[2];
    }

    if(g_Player.m_fEnergy <= 0.0f) g_iViewMode = 0;

    if(g_iViewMode == 1) g_View1Camera.SetPosition(g_Player.m_Control.GetRelativePosition(D3DXVECTOR3(0.0f, 1.0f, 0.0f)));
    
    bRotate = FALSE;
    if(g_iViewMode == 0)
    {
        if(g_Player.m_fEnergy <= 0.0f) goto FallDown;

        if(g_cKeyState[DIK_LEFT])
        {
            bRotate = TRUE;
            g_Player.m_Control.RotateY(-0.075f * fUnit);
        }
        if(g_cKeyState[DIK_RIGHT])
        {
            bRotate = TRUE;
            g_Player.m_Control.RotateY(0.075f * fUnit);
        }

        bWalk = FALSE;
        if(g_cKeyState[DIK_LCONTROL] &&
           g_Player.m_bOnGround)
        {
            g_Player.m_bPush = TRUE;
            fSpeedFactor = 0.5f;
        }
        else
        {
            g_Player.m_bPush = FALSE;
            fSpeedFactor = 1.0f;
        }
        
        if(g_cKeyState[DIK_UP])
        {
            bWalk = TRUE;

            if(!MakeMove(D3DXVECTOR3(0.0f, 0.0f, g_Player.m_fSpeed * fSpeedFactor * fUnit)))
            {
                if(g_Player.m_bPush)
                {
                    iObject = GetObject(g_Player.m_Control.GetRelativePosition(D3DXVECTOR3(0.0f, 0.0f, 1.0f)));
                    if(iObject != -1)
                    {
                        if(g_Object[iObject].m_pType->m_iFlags & OF_PUSHABLE)
                        {
                            vDirection = g_Player.m_Control.m_vAxis[2] * (g_Player.m_fSpeed * fSpeedFactor);
                            if(ObjectFree(g_Object[iObject], g_Object[iObject].m_Control.m_vPosition + vDirection))
                            {
                                g_Object[iObject].m_Control.m_vPosition += vDirection * fUnit;
                                if(g_Object[iObject].m_bOnGround) PlaySound(S_PUSH_OBJECT, FALSE, FALSE, g_Object[iObject].m_Control.m_vPosition);

                                while(TRUE)
                                {
                                    iObject = GetObject(g_Object[iObject].m_Control.m_vPosition + D3DXVECTOR3(0.0f, 1.0f, 0.0f));
                                    if(iObject != -1)
                                    {
                                        if(g_Object[iObject].m_pType->m_iFlags & OF_PUSHABLE)
                                        {
                                            if(ObjectFree(g_Object[iObject], g_Object[iObject].m_Control.m_vPosition + vDirection))
                                            {
                                                g_Object[iObject].m_Control.m_vPosition += vDirection * fUnit;
                                            }
                                            else goto EndObjectTowerCheck;
                                        }
                                    }
                                    else goto EndObjectTowerCheck;
                                }
                            }

EndObjectTowerCheck:
                            MakeMove(D3DXVECTOR3(0.0f, 0.0f, g_Player.m_fSpeed * fSpeedFactor * fUnit));
                        }
                    }
                }
            }

            g_Player.m_fSpeed += ACCELERATION * fUnit;
        }

        if(g_cKeyState[DIK_DOWN])
		{
			bWalk = TRUE;
			MakeMove(D3DXVECTOR3(0.0f, 0.0f, -g_Player.m_fSpeed / 2.0f * fUnit));
			g_Player.m_fSpeed += ACCELERATION * fUnit;
			g_Player.m_bPush = FALSE;
		}

        if(!g_cKeyState[DIK_UP] &&
           !g_cKeyState[DIK_DOWN])
        {
            g_Player.m_fStep = 2.0f;
        }

        if(bWalk)
        {
            if(g_Player.m_bPush)
            {
                g_Player.m_fStep += PLAYER_PUSH_STEP_SPEED * fUnit;
                if((int)(g_Player.m_fStep) >= NUM_PLAYER_PUSH_STEPS) g_Player.m_fStep = 0.0f;
            }
            else
            {
                g_Player.m_fStep += PLAYER_WALK_STEP_SPEED * fUnit;
                if((int)(g_Player.m_fStep) >= NUM_PLAYER_WALK_STEPS) g_Player.m_fStep = 0.0f;
            }        
        }  

        if(bRotate && !bWalk)
        {
            if((int)(g_Player.m_fStep) != 2 &&
               (int)(g_Player.m_fStep) != 6)
            {
                g_Player.m_fStep += PLAYER_WALK_STEP_SPEED * fUnit;
                if((int)(g_Player.m_fStep) >= NUM_PLAYER_WALK_STEPS) g_Player.m_fStep = 0.0f;
            }
        }

FallDown:

        if(g_Player.m_fEnergy <= 0.0f)
        {
            if(g_fDieCountDown > DIE_DELAY * 0.9f) g_Player.m_Control.RotateX(0.05f * fUnit);
            else if(g_fDieCountDown > DIE_DELAY * 0.75f) g_Player.m_Control.RotateX(-0.1f * fUnit);
            else g_vCameraDestination = g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 50.0f, 0.0f);

            g_fDieCountDown -= fUnit;
        }

        if(g_cKeyState[DIK_DELETE]) g_fCameraAngle += 0.05f * fUnit;
        if(g_cKeyState[DIK_END]) g_fCameraAngle -= 0.05f * fUnit;
        if(g_cKeyState[DIK_INSERT]) g_fCameraAngle = 0.0f;
        if(g_cKeyState[DIK_HOME] && g_fLockCameraCountDown <= 0.0f)
        {
            g_bCameraLocked = !g_bCameraLocked;
            if(g_bCameraLocked == FALSE)
            {
                g_fCameraDistance = STANDARD_CAMERA_DISTANCE;
                g_fCameraAngle = 0.0f;         
            }

            g_fLockCameraCountDown = CAMERA_LOCK_DELAY;
        }
        if(g_cKeyState[DIK_NUMPADPLUS])
        {
            g_fCameraDistance -= 0.2f * fUnit;
            if(g_fCameraDistance < MIN_CAMERA_DISTANCE) g_fCameraDistance = MIN_CAMERA_DISTANCE;
        }
        if(g_cKeyState[DIK_NUMPADMINUS]) g_fCameraDistance += 0.2f * fUnit;
        if(g_cKeyState[DIK_PGUP]) g_fCameraHeight += 0.2f * fUnit;
        if(g_cKeyState[DIK_PGDN]) g_fCameraHeight -= 0.2f * fUnit;
    }

    if(g_iViewMode == 1)
    {
        if(g_cKeyState[DIK_LEFT])
        {
            g_Player.m_Control.RotateY(-0.03f * fUnit);
            g_View1Camera.RotateY(-0.03f * fUnit);
        }
        if(g_cKeyState[DIK_RIGHT])
        {
            g_Player.m_Control.RotateY(0.03f * fUnit);
            g_View1Camera.RotateY(0.03f * fUnit);
        }

        if(g_cKeyState[DIK_UP]) g_View1Camera.m_vAxis[2].y -= 0.02f * fUnit;
        if(g_cKeyState[DIK_DOWN]) g_View1Camera.m_vAxis[2].y += 0.02f * fUnit;

        if(g_View1Camera.m_vAxis[2].y < -2.0f) g_View1Camera.m_vAxis[2].y = -2.0f * fUnit;
        if(g_View1Camera.m_vAxis[2].y > 2.0f) g_View1Camera.m_vAxis[2].y = 2.0f * fUnit;
    }
   
    if(g_Player.m_fSpeed > MAX_SPEED) g_Player.m_fSpeed = MAX_SPEED;

    if(g_iViewMode == 0)
    {
        if(g_cKeyState[DIK_LSHIFT])
        {
            if(g_Player.m_fFalling >= 0.0f &&
               g_Player.m_bOnGround &&
               g_Player.m_fJumpCountDown <= 0.0f &&
               IsFreeForPlayer(D3DXVECTOR3(0.0f, 1.1f, 0.0f), FALSE, 1))
            {
                g_Player.m_fFalling = -JUMP_SPEED;
                PlaySound(S_JUMP1 + (rand() % 3), FALSE, TRUE, g_Player.m_Control.m_vPosition);
            }
        }
    }

    if(!IsFreeForPlayer(D3DXVECTOR3(0.0f, -g_Player.m_fFalling * fUnit - 0.1f, 0.0f), FALSE, 1))
    {
        if(!g_Player.m_bOnGround)
        {
            g_Player.m_bOnGround = TRUE;

            if(g_Player.m_fFalling > PLAYER_BEGIN_FALL) g_Player.m_fStep = 2.0f;
            else g_Player.m_fStep = 3.0f;
            fOldStep = g_Player.m_fStep;

            if(g_Player.m_fFalling > 0.0f)
            {
                g_Player.m_fJumpCountDown = JUMP_COUNT_DOWN;

                if(g_fDieCountDown > DIE_DELAY * 0.75f)
                {
                    PlaySound(S_STEP_LEFT, FALSE, TRUE, g_Player.m_Control.m_vPosition);
                    PlaySound(S_STEP_RIGHT, FALSE, TRUE, g_Player.m_Control.m_vPosition);
                }

                x = (int)(g_Player.m_Control.m_vPosition.x);
                y = (int)(g_Player.m_Control.m_vPosition.y - g_Player.m_fFalling - 0.1f);
                z = (int)(-g_Player.m_Control.m_vPosition.z);

                CheckForSwitches(x, y, z, FALSE);

                iFlags = GetFieldFlags(x, y, z);
                if(iFlags & FF_SPRING)
                {
                    g_Player.m_fFalling *= -1.5f;
                    if(-g_Player.m_fFalling > MAX_SPRING_SPEED) g_Player.m_fFalling = -MAX_SPRING_SPEED;

                    PlaySound(S_SPRING, FALSE, TRUE, g_Player.m_Control.m_vPosition);
                }
                else
                {
                    if(g_Player.m_fFalling > PLAYER_BEGIN_HURT_FALL)
                    {
                        g_Player.m_fEnergy -= g_Player.m_fFalling - PLAYER_BEGIN_HURT_FALL;
                        PlaySound(S_LOOSE_ENERGY1 + (rand() % 2), FALSE, TRUE, g_Player.m_Control.m_vPosition);
                    }
                }
            }
        }
    }
    else g_Player.m_bOnGround = FALSE;

    if(MakeMove(D3DXVECTOR3(0.0f, -g_Player.m_fFalling * fUnit, 0.0f)))
    {
        g_Player.m_fFalling += GRAVITY * fUnit;
    }
    else
    {
        if(g_Player.m_fFalling < 0.0f) g_Player.m_fFalling = 0.05f;
        else
        {
            g_Player.m_fFalling = 0.0f;
        }
    }

    if(g_Player.m_bOnGround)
    {
        if(g_Player.m_fJumpCountDown > 0.0f) g_Player.m_fJumpCountDown -= fUnit;
    }

    if(!IsFreeForPlayer(D3DXVECTOR3(0.0f, 0.0f, 0.0f), TRUE, 2))
    {
        if(IsFreeForPlayer(D3DXVECTOR3(0.0f, 0.0f, 0.0f), TRUE, 4))
        {
            g_Player.m_Control.TranslateZ(0.05f * fUnit);
        }
        
        if(IsFreeForPlayer(D3DXVECTOR3(0.0f, 0.0f, 0.0f), TRUE, 3))
        {
            g_Player.m_Control.TranslateZ(-0.05f * fUnit);
        }
    }

    iObject = GetRadiusObject(g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.5f, 0.0f), 0.1f);

    if(iObject != -1)
    {
        if(g_Object[iObject].m_iType == OT_DIAMOND)
        {
            g_Object[iObject].m_bExists = FALSE;
            g_Player.m_iNumDiamonds++;

            if(g_Player.m_iNumDiamonds == g_iLevelDiamonds) ShowLevelExit();

            PlaySound(S_DIAMOND, FALSE, TRUE, g_Object[iObject].m_Control.m_vPosition);
        }
        else if(g_Object[iObject].m_iType == OT_LEVEL_EXIT)
        {
            if(!g_bLevelFinished)
            {
                g_bLevelFinished = TRUE;
                g_fLevelFinished = LEVEL_FINISH_DELAY;

                g_pSound[S_LEVEL_EXIT_SIGNAL][0]->Stop();
                PlaySound(S_LEVEL_FINISHED, FALSE, TRUE, g_Object[iObject].m_Control.m_vPosition);
            }
        }
    }

    iObject = GetRadiusObject(g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.5f, 0.0f), 0.4f);

    if(iObject != -1)
    {
        if(g_Object[iObject].m_iType == OT_SNOWBALL)
        {
            AddEnergy(-0.1f);
            vDirection = g_Object[iObject].m_vDirection;
            vDirection.y = 0.0f;
            if(IsFreeForPlayer(vDirection, FALSE, 1, FALSE)) g_Player.m_Control.m_vPosition += vDirection;
            g_Object[iObject].m_fLife = -1.0f;
        }
    }

    iObject = GetRadiusObject(g_Player.m_Control.m_vPosition - D3DXVECTOR3(0.0f, 0.5f, 0.0f), 0.0f);

    if(iObject != -1)
    {
        if(g_Player.m_bOnGround &&
           (g_Object[iObject].m_pType->m_iFlags & OF_SPIKE))
        {
            AddEnergy(-0.1f);
        }
    }

    iObject = GetRadiusObject(g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.5f, 0.0f), 0.3f);

    if(iObject != -1)
    {
        if(g_Object[iObject].m_pType->m_iFlags & OF_FIRE)
        {
            AddEnergy(-0.1f);
        }
    }

    iObject = GetObject(g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 1.5f, 0.0f));
    
    if(iObject != -1)
    {
        if(g_Object[iObject].m_iType == OT_BLOCK1)
        {
            AddEnergy((float)(fabs(g_Object[iObject].m_vDirection.y)) * -2.0f);
        }
    }

    g_Player.m_fSpeed *= 1.0f - (fSpeedLowering * fUnit);

    UpdatePlayerField();
    FieldType = GetFieldType(g_Player.m_iFX, g_Player.m_iFY, g_Player.m_iFZ);

    iStep = (int)(g_Player.m_fStep);
    iOldStep = (int)(fOldStep);

    if(g_Player.m_bOnGround)
    {
        if(iStep == 1 && iOldStep != 1)
        {
            PlaySound(S_STEP_RIGHT, FALSE, TRUE, g_Player.m_Control.m_vPosition);
        }
        else if(iStep == 5 && iOldStep != 5)
        {
            PlaySound(S_STEP_LEFT, FALSE, TRUE, g_Player.m_Control.m_vPosition);
        }
    }

    if(g_Player.m_fImmortality > 0.0f) g_Player.m_fImmortality -= fUnit;
}

void MoveObjects(float fUnit)
{
    int         iObject;
    int         iFlags;
    int         x;
    int         y;
    int         z;
    float       fSpeed;
    CFieldType  FieldType;
    D3DXVECTOR3 vScreen;
    D3DXMATRIX  mCameraMatrix;

    for(int a = 0; a < MAX_OBJECTS; a++)
    {
        if(g_Object[a].m_bExists)
        {
            if(ObjectFree(g_Object[a], g_Object[a].m_Control.m_vPosition - D3DXVECTOR3(0.0f, 0.01f, 0.0f), g_bObjectsOutOfMap))
            {
                g_Object[a].m_bOnGround = FALSE;
            }
            else g_Object[a].m_bOnGround = TRUE;

            if(g_Object[a].m_iType != OT_SNOWBALL)
            {
                if(ObjectFree(g_Object[a], g_Object[a].m_Control.m_vPosition + g_Object[a].m_vDirection))
                {
                    g_Object[a].m_Control.m_vPosition += g_Object[a].m_vDirection * fUnit;

                    if(g_Player.m_bOnGround)
                    {
                        iObject = GetObject(g_Player.m_Control.m_vPosition - D3DXVECTOR3(0.0f, 0.1f, 0.0f));
                        if(iObject == a)
                        {
                            if(IsFreeForPlayer(g_Object[a].m_vDirection, FALSE, 1, FALSE))
                            {
                                g_Player.m_Control.m_vPosition += g_Object[a].m_vDirection * fUnit;
                            }
                        }
                    }
                }
                else
                {
                    g_Object[a].m_vDirection *= -0.5f;
                    if(D3DXVec3Length(&g_Object[a].m_vDirection) > 0.05f)
                    {
                        if(g_Object[a].m_iType == OT_BLOCK1)
                        {
                            PlaySound(S_BLOCK_FALLING, FALSE, TRUE, g_Object[a].m_Control.m_vPosition);

                            x = (int)(g_Object[a].m_Control.m_vPosition.x);
                            y = (int)(g_Object[a].m_Control.m_vPosition.y - g_Object[a].m_pType->m_fHeight / 2.0f - g_Object[a].m_vDirection.y - 0.2f);
                            z = (int)(-g_Object[a].m_Control.m_vPosition.z);

                            iFlags = GetFieldFlags(x, y, z);
                            if(iFlags & FF_SPRING)
                            {
                                g_Object[a].m_vDirection.y *= 3.0f;
                                if(g_Object[a].m_vDirection.y > MAX_SPRING_SPEED) g_Object[a].m_vDirection.y = MAX_SPRING_SPEED;

                                PlaySound(S_SPRING, FALSE, TRUE, g_Object[a].m_Control.m_vPosition);
                            }
                        }
                    }
                }
            }

            if(!g_Object[a].m_bOnGround)
            {
                if(!(g_Object[a].m_pType->m_iFlags & OF_FLYING))
                {
                    g_Object[a].m_vDirection -= D3DXVECTOR3(0.0f, GRAVITY, 0.0f) * fUnit;
                }
            }
            else
            {
                if(!(g_Object[a].m_pType->m_iFlags & OF_FLYING))
                {
                    if(fabs(g_Object[a].m_vDirection.y) < 0.05f)
                    {
                        g_Object[a].m_Control.m_vPosition.y = (float)((int)(g_Object[a].m_Control.m_vPosition.y + 0.6f)) - 0.5f;
                        g_Object[a].m_vDirection.y = 0.0f;
                    }
                }
            }

            if(g_Object[a].m_iType == OT_BLOCK1)
            {
                if(!ObjectFree(g_Object[a], g_Object[a].m_Control.m_vPosition, FALSE))
                {
                    x = (int)(g_Object[a].m_Control.m_vPosition.x);
                    y = (int)(g_Object[a].m_Control.m_vPosition.y - g_Object[a].m_pType->m_fHeight / 2.0f + 0.1f);
                    z = (int)(-g_Object[a].m_Control.m_vPosition.z);

                    FieldType = GetFieldType(x, y, z);
                    if(FieldType.m_iType != 255)
                    {
                        g_Object[a].m_Control.m_vPosition.y += FieldType.m_vMaximum.y;
                        g_Object[a].m_vDirection.y = 0.0f;
                    }
                }
            }

            if(g_Object[a].m_pType->m_iFlags & OF_SWINGING_X) g_Object[a].m_vDirection.x = (float)(sin(g_fCounter1 / 40.0f + a)) / 50.0f;
            if(g_Object[a].m_pType->m_iFlags & OF_SWINGING_Z) g_Object[a].m_vDirection.z = (float)(sin(g_fCounter1 / 40.0f + a)) / 50.0f;

            if(g_Object[a].m_iType == OT_BLOCK1)
            {
                x = (int)(g_Object[a].m_Control.m_vPosition.x);
                y = (int)(g_Object[a].m_Control.m_vPosition.y - 0.1f);
                z = (int)(-g_Object[a].m_Control.m_vPosition.z);

                CheckForSwitches(x, y, z, TRUE);
            }
            else if(g_Object[a].m_iType == OT_DIAMOND)
            {
                g_Object[a].m_Control.RotateY(0.04f * fUnit);
                g_Object[a].m_vAdd = D3DXVECTOR3(0.0f, (float)(sin(g_Object[a].m_fCounter)) * 0.1f, 0.0f);
                g_Object[a].m_fCounter += 0.1f * fUnit;
            }
            else if(g_Object[a].m_iType == OT_CANNON1)
            {
                D3DXMatrixLookAtLH(&mCameraMatrix, &g_Object[a].m_Control.m_vPosition,
                                                   &D3DXVECTOR3(g_Object[a].m_Control.m_vPosition + g_Object[a].m_Control.m_vAxis[2]),
                                                   &g_Object[a].m_Control.m_vAxis[1]);
                D3DXVec3TransformCoord(&vScreen, &g_Player.m_Control.m_vPosition, &mCameraMatrix);

                if(vScreen.x > 0.1f) g_Object[a].m_Control.RotateY(0.01f);
                else if(vScreen.x < -0.1f) g_Object[a].m_Control.RotateY(-0.01f);
                else
                {
                    if(g_Object[a].m_fReload <= 0.0f)
                    {
                        iObject = CreateObject(OT_SNOWBALL, g_Object[a].m_Control.GetRelativePosition(D3DXVECTOR3(0.0f, 0.1f, 0.5f)));

                        if(iObject != -1)
                        {
                            fSpeed = D3DXVec3Length(&D3DXVECTOR3(g_Object[a].m_Control.m_vPosition - g_Player.m_Control.m_vPosition)) / FloatRandom(10.0f, 14.0f);
                            g_Object[iObject].m_vDirection = g_Object[a].m_Control.m_vAxis[2] / 2.0f + D3DXVECTOR3(0.0f, FloatRandom(0.1f, 0.15f), 0.0f);
                            g_Object[iObject].m_vDirection *= fSpeed;
                            g_Object[iObject].m_iParent = a;
                            g_Object[a].m_fReload = g_Object[a].m_pType->m_fReload;
                            g_Object[a].m_vAdd = g_Object[a].m_Control.m_vAxis[2] / -3.0f;
                            PlaySound(S_SHOOT, FALSE, TRUE, g_Object[a].m_Control.GetRelativePosition(D3DXVECTOR3(0.0f, 0.1f, 0.5f)));
                        }
                    }
                }

                if(g_Object[a].m_fReload > 0.0f) g_Object[a].m_fReload -= fUnit;
                g_Object[a].m_vAdd *= 1.0f - (0.05f * fUnit);
            }
            else if(g_Object[a].m_iType == OT_SNOWBALL)
            {
                if(ObjectFree(g_Object[a], g_Object[a].m_Control.m_vPosition, g_bObjectsOutOfMap) &&
                   g_Object[a].m_Control.m_vPosition.y >= g_fGroundHeight &&
                   g_Object[a].m_fLife >= 0.0f)
                {
                    g_Object[a].m_Control.m_vPosition += g_Object[a].m_vDirection * fUnit;
                }
                else
                {
                    g_Object[a].m_bExists = FALSE;
                    for(int b = 0; b < 40; b++)
                    {
                        CreateParticle(PT_SNOW, g_Object[a].m_Control.m_vPosition, FloatRandom(10.0f, 30.0f), FloatRandom(0.1f, 0.3f));
                    }

                    PlaySound(S_SNOWBALL_HIT, FALSE, TRUE, g_Object[a].m_Control.m_vPosition);
                }
            }

            g_Object[a].m_vDirection *= 1.0f - (0.01f * fUnit);
        }
    }
}

void MoveSnow(float fUnit)
{
    for(int a = 0; a < g_iNumSnowflakes; a++)
    {
        g_Snowflake[a].m_vPosition -= D3DXVECTOR3(FloatRandom(-0.01f, 0.01f), 0.1f, FloatRandom(-0.01f, 0.01f)) * fUnit;

        if(g_Snowflake[a].m_vPosition.y < g_fGroundHeight ||
           !IsFree(g_Snowflake[a].m_vPosition, TRUE, TRUE))
        {
            g_Snowflake[a].m_vPosition.x = FloatRandom(-5.0f, (float)(g_iLevelWidth) + 5.0f);
            g_Snowflake[a].m_vPosition.y = FloatRandom(20.0f, 40.0f);
            g_Snowflake[a].m_vPosition.z = FloatRandom((float)(-g_iLevelDepth) - 5.0f, 5.0f);
        }
    }
}

void MoveParticles(float fUnit)
{
    for(int a = 0; a < MAX_PARTICLES; a++)
    {
        if(g_Particle[a].m_fLife > 0.0f)
        {
            g_Particle[a].m_vPosition += g_Particle[a].m_vDirection * fUnit;
            g_Particle[a].m_vDirection *= 1.0f - (0.01f * fUnit);
            g_Particle[a].m_vDirection.y -= GRAVITY * fUnit;
            g_Particle[a].m_fLife -= fUnit;

            if(g_Particle[a].m_vPosition.y < g_fGroundHeight) g_Particle[a].m_fLife = 0.0f;
        }
    }
}

void MoveMenu(float fUnit)
{
    BOOL    bKey;
    int     iKey;

    if(g_fCounter1 < 200.0f)
    {
        if(g_vCameraDestination.y > 30.0f) g_vCameraDestination.y -= fUnit;
        else g_vCameraDestination.y = 30.0f;
    }
    else
    {
        g_vCameraDestination = g_vCamera;
        g_vCameraDestination.y = 30.0f;
        g_vCameraDestination.z = -10.0f;

        if(g_fCounter1 < 210.0f)
        {
            g_vLookAt.z -= fUnit;
        }

        if(g_fMenuCursorCountDown <= 0.0f)
        {
            bKey = FALSE;

            if(g_cKeyState[DIK_UP]) {g_iMenuCursor--; g_bEnterPassword = FALSE; bKey = TRUE;}
            if(g_cKeyState[DIK_DOWN]) {g_iMenuCursor++; g_bEnterPassword = FALSE; bKey = TRUE;}
            if(g_cKeyState[DIK_RETURN]) {g_iMenuItemSelected = g_iMenuCursor; bKey = TRUE;}
            if(g_cKeyState[DIK_ESCAPE] && g_bEnterPassword) {g_bEnterPassword = FALSE; bKey = TRUE;}
            if(g_bEnterPassword)
            {
                iKey = -1;

                if(g_cKeyState[DIK_A]) iKey = 'A';
                if(g_cKeyState[DIK_B]) iKey = 'B';
                if(g_cKeyState[DIK_C]) iKey = 'C';
                if(g_cKeyState[DIK_D]) iKey = 'D';
                if(g_cKeyState[DIK_E]) iKey = 'E';
                if(g_cKeyState[DIK_F]) iKey = 'F';
                if(g_cKeyState[DIK_G]) iKey = 'G';
                if(g_cKeyState[DIK_H]) iKey = 'H';
                if(g_cKeyState[DIK_I]) iKey = 'I';
                if(g_cKeyState[DIK_J]) iKey = 'J';
                if(g_cKeyState[DIK_K]) iKey = 'K';
                if(g_cKeyState[DIK_L]) iKey = 'L';
                if(g_cKeyState[DIK_M]) iKey = 'M';
                if(g_cKeyState[DIK_N]) iKey = 'N';
                if(g_cKeyState[DIK_O]) iKey = 'O';
                if(g_cKeyState[DIK_P]) iKey = 'P';
                if(g_cKeyState[DIK_Q]) iKey = 'Q';
                if(g_cKeyState[DIK_R]) iKey = 'R';
                if(g_cKeyState[DIK_S]) iKey = 'S';
                if(g_cKeyState[DIK_T]) iKey = 'T';
                if(g_cKeyState[DIK_U]) iKey = 'U';
                if(g_cKeyState[DIK_V]) iKey = 'V';
                if(g_cKeyState[DIK_W]) iKey = 'W';
                if(g_cKeyState[DIK_X]) iKey = 'X';
                if(g_cKeyState[DIK_Y]) iKey = 'Z';
                if(g_cKeyState[DIK_Z]) iKey = 'Y';

                if(iKey != -1 &&
                   g_iPasswordCursor < 254)
                {
                    g_sPassword[g_iPasswordCursor] = iKey;
                    g_sPassword[g_iPasswordCursor + 1] = 0;
                    g_iPasswordCursor++;
                    bKey = TRUE;
                }

                if(g_cKeyState[DIK_BACKSPACE])
                {
                    if(g_iPasswordCursor > 0)
                    {
                        g_iPasswordCursor--;
                        g_sPassword[g_iPasswordCursor] = 0;
                        bKey = TRUE;
                    }
                }
            }

            if(g_iMenuCursor < 0) g_iMenuCursor = 3;
            if(g_iMenuCursor > 3) g_iMenuCursor = 0;

            if(bKey) g_fMenuCursorCountDown = MENU_CURSOR_DELAY;
        }
        else g_fMenuCursorCountDown -= fUnit;
    }

    g_vLookAt = g_vCamera - D3DXVECTOR3(0.0f, 1.0f, 0.0f);

    SetCamera(g_vCamera, g_vLookAt, D3DXVECTOR3(0.0f, 0.0f, 1.0f), g_pD3DDevice);
}

void Move(float fUnit)
{
    D3DXVECTOR3 vCameraDifference;

    g_ObjectType[OT_FIRE].m_pMesh->m_pTexture[0] = g_pFireSprite[(int)(g_fCounter1 / 2.0f) % 4];
    if(g_fEnergyDisplay != g_Player.m_fEnergy) g_fEnergyDisplay += (g_Player.m_fEnergy - g_fEnergyDisplay) / 10.0f * fUnit;

    if(g_bMenu) MoveMenu(fUnit);
    else MovePlayer(fUnit);
    MoveObjects(fUnit);
    if(g_bSnowing) MoveSnow(fUnit);
    MoveParticles(fUnit);
    g_Script.Execute(fUnit);

    vCameraDifference = g_vCameraDestination - g_vCamera;
    if(g_iViewMode == 1) g_vCamera += (vCameraDifference / 10.0f) * fUnit;
    else g_vCamera += (vCameraDifference / 35.0f) * fUnit;
    
    if(g_bLevelFinished) g_fLevelFinished -= fUnit;
    if(g_fLockCameraCountDown > 0.0f) g_fLockCameraCountDown -= fUnit;
    if(g_fMoveCameraCountDown > 0.0f) g_fMoveCameraCountDown -= fUnit;
    if(g_fConfirmQuitCountDown > 0.0f) g_fConfirmQuitCountDown -= fUnit;

    if(g_iLevelNr == 7)
    {
        g_FieldType[1].m_pMesh->m_pTexture[0] = g_FieldType[1].m_pMesh->m_pTexture[1 + (int)(g_fCounter2)];
    }

    g_fCounter1 += fUnit;
    if(g_fCounter1 >= 1000000.0f) g_fCounter1 = 0.0f;

    g_fCounter2 += fUnit;
    if(g_fCounter2 >= 8.0f) g_fCounter2 = 0.0f;
}

void SetCamera()
{
    if(g_fMoveCameraCountDown <= 0.0f)
    {
        if(g_fCameraHeight < MIN_CAMERA_HEIGHT) g_fCameraHeight = MIN_CAMERA_HEIGHT;
        if(g_fCameraHeight > MAX_CAMERA_HEIGHT) g_fCameraHeight = MAX_CAMERA_HEIGHT;

        if(g_fCameraDistance > MAX_CAMERA_DISTANCE) g_fCameraDistance = MAX_CAMERA_DISTANCE;
        if(g_fCameraDistance < -MAX_CAMERA_DISTANCE) g_fCameraDistance = -MAX_CAMERA_DISTANCE;

        if(g_iViewMode == 0)
        {
            g_fCameraRadius = -g_fCameraDistance;

            if(g_bCameraLocked)
            {
                g_vCameraDestination = g_Player.m_Control.m_vPosition + D3DXVECTOR3((float)(sin(g_fCameraAngle)) * g_fCameraRadius, 1.0f + (float)(fabs(g_fCameraDistance / 2.0f)) + g_fCameraHeight, (float)(cos(g_fCameraAngle)) * g_fCameraRadius);
            }
            else
            {
                g_vCameraDestination = g_Player.m_Control.GetRelativePosition(D3DXVECTOR3((float)(sin(g_fCameraAngle)) * g_fCameraRadius, 1.0f + (float)(fabs(g_fCameraDistance / 2.0f)) + g_fCameraHeight, (float)(cos(g_fCameraAngle)) * g_fCameraRadius));
            }

            g_vLookAt = g_Player.m_Control.m_vPosition;
            g_vCameraUp = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
        }
        else if(g_iViewMode == 1)
        {
            g_vCameraDestination = g_View1Camera.m_vPosition;
            g_vLookAt = g_View1Camera.m_vPosition + g_View1Camera.m_vAxis[2];
            g_vCameraUp = g_View1Camera.m_vAxis[1];
        }
                
        if(!g_bCameraLocked)
        {
            if(g_bObjectsOutOfMap)
            {
                if(RayIntersect(g_vCameraDestination, g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE))
                {
                    if(!RayIntersect(g_vCameraDestination + D3DXVECTOR3(0.0f, 2.0f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE) &&
                                     g_fCameraHeight <= MAX_CAMERA_HEIGHT - 2.0f)
                    {
                        g_fCameraHeight += 2.0f;
                    }
                    else
                    {
                        if(g_fCameraDistance > -STANDARD_CAMERA_DISTANCE + 0.5f) g_fCameraDistance -= 0.1f;
                        if(!RayIntersect(g_vCameraDestination - D3DXVECTOR3(0.0f, 2.0f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE))
                        {
                            g_fCameraHeight -= 2.0f;
                        }
                    }
                }
                else
                {
                    if(g_fCameraDistance < STANDARD_CAMERA_DISTANCE - 0.5f) g_fCameraDistance += 0.1f;
                    if(!RayIntersect(g_vCameraDestination - D3DXVECTOR3(0.0f, 2.0f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE) &&
                                     g_fCameraHeight >= MIN_CAMERA_HEIGHT + 2.0f)
                    {
                        g_fCameraHeight -= 2.0f;
                    }
                }
            }
            else
            {
                if(RayIntersect(g_vCameraDestination, g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE))
                {
                    if(!RayIntersect(g_vCameraDestination + D3DXVECTOR3(0.0f, 2.0f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE) &&
                                     g_fCameraHeight <= MAX_CAMERA_HEIGHT - 2.0f)
                    {
                        g_fCameraHeight += 2.0f;
                    }
                    else if(!RayIntersect(g_vCameraDestination - D3DXVECTOR3(0.0f, 2.0f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE) &&
                                          g_fCameraHeight >= MIN_CAMERA_HEIGHT + 2.0f)
                    {
                        g_fCameraHeight -= 2.0f;
                    }
                    else if(g_fCameraDistance > -STANDARD_CAMERA_DISTANCE) g_fCameraDistance -= 0.1f;
                }
                else
                {
                    if(g_fCameraDistance < STANDARD_CAMERA_DISTANCE - 0.5f) g_fCameraDistance += 0.1f;
                    if(g_fCameraDistance > STANDARD_CAMERA_DISTANCE + 0.5f) g_fCameraDistance -= 0.1f;
                    if(g_fCameraHeight > 0.5f && !RayIntersect(g_vCameraDestination - D3DXVECTOR3(0.0f, 0.1f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE)) g_fCameraHeight -= 0.1f;
                    if(g_fCameraHeight < -0.5f && !RayIntersect(g_vCameraDestination + D3DXVECTOR3(0.0f, 0.1f, 0.0f), g_Player.m_Control.m_vPosition + D3DXVECTOR3(0.0f, 0.75f, 0.0f), NULL, TRUE)) g_fCameraHeight += 0.1f;
                }
            }
        }

        if(!g_bObjectsOutOfMap)
        {
            if(g_vCameraDestination.x < 0.1f) g_vCameraDestination.x = 0.1f;
            if(g_vCameraDestination.y < 0.1f) g_vCameraDestination.y = 0.1f;
            if(g_vCameraDestination.z > -0.1f) g_vCameraDestination.z = -0.1f;
            if(g_vCameraDestination.x >= (float)(g_iLevelWidth) + 0.9f) g_vCameraDestination.x = (float)(g_iLevelWidth) + 0.9f;
            if(g_vCameraDestination.y >= (float)(g_iLevelHeight) + 0.9f) g_vCameraDestination.y = (float)(g_iLevelHeight) + 0.9f;
            if(g_vCameraDestination.z <= (float)(-g_iLevelDepth) - 0.9f) g_vCameraDestination.z = (float)(-g_iLevelDepth) - 0.9f;
        }
    }
    else
    {
        g_vCameraUp = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
    }

    g_mCamera = SetCamera(g_vCamera,
                          g_vLookAt,
                          g_vCameraUp,
                          g_pD3DDevice);

    g_vCameraDepth = g_vLookAt - g_vCamera;
    D3DXVec3Normalize(&g_vCameraDepth, &g_vCameraDepth);

    g_vCameraDifference = g_vCameraDestination - g_vCamera;

    SetListener(g_pListener,
                g_vCamera,
                D3DXVECTOR3(0.0f, 0.0f, 0.0f),
                g_vCameraUp,
                g_vCameraDepth,
                1.0f, 1.0f, 10.0f);
}

INT WINAPI WinMain(HINSTANCE hInstance, HINSTANCE, char*, int)
{
    DWORD       dwStartTime;
    DWORD       dwEndTime;
    DWORD       dwDuration;
    MSG         Message;
    WNDCLASSEX  WindowClass = {sizeof(WNDCLASSEX), CS_CLASSDC, MessageProc, 0, 0, 
                               hInstance, LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON)), NULL, NULL, NULL,
                               "D3D Window", NULL};

    DeleteFile("Debug.txt");

    srand(timeGetTime());
    
    g_hInstance = WindowClass.hInstance;

    RegisterClassEx(&WindowClass);

    Debug("Creating the window...\n");
    g_hWindow = CreateWindow("D3D Window", "Blocks2001",
                             WS_OVERLAPPEDWINDOW, 0, 0, (int)(g_fResolutionX), (int)(g_fResolutionY),
                             GetDesktopWindow(), NULL, WindowClass.hInstance, NULL);

    SetTimer(g_hWindow, 0, 20, NULL);
    ShowCursor(FALSE);

    g_mWhiteMaterial.Ambient.r = 1.0f;
    g_mWhiteMaterial.Ambient.g = 1.0f;
    g_mWhiteMaterial.Ambient.b = 1.0f;
    g_mWhiteMaterial.Diffuse.r = 1.0f;
    g_mWhiteMaterial.Diffuse.g = 1.0f;
    g_mWhiteMaterial.Diffuse.b = 1.0f;

    Debug("Initializing the application...\n");
    if(SUCCEEDED(InitApplication()))
    {
        Debug("Initialization complete!\n");
        ShowWindow(g_hWindow, SW_SHOWDEFAULT);
        UpdateWindow(g_hWindow);

        Debug("Starting the game...\n");

        g_iLevelNr = 0;

NewLevel:
        for(int a = 0; a < NUM_SOUNDS; a++)
        {
            for(int b = 0; b < NUM_SOUND_CLONES; b++)
            {
                g_pSound[a][b]->Stop();
            }
        }

        ZeroMemory(&Message, sizeof(Message));

        InitLevel();
        InitLevelFieldTypes();

        while(Message.message != WM_QUIT)
        {
            if(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
            {
                TranslateMessage(&Message);
                DispatchMessage(&Message);
            }
            else
            {
                if(g_Player.m_fEnergy <= 0.0f &&
                   g_fDieCountDown <= 0.0f) goto NewLevel;

                if(g_bMenu)
                {
                    if(g_iMenuItemSelected != -1)
                    {
                        if(g_iMenuItemSelected == 0)
                        {
                            g_iLevelNr = 1;
                            goto NewLevel;
                        }
                        else if(g_iMenuItemSelected == 1)
                        {
                            if(!g_bEnterPassword)
                            {
                                g_bEnterPassword = TRUE;
                                strcpy(g_sPassword, "");
                                g_iPasswordCursor = 0;
                            }
                            else
                            {
                                for(int a = 1; a <= 8; a++)
                                {
                                    if(!strcmp(g_sPassword,
                                               g_sLevelPassword[a]))
                                    {
                                        g_iLevelNr = a;
                                        goto NewLevel;
                                    }
                                }
                                g_bEnterPassword = FALSE;
                            }
                        }
                        else if(g_iMenuItemSelected == 2)
                        {
                            Debug("Exiting...\n");
                            KillTimer(g_hWindow, 0);
                            CleanUp();
                            exit(0);
                        }

                        g_iMenuItemSelected = -1;
                    }

                    if(g_cKeyState[DIK_ESCAPE])
                    {
                        Debug("Exiting...\n");
                        KillTimer(g_hWindow, 0);
                        CleanUp();
                        exit(0);
                    }
                }

                if(g_bConfirmQuit)
                {
                    if(g_cKeyState[DIK_RETURN])
                    {
                        g_iLevelNr = 0;
                        g_bConfirmQuit = FALSE;
                        goto NewLevel;
                    }
                    else if(g_cKeyState[DIK_ESCAPE] &&
                            g_fConfirmQuitCountDown <= 0.0f)
                    {
                        g_bConfirmQuit = FALSE;
                        g_fConfirmQuitCountDown = CONFIRM_QUIT_DELAY;
                    }
                }
                else if(g_bConfirmRestart)
                {
                    if(g_cKeyState[DIK_RETURN])
                    {
                        g_bConfirmRestart = FALSE;
                        goto NewLevel;
                    }
                    else if(g_cKeyState[DIK_ESCAPE])
                    {
                        g_bConfirmRestart = FALSE;
                        g_fConfirmQuitCountDown = CONFIRM_QUIT_DELAY;
                    }
                }
                
                if(!g_bMenu) SetCamera();

                dwStartTime = timeGetTime();
                Render();
                dwEndTime = timeGetTime();
                dwDuration = dwEndTime - dwStartTime;

                if(!g_bPaused)
                {
                    if(dwDuration >= 25)
                    {
                        for(int a = 0; a < (int)((float)(dwDuration) / 5.0f); a++)
                        {
                            Move(0.2f);
                        }
                    }
                    else
                    {
                        Delay(25 - dwDuration);
                        Move(1.0f);
                    }

                    if(g_fLevelFinished <= -LEVEL_FINISH_DELAY)
                    {
                        g_bPaused = FALSE;

                        if(g_iLevelNr != 8)
                        {
                            g_iLevelNr++;
                            goto NewLevel;
                        }
                        else
                        {
                            g_Player.m_Control.m_vPosition.y += 0.05f;
                            g_Player.m_fJumpCountDown = 1;
                            g_Player.m_fFalling = 0.0f;
                            if(!g_bCreditsMusicPlaying)
                            {
                                PlayMusic("Data\\Credits.mid", TRUE);
                                g_bCreditsMusicPlaying = TRUE;
                            }
                        }
                    }
                }

                if(!g_bMenu &&
                   g_cKeyState[DIK_P])
                {
                    g_bPaused = !g_bPaused;
                    Delay(100);
                }

                if(g_bMenu) g_bPaused = FALSE;
            }
        }
    }
    else
    {
        Debug("ERROR: Could not initialize the application!\n");
        ShowCursor(TRUE);
        MessageBox(g_hWindow, "Das Spiel konnte nicht gestartet werden.\nUm weitere Informationen zu erhalten, lesen Sie die Datei 'DEBUG.TXT'.", "Fehler", MB_OK);
        CleanUp();
    }

    return 0;
}

void CScript::Execute(float fUnit)
{
    int     a;
    BOOL    bNext;

ExecuteCommand:

    a = m_iCurrentCommand;

    bNext = FALSE;

    if(m_Command[a].m_bExists)
    {
        if(m_Command[a].m_iType == CT_SET_FIELD)
        {
            g_iLevel[m_Command[a].m_iParameter[0]]
                    [m_Command[a].m_iParameter[1]]
                    [m_Command[a].m_iParameter[2]] = m_Command[a].m_iParameter[3];

            InitFieldFlagsArray();

            bNext = TRUE;
        }
        else if(m_Command[a].m_iType == CT_SET_CAMERA)
        {
            g_vCameraDestination = D3DXVECTOR3(m_Command[a].m_fParameter[0],
                                               m_Command[a].m_fParameter[1],
                                               m_Command[a].m_fParameter[2]);

            g_vLookAt = D3DXVECTOR3(m_Command[a].m_fParameter[3],
                                    m_Command[a].m_fParameter[4],
                                    m_Command[a].m_fParameter[5]);

            g_fMoveCameraCountDown = m_Command[a].m_fParameter[6];
            bNext = TRUE;
        }
        else if(m_Command[a].m_iType == CT_DELAY)
        {
            m_Command[a].m_fParameter[0] -= fUnit;
            if(m_Command[a].m_fParameter[0] <= 0.0f) bNext = TRUE;
        }
        else if(m_Command[a].m_iType == CT_PLAY_SOUND)
        {
            PlaySound(m_Command[a].m_iParameter[0], FALSE, TRUE,
                      D3DXVECTOR3(m_Command[a].m_fParameter[0],
                                  m_Command[a].m_fParameter[1],
                                  m_Command[a].m_fParameter[2]));
            bNext = TRUE;
        }
    }

    if(bNext)
    {
        m_iCurrentCommand++;
        goto ExecuteCommand;
    }
}

void CScript::AddSetFieldCommand(int x,
                                 int y,
                                 int z,
                                 int iField)
{
    int a;
    
    a = GetFreeCommand();
    if(a != -1)
    {
        m_Command[a].m_bExists = TRUE;
        m_Command[a].m_iType = CT_SET_FIELD;
        m_Command[a].m_iParameter[0] = x;
        m_Command[a].m_iParameter[1] = y;
        m_Command[a].m_iParameter[2] = z;
        m_Command[a].m_iParameter[3] = iField;
    }
}

void CScript::AddSetCameraCommand(D3DXVECTOR3 vCamera,
                                  D3DXVECTOR3 vLookAt,
                                  float fDuration)
{
    int a;

    a = GetFreeCommand();
    if(a != -1)
    {
        m_Command[a].m_bExists = TRUE;
        m_Command[a].m_iType = CT_SET_CAMERA;
        m_Command[a].m_fParameter[0] = vCamera.x;
        m_Command[a].m_fParameter[1] = vCamera.y;
        m_Command[a].m_fParameter[2] = vCamera.z;
        m_Command[a].m_fParameter[3] = vLookAt.x;
        m_Command[a].m_fParameter[4] = vLookAt.y;
        m_Command[a].m_fParameter[5] = vLookAt.z;
        m_Command[a].m_fParameter[6] = fDuration;
    }
}

void CScript::AddDelayCommand(float fDelay)
{
    int a;

    a = GetFreeCommand();
    if(a != -1)
    {
        m_Command[a].m_bExists = TRUE;
        m_Command[a].m_iType = CT_DELAY;
        m_Command[a].m_fParameter[0] = fDelay;
    }
}

void CScript::AddPlaySoundCommand(int iSound,
                                  D3DXVECTOR3 vPosition)
{
    int a;

    a = GetFreeCommand();
    if(a != -1)
    {
        m_Command[a].m_bExists = TRUE;
        m_Command[a].m_iType = CT_PLAY_SOUND;
        m_Command[a].m_iParameter[0] = iSound;
        m_Command[a].m_fParameter[0] = vPosition.x;
        m_Command[a].m_fParameter[1] = vPosition.y;
        m_Command[a].m_fParameter[2] = vPosition.z;
    }
}

int CScript::GetFreeCommand()
{
    for(int a = 0; a < MAX_COMMANDS; a++)
    {
        if(!m_Command[a].m_bExists)
        {
            return a;
        }
    }

    return -1;
}